@@ -15,6 +15,8 @@ Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA. */
 #include "pari.h"
 #include "paripriv.h"
 
+#define DEBUGLEVEL DEBUGLEVEL_mat
+
 /***********************************************************************/
 /**                                                                   **/
 /**                             F2v                                   **/
@@ -81,29 +83,7 @@ F2v_slice(GEN x, long a, long b)
 /* x[a..b,], a <= b */
 GEN
 F2m_rowslice(GEN x, long a, long b)
-{
-  long i, l;
-  GEN y = cgetg_copy(x, &l);
-  for (i = 1; i < l; i++) gel(y,i) = F2v_slice(gel(x,i),a,b);
-  return y;
-}
-
-GEN
-F2m_to_ZM(GEN z)
-{
-  long i, l = lg(z);
-  GEN x = cgetg(l,t_MAT);
-  for (i=1; i<l; i++) gel(x,i) = F2c_to_ZC(gel(z,i));
-  return x;
-}
-GEN
-F2m_to_mod(GEN z)
-{
-  long i, l = lg(z);
-  GEN x = cgetg(l,t_MAT);
-  for (i=1; i<l; i++) gel(x,i) = F2c_to_mod(gel(z,i));
-  return x;
-}
+{ pari_APPLY_same(F2v_slice(gel(x,i),a,b)) }
 
 GEN
 F2v_to_Flv(GEN x)
@@ -118,20 +98,17 @@ F2v_to_Flv(GEN x)
 }
 
 GEN
-F2m_to_Flm(GEN z)
-{
-  long i, l = lg(z);
-  GEN x = cgetg(l,t_MAT);
-  for (i=1; i<l; i++) gel(x,i) = F2v_to_Flv(gel(z,i));
-  return x;
-}
+F2m_to_ZM(GEN x) { pari_APPLY_same(F2c_to_ZC(gel(x,i))) }
+GEN
+F2m_to_mod(GEN x) { pari_APPLY_same(F2c_to_mod(gel(x,i))) }
+GEN
+F2m_to_Flm(GEN x) { pari_APPLY_same(F2v_to_Flv(gel(x,i))) }
 
 GEN
 ZV_to_F2v(GEN x)
 {
-  long l = lg(x)-1;
+  long i, j, k, l = lg(x)-1;
   GEN z = cgetg(nbits2lg(l), t_VECSMALL);
-  long i,j,k;
   z[1] = l;
   for(i=1,k=1,j=BITS_IN_LONG; i<=l; i++,j++)
   {
@@ -296,6 +273,15 @@ F2v_hamming(GEN H)
   return n;
 }
 
+int
+F2v_subset(GEN x, GEN y)
+{
+  long i, n = lg(y);
+  for (i = 2; i < n; i ++)
+    if ((x[i] & y[i]) != x[i]) return 0;
+  return 1;
+}
+
 GEN
 matid_F2m(long n)
 {
@@ -413,12 +399,10 @@ F2m_ker_sp(GEN x, long deplin)
     if (j>m)
     {
       if (deplin) {
-        GEN c = zero_F2v(n);
+        GEN v = zero_F2v(n);
         for (i=1; i<k; i++)
-          if (F2v_coeff(xk, d[i]))
-            F2v_set(c, i);
-        F2v_set(c, k);
-        return c;
+          if (F2v_coeff(xk, d[i])) F2v_set(v, i);
+        F2v_set(v, k); return v;
       }
       r++; d[k] = 0;
     }
@@ -441,8 +425,7 @@ F2m_ker_sp(GEN x, long deplin)
   {
     GEN C = gel(y,j); while (d[k]) k++;
     for (i=1; i<k; i++)
-      if (d[i] && F2m_coeff(x,d[i],k))
-        F2v_set(C,i);
+      if (d[i] && F2m_coeff(x,d[i],k)) F2v_set(C,i);
     F2v_set(C, k);
   }
   return y;
