@@ -15,6 +15,8 @@ Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA. */
 #include "pari.h"
 #include "paripriv.h"
 
+#define DEBUGLEVEL DEBUGLEVEL_fflog
+
 /* Not so fast arithmetic with polynomials over F_2 */
 
 /***********************************************************************/
@@ -748,17 +750,19 @@ F2xn_inv_basecase1(ulong x)
   w = u&v; w ^= w >> 2; w ^= w >> 1; v = (w&1UL)|(v<<1);
   w = u&v; w ^= w >> 2; w ^= w >> 1; v = (w&1UL)|(v<<1);
   for (i=1;i<=4;i++)
-   { w = u&v; w ^= w >> 4; w ^= w >> 2; w ^= w >> 1; v = (w&1UL)|(v<<1); }
+  { w = u&v; w ^= w >> 4; w ^= w >> 2; w ^= w >> 1; v = (w&1UL)|(v<<1); }
   for (i=1;i<=8;i++)
-   { w = u&v; w ^= w >> 8; w ^= w >> 4; w ^= w >> 2; w ^= w >> 1; v = (w&1UL)|(v<<1); }
+  { w = u&v; w ^= w >> 8; w ^= w >> 4; w ^= w >> 2; w ^= w >> 1;
+    v = (w&1UL)|(v<<1); }
   for (i=1;i<=16;i++)
-   { w = u&v; w ^= w >> 16; w ^= w >> 8; w ^= w >> 4; w ^= w >> 2; w ^= w >> 1; v = (w&1UL)|(v<<1); }
+  { w = u&v; w ^= w >> 16; w ^= w >> 8; w ^= w >> 4; w ^= w >> 2; w ^= w >> 1;
+    v = (w&1UL)|(v<<1); }
 #ifdef LONG_IS_64BIT
   for (i=1; i<=32; i++)
-   { w = u&v; w ^= w >> 32; w ^= w >> 16; w ^= w >> 8; w ^= w >> 4; w ^= w >> 2; w ^= w >> 1;
-   v = (w&1UL)|(v<<1); }
+  { w = u&v; w ^= w >> 32; w ^= w >> 16; w ^= w >> 8; w ^= w >> 4; w ^= w >> 2;
+    w ^= w >> 1; v = (w&1UL)|(v<<1); }
 #endif
-  return (F2x_recip1(v)<<1)|1UL;
+  return (F2x_recip1(v)<<1) | 1UL;
 }
 
 static GEN
@@ -768,15 +772,22 @@ F2xn_inv1(GEN v, long e)
   return mkvecsmall2(v[1],F2xn_inv_basecase1(uel(v,2))&mask);
 }
 
+static GEN
+F2xn_div1(GEN g, GEN f, long e)
+{
+  GEN fi = F2xn_inv1(f, e);
+  return g ? F2xn_mul(g, fi, e): fi;
+}
+
 GEN
-F2xn_inv(GEN f, long e)
+F2xn_div(GEN g, GEN f, long e)
 {
   pari_sp av = avma, av2;
   ulong mask;
   GEN W;
   long n=1;
-  if (lg(f)==2) pari_err_INV("Flxn_inv",f);
-  if (e <= BITS_IN_LONG) return F2xn_inv1(f, e);
+  if (lg(f)<=2) pari_err_INV("Flxn_inv",f);
+  if (e <= BITS_IN_LONG) return F2xn_div1(g, f, e);
   W = F2xn_inv1(f, BITS_IN_LONG);
   mask = quadratic_prec_mask(divsBIL(e+BITS_IN_LONG-1));
   n = BITS_IN_LONG;
@@ -788,8 +799,16 @@ F2xn_inv(GEN f, long e)
     n<<=1; if (mask & 1) n--;
     mask >>= 1;
     fr = F2xn_red(f, n);
-    u = F2x_shift(F2xn_mul(W, fr, n), -n2);
-    W = F2x_add(W, F2x_shift(F2xn_mul(u, W, n-n2), n2));
+    if (mask>1 || !g)
+    {
+      u = F2x_shift(F2xn_mul(W, fr, n), -n2);
+      W = F2x_add(W, F2x_shift(F2xn_mul(u, W, n-n2), n2));
+    } else
+    {
+      GEN y = F2xn_mul(g, W, n), yt =  F2xn_red(y, n-n2);
+      u = F2xn_mul(yt, F2x_shift(F2xn_mul(fr,  W, n), -n2), n-n2);
+      W = F2x_add(y, F2x_shift(u, n2));
+    }
     if (gc_needed(av2,2))
     {
       if(DEBUGMEM>1) pari_warn(warnmem,"F2xn_inv, e = %ld", n);
@@ -799,6 +818,10 @@ F2xn_inv(GEN f, long e)
   return gerepileupto(av, F2xn_red(W,e));
 }
 
+GEN
+F2xn_inv(GEN f, long e)
+{ return F2xn_div(NULL, f, e); }
+
 GEN
 F2x_get_red(GEN T)
 {
@@ -1400,8 +1423,7 @@ F2xq_log_find_rel(GEN b, long r, GEN T, GEN *g, ulong *e)
         GEN G = F2x_factorel(gcoeff(M,1,1));
         GEN rel = mkmat2(vecsmall_concat(gel(F, 1),gel(G, 1)),
                          vecsmall_concat(gel(F, 2),zv_neg(gel(G, 2))));
-        gerepileall(av, 2, g, &rel);
-        return rel;
+        return gc_all(av, 2, &rel, g);
       }
     }
     if (gc_needed(av,2))
@@ -2491,8 +2513,7 @@ F2xqX_divrem(GEN x, GEN S, GEN T, GEN *pr)
     GEN q = F2xqX_divrem_Barrett(x,mg,y,T,pr);
     if (!q) return gc_NULL(av);
     if (!pr || pr==ONLY_DIVIDES) return gerepilecopy(av, q);
-    gerepileall(av,2,&q,pr);
-    return q;
+    return gc_all(av,2,&q,pr);
   }
 }
 
@@ -2729,16 +2750,15 @@ F2xqX_extgcd_halfgcd(GEN x, GEN y, GEN T,  GEN *ptu, GEN *ptv)
 GEN
 F2xqX_extgcd(GEN x, GEN y, GEN T,  GEN *ptu, GEN *ptv)
 {
+  pari_sp av = avma;
   GEN d;
-  pari_sp ltop=avma;
   x = F2xqX_red(x, T);
   y = F2xqX_red(y, T);
   if (lg(y)>F2xqX_EXTGCD_LIMIT)
     d = F2xqX_extgcd_halfgcd(x, y, T, ptu, ptv);
   else
     d = F2xqX_extgcd_basecase(x, y, T, ptu, ptv);
-  gerepileall(ltop,ptu?3:2,&d,ptv,ptu);
-  return d;
+  return gc_all(av, ptu?3:2, &d, ptv, ptu);
 }
 
 static GEN
