@@ -1193,8 +1193,8 @@ Flxq_ell_to_a4a6(GEN E, GEN T, ulong p)
   c4 = Rg_to_Flxq(ell_get_c4(E),T,p);
   c6 = Rg_to_Flxq(ell_get_c6(E),T,p);
   retmkvec3(Flx_neg(Flx_mulu(c4, 27, p), p), Flx_neg(Flx_mulu(c6, 54, p), p),
-            mkvec4(Fl_to_Flx(6%p,T[1]),Flx_mulu(b2,3,p),
-                   Flx_mulu(a1,3,p),Flx_mulu(a3,108,p)));
+            mkvec4(Fl_to_Flx(6%p,T[1]), Flx_triple(b2,p), Flx_triple(a1,p),
+                   Flx_mulu(a3,108,p)));
 }
 
 static GEN
@@ -2035,6 +2035,7 @@ ffgen(GEN T, long v)
   switch(typ(T))
   {
     case t_FFELT:
+      if (v < 0 || FF_var(T)==v) return FF_gen(T);
       p = FF_p_i(T); T = FF_mod(T); d = degpol(T);
       break;
     case t_POL:
@@ -2174,6 +2175,7 @@ GEN
 fforder(GEN x, GEN o)
 {
   if (typ(x)!=t_FFELT) pari_err_TYPE("fforder",x);
+  if (FF_equal0(x)) pari_err_DOMAIN("fforder", "x", "=", gen_0, x);
   return FF_order(x,o);
 }
 
