@@ -15,6 +15,8 @@ Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA. */
 #include "pari.h"
 #include "paripriv.h"
 
+#define DEBUGLEVEL DEBUGLEVEL_ellcard
+
 /* Not so fast arithmetic with points over elliptic curves over Fp */
 
 /***********************************************************************/
@@ -153,9 +155,9 @@ _FpJ_mul(void *E, GEN P, GEN n)
   pari_sp av = avma;
   struct _FpE *e=(struct _FpE *) E;
   long s = signe(n);
-  if (!s || ell_is_inf(P)) return ellinf();
+  if (!s || signe(gel(P,3))==0) return mkvec3(gen_1, gen_1, gen_0);
   if (s<0) P = FpJ_neg(P, e->p);
-  if (is_pm1(n)) return s>0? gcopy(P): P;
+  if (is_pm1(n)) return s > 0 ? gcopy(P): P;
   return gerepilecopy(av, gen_pow_i(P, n, e, &_FpJ_dbl, &_FpJ_add));
 }
 
@@ -590,20 +592,35 @@ FpE_Miller(GEN Q, GEN P, GEN m, GEN a4, GEN p)
 GEN
 FpE_weilpairing(GEN P, GEN Q, GEN m, GEN a4, GEN p)
 {
-  pari_sp ltop = avma;
+  pari_sp av = avma;
   GEN N, D, w;
   if (ell_is_inf(P) || ell_is_inf(Q) || ZV_equal(P,Q)) return gen_1;
+  if (lgefint(p)==3 && lgefint(m)==3)
+  {
+    ulong pp = p[2];
+    GEN Pp = ZV_to_Flv(P, pp), Qp = ZV_to_Flv(Q, pp);
+    ulong w = Fle_weilpairing(Pp, Qp, itou(m), umodiu(a4, pp), pp);
+    set_avma(av); return utoi(w);
+  }
   N = FpE_Miller(P, Q, m, a4, p);
   D = FpE_Miller(Q, P, m, a4, p);
   w = Fp_div(N, D, p);
   if (mpodd(m)) w  = Fp_neg(w, p);
-  return gerepileuptoint(ltop, w);
+  return gerepileuptoint(av, w);
 }
 
 GEN
 FpE_tatepairing(GEN P, GEN Q, GEN m, GEN a4, GEN p)
 {
   if (ell_is_inf(P) || ell_is_inf(Q)) return gen_1;
+  if (lgefint(p)==3 && lgefint(m)==3)
+  {
+    pari_sp av = avma;
+    ulong pp = p[2];
+    GEN Pp = ZV_to_Flv(P, pp), Qp = ZV_to_Flv(Q, pp);
+    ulong w = Fle_tatepairing(Pp, Qp, itou(m), umodiu(a4, pp), pp);
+    set_avma(av); return utoi(w);
+  }
   return FpE_Miller(P, Q, m, a4, p);
 }
 
@@ -1192,8 +1209,7 @@ Fl_ellcard_Shanks(ulong c4, ulong c6, ulong p)
         if (table[m].x < uel(ftest,1)) l=m+1; else r=m;
       }
       if (r < s && table[r].x == uel(ftest,1)) break;
-      if (Fle_add_inplace(ftest, fg, cp4, p))
-        pari_err_PRIME("ellap",utoi(p));
+      if (Fle_add_inplace(ftest, fg, cp4, p)) pari_err_PRIME("ellap",utoi(p));
     }
     h += table[r].i * B;
     if (table[r].y == uel(ftest,2))
