@@ -15,6 +15,8 @@ Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA. */
 #include "pari.h"
 #include "paripriv.h"
 
+#define DEBUGLEVEL DEBUGLEVEL_mat
+
 /********************************************************************/
 /**                                                                **/
 /**                           REDUCTION                            **/
@@ -34,19 +36,6 @@ GEN
 FpC_center(GEN x, GEN p, GEN pov2)
 { pari_APPLY_type(t_COL, Fp_center(gel(x,i), p, pov2)) }
 
-/* assume 0 <= u < p and ps2 = p>>1 */
-INLINE void
-Fp_center_inplace(GEN u, GEN p, GEN ps2)
-{ if (abscmpii(u,ps2) > 0) subiiz(u,p,u); }
-
-void
-FpC_center_inplace(GEN z, GEN p, GEN ps2)
-{
-  long i,l = lg(z);
-  for (i=1; i<l; i++)
-    Fp_center_inplace(gel(z,i), p, ps2);
-}
-
 GEN
 Flv_center(GEN x, ulong p, ulong ps2)
 { pari_APPLY_ulong(Fl_center(uel(x,i),p,ps2)) }
@@ -60,11 +49,41 @@ GEN
 FpM_center(GEN x, GEN p, GEN pov2)
 { pari_APPLY_same(FpC_center(gel(x,i), p,pov2)) }
 
+/* p != 3; assume entries in [0,p[ and ps2 = p>>1. */
+static void
+_FpC_center_inplace(GEN z, GEN p, GEN ps2)
+{
+  long i, l = lg(z);
+  for (i = 1; i < l; i++)
+  { /* HACK: assume p != 3, which ensures u = gen_[0-2] is never written to */
+    GEN u = gel(z,i);
+    if (abscmpii(u, ps2) > 0) subiiz(u, p, u);
+  }
+}
+static void
+_F3C_center_inplace(GEN z)
+{
+  long i, l = lg(z);
+  for (i = 1; i < l; i++) /* z[i] = 0, 1 : no-op */
+    if (equaliu(gel(z,i), 2)) gel(z,i) = gen_m1;
+}
+void
+FpC_center_inplace(GEN z, GEN p, GEN ps2)
+{
+  if (equaliu(p, 3))
+    _F3C_center_inplace(z);
+  else
+    _FpC_center_inplace(z, p, ps2);
+}
+
 void
 FpM_center_inplace(GEN z, GEN p, GEN pov2)
 {
   long i, l = lg(z);
-  for (i=1; i<l; i++) FpC_center_inplace(gel(z,i), p, pov2);
+  if (equaliu(p, 3))
+    for (i = 1; i < l; i++) _F3C_center_inplace(gel(z,i));
+  else
+    for (i = 1; i < l; i++) _FpC_center_inplace(gel(z,i), p, pov2);
 }
 GEN
 Flm_center(GEN x, ulong p, ulong ps2)
@@ -892,15 +911,18 @@ Flv_dotproduct_pre(GEN x, GEN y, ulong p, ulong pi)
 }
 
 ulong
-Flx_dotproduct(GEN x, GEN y, ulong p)
+Flx_dotproduct_pre(GEN x, GEN y, ulong p, ulong pi)
 {
   long lx = minss(lgpol(x), lgpol(y));
   if (lx == 0) return 0;
-  if (SMALL_ULONG(p))
-    return Flv_dotproductspec_SMALL(x+2, y+2, p, lx);
-  else
+  if (pi)
     return Flv_dotproductspec_i(x+2, y+2, p, get_Fl_red(p), lx);
+  else
+    return Flv_dotproductspec_SMALL(x+2, y+2, p, lx);
 }
+ulong
+Flx_dotproduct(GEN x, GEN y, ulong p)
+{ return Flx_dotproduct_pre(x, y, p, SMALL_ULONG(p)? 0: get_Fl_red(p)); }
 
 GEN
 FpM_FpC_mul(GEN x, GEN y, GEN p)
@@ -922,6 +944,7 @@ Flm_Flc_mul(GEN x, GEN y, ulong p)
     return Flm_Flc_mul_i(x, y, lx, l, p, get_Fl_red(p));
 }
 
+/* allow pi = 0 */
 GEN
 Flm_Flc_mul_pre(GEN x, GEN y, ulong p, ulong pi)
 {
@@ -937,6 +960,7 @@ Flm_Flc_mul_pre(GEN x, GEN y, ulong p, ulong pi)
   return z;
 }
 
+/* allow pi = 0 */
 GEN
 Flm_Flc_mul_pre_Flx(GEN x, GEN y, ulong p, ulong pi, long sv)
 {
@@ -1321,7 +1345,7 @@ zCs_to_ZC(GEN R, long nbrow)
 
 GEN
 zMs_to_ZM(GEN x, long nbrow)
-{ pari_APPLY_same(zCs_to_ZC(gel(x, i), nbrow)) }
+{ pari_APPLY_type(t_MAT, zCs_to_ZC(gel(x, i), nbrow)) }
 
 /* Solve equation f(X) = B (mod p) where B is a FpV, and f is an endomorphism.
  * Return either a solution as a t_COL, or a kernel vector as a t_VEC. */
