@@ -40,18 +40,20 @@ RgX_to_ser_i(GEN x, long l, long v, int copy)
     if (w)
     { /* keep type information, e.g. Mod(0,3) + x */
       GEN z = gel(x,2); /* = 0 */
-      x += w; gel(y,2) = gadd(gel(x,2), z); i++;
+      if (lx <= 2) gel(y,2) = copy? gcopy(z): z;
+      else { x += w; gel(y,2) = gadd(gel(x,2), z); }
+      i++;
     }
   }
   y[1] = evalvarn(vx) | evalvalp(v); /* must come here because of LONG_MAX */
   if (lx > l) lx = l;
-  /* 3 <= lx <= l */
+  /* 2 <= lx <= l */
   if (copy)
     for (; i <lx; i++) gel(y,i) = gcopy(gel(x,i));
   else
     for (; i <lx; i++) gel(y,i) = gel(x,i);
   for (     ; i < l; i++) gel(y,i) = gen_0;
-  return normalize(y);
+  return normalizeser(y);
 }
 /* enlarge/truncate t_POL x to a t_SER with lg l */
 GEN
@@ -61,7 +63,7 @@ RgX_to_ser_inexact(GEN x, long l)
 {
   long i, lx = lg(x);
   int first = 1;
-  for (i = 2; i < lx && gequal0(gel(x,i)); i++) /* ~ RgX_valrem + normalize */
+  for (i = 2; i < lx && gequal0(gel(x,i)); i++) /* ~ RgX_valrem + normalizeser */
     if (first && !isexactzero(gel(x,i)))
     {
       pari_warn(warner,"normalizing a series with 0 leading term");
@@ -69,17 +71,51 @@ RgX_to_ser_inexact(GEN x, long l)
     }
   return RgX_to_ser_i(x, l, i - 2, 0);
 }
-GEN
-rfrac_to_ser(GEN x, long l)
+/* *pd t_POL normalized wrt exact zeros; normalize fully, keeping type
+ * information */
+static long
+RgX_valrem_type(GEN *pd, long *warn)
 {
-  GEN d = gel(x,2);
-  if (l == 2)
-  {
-    long v = varn(d);
-    return zeroser(varn(d), gvaluation(x, pol_x(v)));
-  }
-  return gdiv(gel(x,1), RgX_to_ser(d, l));
+  GEN d = *pd, z = gel(d,2);
+  long v;
+  if (!gequal0(z)) return 0;
+  *warn = 1;
+  if (!signe(d)) { *pd = scalarpol_shallow(z, varn(d)); return degpol(d); }
+  v = RgX_valrem_inexact(d, &d);
+  if (lg(d) > 2)
+    gel(d,2) = gadd(gel(d,2), z);
+  else
+    d = scalarpol_shallow(z, varn(d));
+  *pd = d; return v;
 }
+static GEN
+_rfrac_to_ser(GEN x, long l, long copy)
+{
+  GEN a = gel(x,1), d = gel(x,2);
+  long warn = 0, v = varn(d), e;
+  if (l == 2) return zeroser(v, gvaluation(x, pol_x(v)));
+  e = - RgX_valrem(d, &d);
+  e -= RgX_valrem_type(&d, &warn);
+  if (!signe(d)) pari_err_INV("rfrac_to_ser", gel(x,2));
+  if (typ(a) != t_POL || varn(a) != v)
+  {
+    a = RgX_Rg_mul(RgXn_inv(d, l - 2), a);
+    e += RgX_valrem_type(&a, &warn);
+  }
+  else
+  {
+    e += RgX_valrem(a, &a);
+    e += RgX_valrem_type(&a, &warn);
+    a = RgXn_div(a, d, l - 2);
+  }
+  if (warn) pari_warn(warner,"normalizing a series with 0 leading term");
+  a = RgX_to_ser_i(a, l, 0, copy);
+  setvalp(a, valp(a) + e); return a;
+}
+GEN
+rfrac_to_ser(GEN x, long l) { return _rfrac_to_ser(x, l, 1); }
+GEN
+rfrac_to_ser_i(GEN x, long l) { return _rfrac_to_ser(x, l, 0); }
 
 static GEN
 RgV_to_ser_i(GEN x, long v, long l, int copy)
@@ -87,14 +123,14 @@ RgV_to_ser_i(GEN x, long v, long l, int copy)
   long j, lx = minss(lg(x), l-1);
   GEN y;
   if (lx == 1) return zeroser(v, l-2);
-  y = cgetg(l, t_SER); y[1] = evalvarn(v)|evalvalp(0);
+  y = cgetg(l, t_SER); y[1] = evalsigne(1)|evalvarn(v)|evalvalp(0);
   x--;
   if (copy)
     for (j = 2; j <= lx; j++) gel(y,j) = gcopy(gel(x,j));
   else
     for (j = 2; j <= lx; j++) gel(y,j) = gel(x,j);
   for (     ; j < l;   j++) gel(y,j) = gen_0;
-  return normalize(y);
+  return normalizeser(y);
 }
 GEN
 RgV_to_ser(GEN x, long v, long l) { return RgV_to_ser_i(x, v, l, 0); }
@@ -134,7 +170,7 @@ rfracrecip_to_ser_absolute(GEN R, long N)
   GEN n = gel(R,1), d = gel(R,2);
   long v = rfracrecip(&n, &d); /* R(1/x) = x^v * n/d, val(n) = val(d) = 0 */
   if (N <= v) return zeroser(varn(d), N);
-  R = gdiv(n, RgX_to_ser(d, N-v+2));
+  R = rfrac_to_ser_i(mkrfrac(n, d), N-v+2);
   setvalp(R, v); return R;
 }
 
@@ -142,19 +178,18 @@ rfracrecip_to_ser_absolute(GEN R, long N)
 GEN
 scalarser(GEN x, long v, long prec)
 {
-  long i, l;
+  long i, l, s;
   GEN y;
 
-  if (gequal0(x))
+  if (isexactzero(x))
   {
     if (isrationalzero(x)) return zeroser(v, prec);
-    if (!isexactzero(x)) prec--;
     y = cgetg(3, t_SER);
     y[1] = evalsigne(0) | _evalvalp(prec) | evalvarn(v);
     gel(y,2) = gcopy(x); return y;
   }
-  l = prec + 2; y = cgetg(l, t_SER);
-  y[1] = evalsigne(1) | _evalvalp(0) | evalvarn(v);
+  l = prec + 2; y = cgetg(l, t_SER); s = !gequal0(x);
+  y[1] = evalsigne(s) | _evalvalp(0) | evalvarn(v);
   gel(y,2) = gcopy(x); for (i=3; i<l; i++) gel(y,i) = gen_0;
   return y;
 }
@@ -196,8 +231,8 @@ toser_i(GEN x)
   switch(typ(x))
   {
     case t_SER: return x;
-    case t_POL: return RgX_to_ser(x, precdl+2);
-    case t_RFRAC: return rfrac_to_ser(x, precdl+2);
+    case t_POL: return RgX_to_ser_inexact(x, precdl+2);
+    case t_RFRAC: return rfrac_to_ser_i(x, precdl+2);
   }
   return NULL;
 }
@@ -222,8 +257,7 @@ gtoser(GEN x, long v, long prec)
   {
     case t_POL: return poltoser(x, v, prec);
     case t_RFRAC: return rfractoser(x, v, prec);
-    case t_QFR:
-    case t_QFI: return RgV_to_ser_i(x, v, 4+1, 1);
+    case t_QFB: return RgV_to_ser_i(x, v, 4+1, 1);
     case t_VECSMALL: x = zv_to_ZV(x);/*fall through*/
     case t_VEC: case t_COL:
       if (varncmp(gvar(x), v) <= 0) pari_err_PRIORITY("Ser", x, "<=", v);
@@ -243,8 +277,7 @@ gtoser_prec(GEN x, long v, long prec)
   {
     case t_SER: if (varn(x) != v) break;
                 return gerepilecopy(av, sertoser(x, prec));
-    case t_QFR:
-    case t_QFI:
+    case t_QFB:
       x = RgV_to_ser_i(mkvec3(gel(x,1),gel(x,2),gel(x,3)), v, prec+2, 1);
       return gerepileupto(av, x);
     case t_VECSMALL: x = zv_to_ZV(x);/*fall through*/
