@@ -45,35 +45,24 @@ static long
 ZV_max_lg_i(GEN x, long m)
 {
   long i, prec = 2;
-  for (i=1; i<m; i++) { long l = lgefint(gel(x,i)); if (l > prec) prec = l; }
+  for (i = 1; i < m; i++) prec = maxss(prec, lgefint(gel(x,i)));
   return prec;
 }
-
 long
-ZV_max_lg(GEN x)
-{ return ZV_max_lg_i(x, lg(x)); }
+ZV_max_lg(GEN x) { return ZV_max_lg_i(x, lg(x)); }
 
 static long
 ZM_max_lg_i(GEN x, long n, long m)
 {
-  long prec = 2;
-  if (n != 1)
-  {
-    long j;
-    for (j=1; j<n; j++)
-    {
-      long l = ZV_max_lg_i(gel(x,j), m);
-      if (l > prec) prec = l;
-    }
-  }
+  long j, prec = 2;
+  for (j = 1; j < n; j++) prec = maxss(prec, ZV_max_lg_i(gel(x,j), m));
   return prec;
 }
-
 long
 ZM_max_lg(GEN x)
 {
   long n = lg(x);
-  if (n==1) return 2;
+  if (n == 1) return 2;
   return ZM_max_lg_i(x, n, lgcols(x));
 }
 
@@ -198,10 +187,8 @@ ZM_nm_mul(GEN x, GEN y)
 
 /* Strassen-Winograd algorithm */
 
-/*
-  Return A[ma+1..ma+da, na+1..na+ea] - B[mb+1..mb+db, nb+1..nb+eb]
-  as an (m x n)-matrix, padding the input with zeroes as necessary.
-*/
+/* Return A[ma+1..ma+da, na+1..na+ea] - B[mb+1..mb+db, nb+1..nb+eb]
+ * as an (m x n)-matrix, padding the input with zeroes as necessary. */
 static GEN
 add_slices(long m, long n,
            GEN A, long ma, long da, long na, long ea,
@@ -215,36 +202,26 @@ add_slices(long m, long n,
     for (i = 1; i <= min_d; i++)
       gel(C, i) = addii(gcoeff(A, ma + i, na + j),
                         gcoeff(B, mb + i, nb + j));
-    for (; i <= da; i++)
-      gel(C, i) = gcoeff(A, ma + i, na + j);
-    for (; i <= db; i++)
-      gel(C, i) = gcoeff(B, mb + i, nb + j);
-    for (; i <= m; i++)
-      gel(C, i) = gen_0;
+    for (; i <= da; i++) gel(C, i) = gcoeff(A, ma + i, na + j);
+    for (; i <= db; i++) gel(C, i) = gcoeff(B, mb + i, nb + j);
+    for (; i <= m; i++)  gel(C, i) = gen_0;
   }
   for (; j <= ea; j++) {
     gel(M, j) = C = cgetg(m + 1, t_COL);
-    for (i = 1; i <= da; i++)
-      gel(C, i) = gcoeff(A, ma + i, na + j);
-    for (; i <= m; i++)
-      gel(C, i) = gen_0;
+    for (i = 1; i <= da; i++) gel(C, i) = gcoeff(A, ma + i, na + j);
+    for (; i <= m; i++) gel(C, i) = gen_0;
   }
   for (; j <= eb; j++) {
     gel(M, j) = C = cgetg(m + 1, t_COL);
-    for (i = 1; i <= db; i++)
-      gel(C, i) = gcoeff(B, mb + i, nb + j);
-    for (; i <= m; i++)
-      gel(C, i) = gen_0;
+    for (i = 1; i <= db; i++) gel(C, i) = gcoeff(B, mb + i, nb + j);
+    for (; i <= m; i++) gel(C, i) = gen_0;
   }
-  for (; j <= n; j++)
-    gel(M, j) = zerocol(m);
+  for (; j <= n; j++) gel(M, j) = zerocol(m);
   return M;
 }
 
-/*
-  Return A[ma+1..ma+da, na+1..na+ea] - B[mb+1..mb+db, nb+1..nb+eb]
-  as an (m x n)-matrix, padding the input with zeroes as necessary.
-*/
+/* Return A[ma+1..ma+da, na+1..na+ea] - B[mb+1..mb+db, nb+1..nb+eb]
+ * as an (m x n)-matrix, padding the input with zeroes as necessary. */
 static GEN
 subtract_slices(long m, long n,
                 GEN A, long ma, long da, long na, long ea,
@@ -258,29 +235,21 @@ subtract_slices(long m, long n,
     for (i = 1; i <= min_d; i++)
       gel(C, i) = subii(gcoeff(A, ma + i, na + j),
                         gcoeff(B, mb + i, nb + j));
-    for (; i <= da; i++)
-      gel(C, i) = gcoeff(A, ma + i, na + j);
-    for (; i <= db; i++)
-      gel(C, i) = negi(gcoeff(B, mb + i, nb + j));
-    for (; i <= m; i++)
-      gel(C, i) = gen_0;
+    for (; i <= da; i++) gel(C, i) = gcoeff(A, ma + i, na + j);
+    for (; i <= db; i++) gel(C, i) = negi(gcoeff(B, mb + i, nb + j));
+    for (; i <= m; i++) gel(C, i) = gen_0;
   }
   for (; j <= ea; j++) {
     gel(M, j) = C = cgetg(m + 1, t_COL);
-    for (i = 1; i <= da; i++)
-      gel(C, i) = gcoeff(A, ma + i, na + j);
-    for (; i <= m; i++)
-      gel(C, i) = gen_0;
+    for (i = 1; i <= da; i++) gel(C, i) = gcoeff(A, ma + i, na + j);
+    for (; i <= m; i++) gel(C, i) = gen_0;
   }
   for (; j <= eb; j++) {
     gel(M, j) = C = cgetg(m + 1, t_COL);
-    for (i = 1; i <= db; i++)
-      gel(C, i) = negi(gcoeff(B, mb + i, nb + j));
-    for (; i <= m; i++)
-      gel(C, i) = gen_0;
+    for (i = 1; i <= db; i++) gel(C, i) = negi(gcoeff(B, mb + i, nb + j));
+    for (; i <= m; i++) gel(C, i) = gen_0;
   }
-  for (; j <= n; j++)
-    gel(M, j) = zerocol(m);
+  for (; j <= n; j++) gel(M, j) = zerocol(m);
   return M;
 }
 
@@ -422,9 +391,7 @@ ZM_mul_slice(GEN A, GEN B, GEN P, GEN *mod)
   for(i=1; i <= n; i++)
     gel(H,i) = Flm_mul(gel(A,i),gel(B,i),P[i]);
   H = nmV_chinese_center_tree_seq(H, P, T, ZV_chinesetree(P, T));
-  *mod = gmael(T, lg(T)-1, 1);
-  gerepileall(av, 2, &H, mod);
-  return H;
+  *mod = gmael(T, lg(T)-1, 1); return gc_all(av, 2, &H, mod);
 }
 
 GEN
@@ -435,40 +402,15 @@ ZM_mul_worker(GEN P, GEN A, GEN B)
   return V;
 }
 
-static long
-ZC_expi(GEN x)
-{
-  long i, l = lg(x), m=-1;
-  for(i = 1; i < l; i++)
-  {
-    GEN xi = gel(x,i);
-    long e = signe(xi) ? expi(xi): -1;
-    if (e > m) m = e;
-  }
-  return m;
-}
-
-static long
-ZM_expi(GEN x)
-{
-  long i, l = lg(x), m=-1;
-  for(i = 1; i < l; i++)
-  {
-    long e = ZC_expi(gel(x,i));
-    if (e > m) m = e;
-  }
-  return m;
-}
-
 static GEN
-ZM_mul_fast(GEN A, GEN B, long lA, long lB)
+ZM_mul_fast(GEN A, GEN B, long lA, long lB, long sA, long sB)
 {
   pari_sp av = avma;
   forprime_t S;
   GEN H, worker;
-  long h, mA = ZM_expi(A), mB = ZM_expi(B);
-  if (mA==-1 || mB==-1) return zeromat(nbrows(A),lB-1);
-  h = 3 + mA + mB + expu(lA-1);
+  long h;
+  if (sA == 2 || sB == 2) return zeromat(nbrows(A),lB-1);
+  h = 1 + (sA + sB - 4) * BITS_IN_LONG + expu(lA-1);
   init_modular_big(&S);
   worker = snm_closure(is_entry("_ZM_mul_worker"), mkvec2(A,B));
   H = gen_crt("ZM_mul", worker, &S, NULL, h, 0, NULL,
@@ -476,29 +418,35 @@ ZM_mul_fast(GEN A, GEN B, long lA, long lB)
   return gerepileupto(av, H);
 }
 
-/* Strassen-Winograd used for dim >= ZM_sw_bound */
+/* s = min(log_BIL |x|, log_BIL |y|), use Strassen-Winograd when
+ * min(dims) > B */
+static long
+sw_bound(long s)
+{ return s > 60 ? 2: s > 25 ? 4: s > 15 ? 8 : s > 8 ? 16 : 32; }
+
 static GEN
 ZM_mul_i(GEN x, GEN y, long l, long lx, long ly)
 {
-  if (lx > 70 && ly > 70 && l > 70)
-    return ZM_mul_fast(x, y, lx, ly);
+  long sx, sy, B;
+  if (lx==3 && l==3 && ly==3) return ZM2_mul(x,y);
+  sx = ZM_max_lg_i(x, lx, l);
+  sy = ZM_max_lg_i(y, ly, lx);
+  if ((lx > 70 && ly > 70 && l > 70) && (sx <= 10 * sy && sy <= 10 * sx))
+    return ZM_mul_fast(x, y, lx, ly, sx, sy);
+
+  B = sw_bound(minss(sx, sy));
+  if (l <= B || lx <= B || ly <= B)
+    return ZM_mul_classical(x, y, l, lx, ly);
   else
-  {
-    long s = maxss(ZM_max_lg_i(x,lx,l), ZM_max_lg_i(y,ly,lx));
-    long ZM_sw_bound = s > 60 ? 2: s > 25 ? 4: s>15 ? 8 : s > 8 ? 16 : 32;
-    if (l <= ZM_sw_bound || lx <= ZM_sw_bound || ly <= ZM_sw_bound)
-      return ZM_mul_classical(x, y, l, lx, ly);
-    else
-      return ZM_mul_sw(x, y, l - 1, lx - 1, ly - 1);
-  }
+    return ZM_mul_sw(x, y, l - 1, lx - 1, ly - 1);
 }
 
 GEN
 ZM_mul(GEN x, GEN y)
 {
-  long lx=lg(x), ly=lg(y);
-  if (ly==1) return cgetg(1,t_MAT);
-  if (lx==1) return zeromat(0, ly-1);
+  long lx = lg(x), ly = lg(y);
+  if (ly == 1) return cgetg(1,t_MAT);
+  if (lx == 1) return zeromat(0, ly-1);
   return ZM_mul_i(x, y, lgcols(x), lx, ly);
 }
 
@@ -651,17 +599,12 @@ ZM_transmul(GEN x, GEN y)
 static GEN
 ZM_sqr_i(GEN x, long l)
 {
-  if (l > 70)
-    return ZM_mul_fast(x, x, l, l);
+  long s = ZM_max_lg_i(x, l, l);
+  if (l > 70) return ZM_mul_fast(x, x, l, l, s, s);
+  if (l <= sw_bound(s))
+    return ZM_mul_classical(x, x, l, l, l);
   else
-  {
-    long s = ZM_max_lg_i(x,l,l);
-    long ZM_sw_bound = s > 60 ? 2: s > 25 ? 4: s>15 ? 8 : s > 8 ? 16 : 32;
-    if (l <= ZM_sw_bound)
-      return ZM_mul_classical(x, x, l, l, l);
-    else
-      return ZM_mul_sw(x, x, l - 1, l - 1, l - 1);
-  }
+    return ZM_mul_sw(x, x, l - 1, l - 1, l - 1);
 }
 
 GEN
@@ -890,13 +833,17 @@ GEN
 ZC_Z_divexact(GEN x, GEN c)
 { pari_APPLY_type(t_COL, diviiexact(gel(x,i), c)) }
 GEN
-ZC_u_divexact(GEN x, ulong c)
+ZC_divexactu(GEN x, ulong c)
 { pari_APPLY_type(t_COL, diviuexact(gel(x,i), c)) }
 
 GEN
 ZM_Z_divexact(GEN x, GEN c)
 { pari_APPLY_same(ZC_Z_divexact(gel(x,i), c)) }
 
+GEN
+ZM_divexactu(GEN x, ulong c)
+{ pari_APPLY_same(ZC_divexactu(gel(x,i), c)) }
+
 GEN
 ZC_Z_mul(GEN x, GEN c)
 {
@@ -1090,6 +1037,7 @@ ZV_neg_inplace(GEN M)
 GEN
 ZC_neg(GEN x)
 { pari_APPLY_type(t_COL, negi(gel(x,i))) }
+
 GEN
 zv_neg(GEN x)
 { pari_APPLY_long(-x[i]) }
@@ -1101,6 +1049,9 @@ zv_neg_inplace(GEN M)
   return M;
 }
 GEN
+zv_abs(GEN x)
+{ pari_APPLY_ulong(labs(x[i])) }
+GEN
 ZM_neg(GEN x)
 { pari_APPLY_same(ZC_neg(gel(x,i))) }
 
