@@ -445,6 +445,49 @@ ZX_translate(GEN P, GEN c)
   }
 }
 
+/* P(ax + b) */
+GEN
+ZX_affine(GEN P, GEN a, GEN b)
+{
+  if (signe(b)) P = ZX_translate(P, b);
+  return ZX_unscale(P, a);
+}
+
+GEN
+ZX_Z_eval(GEN x, GEN y)
+{
+  long i = lg(x)-1, j;
+  pari_sp av = avma;
+  GEN t, r;
+
+  if (i<=2) return (i==2)? icopy(gel(x,2)): gen_0;
+  if (!signe(y)) return icopy(gel(x,2));
+
+  t = gel(x,i); i--;
+#if 0 /* standard Horner's rule */
+  for ( ; i>=2; i--)
+    t = addii(mulii(t,y),gel(x,i));
+#endif
+  /* specific attention to sparse polynomials */
+  for ( ; i>=2; i = j-1)
+  {
+    for (j = i; !signe(gel(x,j)); j--)
+      if (j==2)
+      {
+        if (i != j) y = powiu(y, i-j+1);
+        return gerepileuptoint(av, mulii(t,y));
+      }
+    r = (i==j)? y: powiu(y, i-j+1);
+    t = addii(mulii(t,r), gel(x,j));
+    if (gc_needed(av,2))
+    {
+      if (DEBUGMEM>1) pari_warn(warnmem,"ZX_Z_eval: i = %ld",i);
+      t = gerepileuptoint(av, t);
+    }
+  }
+  return gerepileuptoint(av, t);
+}
+
 /* Return 2^(n degpol(P))  P(x >> n) */
 GEN
 ZX_rescale2n(GEN P, long n)
@@ -677,17 +720,39 @@ GEN
 ZX_mulspec(GEN x, GEN y, long nx, long ny)
 {
   pari_sp av;
-  long ex, ey, vx, vy;
-  GEN z;
+  long ex, ey, vx, vy, v;
   if (!nx || !ny) return pol_0(0);
   vx = ZX_valspec(x,nx); nx-=vx; x += vx;
   vy = ZX_valspec(y,ny); ny-=vy; y += vy;
-  if (nx==1) return Z_ZX_mulshiftspec(gel(x,0), y, ny, vx+vy);
-  if (ny==1) return Z_ZX_mulshiftspec(gel(y,0), x, nx, vy+vx);
+  v = vx + vy;
+  if (nx==1) return Z_ZX_mulshiftspec(gel(x,0), y, ny, v);
+  if (ny==1) return Z_ZX_mulshiftspec(gel(y,0), x, nx, v);
+  if (nx == 2 && ny == 2)
+  {
+    GEN a0 = gel(x,0), a1 = gel(x,1), A0, A1, A2;
+    GEN b0 = gel(y,0), b1 = gel(y,1), z = cgetg(5 + v, t_POL);
+    long i;
+    z[1] = evalvarn(0) | evalsigne(1);
+    A0 = mulii(a0, b0);
+    A2 = mulii(a1, b1); av = avma;
+    A1 = gerepileuptoint(av, subii(addii(A0,A2),
+                                   mulii(subii(a1,a0), subii(b1,b0))));
+    i = 4 + v;
+    gel(z,i--) = A2;
+    gel(z,i--) = A1;
+    gel(z,i--) = A0; while (i > 1) gel(z, i--) = gen_0;
+    return z;
+  }
+#if 0
+  /* generically slower even when degrees differ a lot; sometimes about twice
+   * faster when bitsize is moderate */
+  if (DEBUGVAR)
+    return RgX_mulspec(x - vx, y - vy, nx + vx, ny + vy);
+#endif
   av = avma;
-  ex = ZX_expispec(x, nx); ey = ZX_expispec(y, ny);
-  z  = ZX_mulspec_mulii(x,y,nx,ny,ex,ey,vx+vy);
-  return gerepileupto(av, z);
+  ex = ZX_expispec(x, nx);
+  ey = ZX_expispec(y, ny);
+  return gerepileupto(av,  ZX_mulspec_mulii(x,y,nx,ny,ex,ey,v));
 }
 GEN
 ZX_mul(GEN x, GEN y)
@@ -788,6 +853,34 @@ ZX_mod_Xnm1(GEN T, ulong n)
   return normalizepol_lg(S, l);
 }
 
+static GEN
+_ZX_mul(void* E, GEN x, GEN y)
+{ (void) E; return ZX_mul(x, y); }
+static GEN
+_ZX_sqr(void *E, GEN x)
+{ (void) E; return ZX_sqr(x); }
+
+static GEN
+_ZX_divrem(void * E, GEN x, GEN y, GEN *r)
+{ (void) E; return RgX_divrem(x, y, r); }
+
+static GEN
+_ZX_add(void * E, GEN x, GEN y)
+{ (void) E; return ZX_add(x, y); }
+
+static struct bb_ring ZX_ring = { _ZX_add,_ZX_mul,_ZX_sqr };
+
+GEN
+ZX_digits(GEN x, GEN T)
+{
+  long d = degpol(T), n = (lgpol(x)+d-1)/d;
+  return gen_digits(x, T, n, NULL, &ZX_ring, _ZX_divrem);
+}
+
+GEN
+ZXV_ZX_fromdigits(GEN x, GEN T)
+{ return gen_fromdigits(x,T, NULL, &ZX_ring); }
+
 /*******************************************************************/
 /*                                                                 */
 /*                                ZXV                              */
@@ -828,6 +921,15 @@ ZXV_dotproduct(GEN x, GEN y)
   return gerepileupto(av, c);
 }
 
+GEN
+ZXC_to_FlxC(GEN x, ulong p, long sv)
+{ pari_APPLY_type(t_COL,typ(gel(x,i))==t_INT ?
+                  Z_to_Flx(gel(x,i), p, sv): ZX_to_Flx(gel(x,i), p)) }
+
+GEN
+ZXM_to_FlxM(GEN x, ulong p, long sv)
+{ pari_APPLY_same(ZXC_to_FlxC(gel(x,i), p, sv)) }
+
 /*******************************************************************/
 /*                                                                 */
 /*                                ZXQM                             */
@@ -945,9 +1047,9 @@ ZXQM_mul(GEN x, GEN y, GEN T)
     z = ZM_mul(simplify_shallow(x),simplify_shallow(y));
   else
   {
-    long ex = ZXM_expi(x), ey = ZXM_expi(y), d= degpol(T), n = lg(x)-1;
-    long e = ex + ey + expu(d) + expu(n) + 4;
-    long N = divsBIL(e)+1;
+    long e, N, ex = ZXM_expi(x), ey = ZXM_expi(y), n = lg(x)-1;
+    e = ex + ey + expu(d) + expu(n) + 4;
+    N = divsBIL(e)+1;
     z = ZM_mul(ZXM_eval2BIL(x,N), ZXM_eval2BIL(y,N));
     z = ZM_mod2BIL_ZXQM(z, N, T);
   }
@@ -1137,7 +1239,7 @@ ZXX_Z_divexact(GEN y, GEN x)
  * P(X,Y) = sum_{0<=i<lP} P_i(X) * Y^i, where deg P_i < n.
  * Returns P(X,X^(2n-1)) */
 GEN
-ZXX_to_Kronecker_spec(GEN P, long lP, long n)
+RgXX_to_Kronecker_spec(GEN P, long lP, long n)
 {
   long i, k, N = (n<<1) + 1;
   GEN y;
@@ -1156,7 +1258,7 @@ ZXX_to_Kronecker_spec(GEN P, long lP, long n)
     {
       long l = lg(c);
       if (l-3 >= n)
-        pari_err_BUG("ZXX_to_Kronecker, P is not reduced mod Q");
+        pari_err_BUG("RgXX_to_Kronecker, P is not reduced mod Q");
       for (j=2; j < l; j++) gel(y,k++) = gel(c,j);
     }
     if (i == lP-1) break;
@@ -1189,18 +1291,18 @@ Kronecker_to_ZXX(GEN z, long n, long v)
 }
 
 GEN
-ZXX_to_Kronecker(GEN P, long n)
+RgXX_to_Kronecker(GEN P, long n)
 {
-  GEN z = ZXX_to_Kronecker_spec(P+2, lgpol(P), n);
+  GEN z = RgXX_to_Kronecker_spec(P+2, lgpol(P), n);
   setvarn(z,varn(P)); return z;
 }
 GEN
 ZXX_mul_Kronecker(GEN x, GEN y, long n)
-{ return ZX_mul(ZXX_to_Kronecker(x,n), ZXX_to_Kronecker(y,n)); }
+{ return ZX_mul(RgXX_to_Kronecker(x,n), RgXX_to_Kronecker(y,n)); }
 
 GEN
 ZXX_sqr_Kronecker(GEN x, long n)
-{ return ZX_sqr(ZXX_to_Kronecker(x,n)); }
+{ return ZX_sqr(RgXX_to_Kronecker(x,n)); }
 
 /* shallow, n = deg(T) */
 GEN
