@@ -317,30 +317,13 @@ ok_elt(GEN x, GEN xZ, GEN y)
   return gc_bool(av, ZM_equal(x, ZM_hnfmodid(y, xZ)));
 }
 
+/* a + s * b, a and b ZM, s integer */
 static GEN
-addmul_col(GEN a, long s, GEN b)
+addmul_mat(GEN a, GEN s, GEN b)
 {
-  long i,l;
-  if (!s) return a? leafcopy(a): a;
-  if (!a) return gmulsg(s,b);
-  l = lg(a);
-  for (i=1; i<l; i++)
-    if (signe(gel(b,i))) gel(a,i) = addii(gel(a,i), mulsi(s, gel(b,i)));
-  return a;
-}
-
-/* a <-- a + s * b, all coeffs integers */
-static GEN
-addmul_mat(GEN a, long s, GEN b)
-{
-  long j,l;
-  /* copy otherwise next call corrupts a */
-  if (!s) return a? RgM_shallowcopy(a): a;
-  if (!a) return gmulsg(s,b);
-  l = lg(a);
-  for (j=1; j<l; j++)
-    (void)addmul_col(gel(a,j), s, gel(b,j));
-  return a;
+  if (!signe(s)) return a;
+  if (!equali1(s)) b = ZM_Z_mul(b, s);
+  return a? ZM_add(a, b): b;
 }
 
 static GEN
@@ -348,9 +331,9 @@ get_random_a(GEN nf, GEN x, GEN xZ)
 {
   pari_sp av;
   long i, lm, l = lg(x);
-  GEN a, z, beta, mul;
+  GEN z, beta, mul;
 
-  beta= cgetg(l, t_VEC);
+  beta= cgetg(l, t_MAT);
   mul = cgetg(l, t_VEC); lm = 1; /* = lg(mul) */
   /* look for a in x such that a O/xZ = x O/xZ */
   for (i = 2; i < l; i++)
@@ -364,22 +347,19 @@ get_random_a(GEN nf, GEN x, GEN xZ)
     gel(mul,lm) = t; lm++;
   }
   setlg(mul, lm);
-  setlg(beta,lm);
-  z = cgetg(lm, t_VECSMALL);
+  setlg(beta,lm); z = cgetg(lm, t_VEC);
   for(av = avma;; set_avma(av))
   {
-    for (a=NULL,i=1; i<lm; i++)
+    GEN a = NULL;
+    for (i = 1; i < lm; i++)
     {
-      long t = random_bits(4) - 7; /* in [-7,8] */
-      z[i] = t;
-      a = addmul_mat(a, t, gel(mul,i));
+      gel(z,i) = randomi(xZ);
+      a = addmul_mat(a, gel(z,i), gel(mul,i));
     }
     /* a = matrix (NOT HNF) of ideal generated by beta.z in O/xZ */
     if (a && ok_elt(x,xZ, a)) break;
   }
-  for (a=NULL,i=1; i<lm; i++)
-    a = addmul_col(a, z[i], gel(beta,i));
-  return a;
+  return ZM_ZC_mul(beta, z);
 }
 
 /* x square matrix, assume it is HNF */
@@ -2579,7 +2559,7 @@ GEN
 idealred0(GEN nf, GEN I, GEN vdir)
 {
   pari_sp av = avma;
-  GEN G, aI, IZ, J, y, yZ, my, yi, c1 = NULL;
+  GEN G, aI, IZ, J, y, my, yi, c1 = NULL;
   long N;
 
   nf = checknf(nf);
@@ -2599,6 +2579,7 @@ idealred0(GEN nf, GEN I, GEN vdir)
       I = idealhnf_two(nf,I);
       break;
     case id_MAT:
+      if (lg(I)-1 != N) pari_err_DIM("idealred");
       I = Q_primitive_part(I, &c1);
       IZ = gcoeff(I,1,1);
       if (is_pm1(IZ))
@@ -2626,13 +2607,12 @@ idealred0(GEN nf, GEN I, GEN vdir)
   c1 = mul_content(c1, IZ);
   if (equali1(c1)) c1 = NULL; /* can be simplified with IZ */
   yi = ZM_gauss(my, col_ei(N,1)); /* y^-1 */
-  yZ = Q_denom(yi); /* (y) \cap Z */
-  I = hnfmodid(I, yZ);
+  I = hnfmodid(I, Q_denom(yi)); /* denom(yi) generates (y) \cap Z */
   if (!aI) return gerepileupto(av, I);
   if (typ(aI) == t_MAT) /* yi is not integral and usually larger than y */
     aI = famat_div(aI, y);
-  else if (c1)
-    c1 = RgC_Rg_mul(yi, c1);
+  else
+    c1 = c1? RgC_Rg_mul(yi, c1): yi;
 END:
   if (c1) aI = ext_mul(nf, aI,c1);
   return gerepilecopy(av, mkvec2(I, aI));
