@@ -338,7 +338,7 @@ pollegendre_eval0(long n, GEN x, long flag)
   { /* u = P_i(x), v = P_{i-1}(x), compute t = P_{i+1}(x) */
     GEN t;
     if ((i & 0xff) == 0) gerepileall(av,2,&u, &v);
-    t = gdivgs(gsub(gmul(gmulsg(2*i+1,x), u), gmulsg(i,v)), i+1);
+    t = gdivgu(gsub(gmul(gmulsg(2*i+1,x), u), gmulsg(i,v)), i+1);
     v = u; u = t;
   }
   if (flag) return gerepilecopy(av, mkvec2(v, u));
@@ -366,7 +366,7 @@ pollaguerre(long n, GEN a, long v)
     if (i)
     {
       c2 = divis(c2,-i);
-      c1 = gdivgs(gmul(c1, gaddsg(i,a)), n+1-i);
+      c1 = gdivgu(gmul(c1, gaddsg(i,a)), n+1-i);
     }
   }
   return gerepilecopy(av, L);
@@ -409,7 +409,7 @@ pollaguerre_eval0(long n, GEN a, GEN x, long flag)
   { /* u = P_i(x), v = P_{i-1}(x), compute t = P_{i+1}(x) */
     GEN t;
     if ((i & 0xff) == 0) gerepileall(av,2,&u, &v);
-    t = gdivgs(gsub(gmul(gsub(gaddsg(2*i+1,a),x), u), gmul(gaddsg(i,a),v)), i+1);
+    t = gdivgu(gsub(gmul(gsub(gaddsg(2*i+1,a),x), u), gmul(gaddsg(i,a),v)), i+1);
     v = u; u = t;
   }
   if (flag) return gerepilecopy(av, mkvec2(v, u));
@@ -493,7 +493,7 @@ polcyclo_eval(long n, GEN x)
   } else {
     if (gequal1(x))
     { /* n is prime, return n; multiply by x to keep the type */
-      if (l == 1) return gerepileupto(av, gmulgs(x,n));
+      if (l == 1) return gerepileupto(av, gmulgu(x,n));
       return gerepilecopy(av, x); /* else 1 */
     }
     if (gequalm1(x)) return gerepileupto(av, gneg(x)); /* -1 */
@@ -553,7 +553,7 @@ polcyclo_eval(long n, GEN x)
     if (!(bitmask_q & (bitmask_q-1))) /* detects power of 2, since bitmask!=0 */
     {
       i = vals(bitmask_q)+1; /* q = P[i] */
-      ypos = gmulgs(ypos, P[i]);
+      ypos = gmulgu(ypos, P[i]);
     }
   }
   return gerepileupto(av, ypos);
@@ -686,9 +686,7 @@ gprec_w(GEN x, long pr)
       for (i=2; i<lx; i++) gel(y,i) = gprec_w(gel(x,i),pr);
       break;
     case t_POLMOD: case t_RFRAC: case t_VEC: case t_COL: case t_MAT:
-      y = cgetg_copy(x, &lx);
-      for (i=1; i<lx; i++) gel(y,i) = gprec_w(gel(x,i),pr);
-      break;
+      pari_APPLY_same(gprec_w(gel(x,i), pr));
     default: return x;
   }
   return y;
@@ -717,9 +715,7 @@ gprec_wensure(GEN x, long pr)
       for (i=2; i<lx; i++) gel(y,i) = gprec_wensure(gel(x,i),pr);
       break;
     case t_POLMOD: case t_RFRAC: case t_VEC: case t_COL: case t_MAT:
-      y = cgetg_copy(x, &lx);
-      for (i=1; i<lx; i++) gel(y,i) = gprec_wensure(gel(x,i),pr);
-      break;
+      pari_APPLY_same(gprec_wensure(gel(x,i), pr));
     default: return x;
   }
   return y;
@@ -748,9 +744,7 @@ gprec_wtrunc(GEN x, long pr)
       for (i=2; i<lx; i++) gel(y,i) = gprec_wtrunc(gel(x,i),pr);
       break;
     case t_POLMOD: case t_RFRAC: case t_VEC: case t_COL: case t_MAT:
-      y = cgetg_copy(x, &lx);
-      for (i=1; i<lx; i++) gel(y,i) = gprec_wtrunc(gel(x,i),pr);
-      break;
+      pari_APPLY_same(gprec_wtrunc(gel(x,i), pr));
     default: return x;
   }
   return y;
@@ -819,7 +813,10 @@ convol(GEN x, GEN y)
   ex = valp(x);
   ey = valp(y);
   if (ser_isexactzero(x))
-    return scalarser(gadd(Rg_get_0(x), Rg_get_0(y)), varn(x), maxss(ex,ey));
+  {
+    z = scalarser(gadd(Rg_get_0(x), Rg_get_0(y)), varn(x), 1);
+    setvalp(z, maxss(ex,ey)); return z;
+  }
   lx = lg(x) + ex; x -= ex;
   ly = lg(y) + ey; y -= ey;
   /* inputs shifted: x[i] and y[i] now correspond to monomials of same degree */
@@ -830,7 +827,7 @@ convol(GEN x, GEN y)
   z = cgetg(lx - ex, t_SER);
   z[1] = evalvalp(ex) | evalvarn(vx);
   for (j = ex+2; j<lx; j++) gel(z,j-ex) = gmul(gel(x,j),gel(y,j));
-  return normalize(z);
+  return normalizeser(z);
 }
 
 /***********************************************************************/
@@ -944,17 +941,72 @@ dirdiv(GEN x, GEN y)
 /*******************************************************************/
 /**                      BINOMIAL COEFFICIENTS                    **/
 /*******************************************************************/
+/* Lucas's formula for v_p(\binom{n}{k}), used in the tough case p <= sqrt(n) */
+static long
+binomial_lval(ulong n, ulong k, ulong p)
+{
+  ulong r = 0, e = 0;
+  do
+  {
+    ulong a = n % p, b = k % p + r;
+    n /= p; k /= p;
+    if (a < b) { e++; r = 1; } else r = 0;
+  } while (n);
+  return e;
+}
 GEN
 binomialuu(ulong n, ulong k)
 {
-  pari_sp ltop = avma;
-  GEN z;
+  pari_sp av = avma;
+  ulong p, nk, sn;
+  long c, l;
+  forprime_t T;
+  GEN v, z;
   if (k > n) return gen_0;
-  k = minuu(k,n-k);
+  nk = n-k; if (k > nk) lswap(nk, k);
   if (!k) return gen_1;
   if (k == 1) return utoipos(n);
-  z = diviiexact(mulu_interval(n-k+1, n), mulu_interval(2UL, k));
-  return gerepileuptoint(ltop,z);
+  if (k == 2) return muluu(odd(n)? n: n-1, n>>1);
+  if (k < 1000 || ((double)k/ n) * log((double)n) < 0.5)
+  { /* k "small" */
+    z = diviiexact(mulu_interval(n-k+1, n), mulu_interval(2UL, k));
+    return gerepileuptoint(av, z);
+  }
+  sn = usqrt(n);
+  /* use Lucas's formula, k <= n/2 */
+  l = minuu(1UL << 20, n); v = cgetg(l+1, t_VECSMALL); c = 1;
+  u_forprime_init(&T, nk+1, n);
+  while ((p = u_forprime_next(&T))) /* all primes n-k < p <= n occur, v_p = 1 */
+  {
+    if (c == l) { ulong L = l << 1; v = vecsmall_lengthen(v, L); l = L; }
+    v[c++] = p;
+  }
+  u_forprime_init(&T, sn+1, n >> 1);
+  while ((p = u_forprime_next(&T))) /* p^2 > n, v_p <= 1 */
+    if (n % p < k % p)
+    {
+      if (c == l) { ulong L = l << 1; v = vecsmall_lengthen(v, L); l = L; }
+      v[c++] = p;
+    }
+  setlg(v, c); z = zv_prod_Z(v);
+  u_forprime_init(&T, 3, sn);
+  l = minuu(1UL << 20, sn); v = cgetg(l + 1, t_VEC); c = 1;
+  while ((p = u_forprime_next(&T))) /* p <= sqrt(n) */
+  {
+    ulong e = binomial_lval(n, k, p);
+    if (e)
+    {
+      if (c == l) { ulong L = l << 1; v = vec_lengthen(v, L); l = L; }
+      gel(v, c++) = powuu(p, e);
+    }
+  }
+  setlg(v, c); z = mulii(z, ZV_prod(v));
+  { /* p = 2 */
+    ulong e = hammingl(k);
+    e += (k == nk)? e: hammingl(nk);
+    e -= hammingl(n); if (e) z = shifti(z, e);
+  }
+  return gerepileuptoint(av, z);
 }
 
 GEN
@@ -976,31 +1028,24 @@ binomial(GEN n, long k)
   {
     if (signe(n) > 0)
     {
-      GEN z = subiu(n,k);
-      if (cmpis(z,k) < 0)
+      GEN z = subiu(n, k);
+      if (cmpiu(z, k) < 0)
       {
-        k = itos(z); set_avma(av);
-        if (k <= 1)
+        switch(signe(z))
         {
-          if (k < 0) return gen_0;
-          if (k == 0) return gen_1;
-          return icopy(n);
+          case -1: return gc_const(av, gen_0);
+          case 0: return gc_const(av, gen_1);
         }
+        k = z[2];
+        if (k == 1) { set_avma(av); return icopy(n); }
       }
+      set_avma(av);
+      if (lgefint(n) == 3) return binomialuu(n[2],(ulong)k);
     }
     /* k > 1 */
-    if (lgefint(n) == 3 && signe(n) > 0)
-    {
-      y = binomialuu(itou(n),(ulong)k);
-      return gerepileupto(av, y);
-    }
-    else
-    {
-      y = cgetg(k+1,t_VEC);
-      for (i=1; i<=k; i++) gel(y,i) = subiu(n,i-1);
-      y = ZV_prod(y);
-    }
-    y = diviiexact(y, mpfact(k));
+    y = cgetg(k+1,t_VEC); gel(y,1) = n;
+    for (i = 2; i <= k; i++) gel(y,i) = subiu(n,i-1);
+    y = diviiexact(ZV_prod(y), mpfact(k));
     return gerepileuptoint(av, y);
   }
 
@@ -1150,6 +1195,44 @@ polrecip(GEN x)
 /**                  POLYNOMIAL INTERPOLATION                      **/
 /**                                                                **/
 /********************************************************************/
+/* given complex roots L[i], i <= n of some monic T in C[X], return
+ * the T'(L[i]), computed stably via products of differences */
+GEN
+vandermondeinverseinit(GEN L)
+{
+  long i, j, l = lg(L);
+  GEN V = cgetg(l, t_VEC);
+  for (i = 1; i < l; i++)
+  {
+    pari_sp av = avma;
+    GEN W = cgetg(l-1,t_VEC);
+    long k = 1;
+    for (j = 1; j < l; j++)
+      if (i != j) gel(W, k++) = gsub(gel(L,i), gel(L,j));
+    gel(V,i) = gerepileupto(av, RgV_prod(W));
+  }
+  return V;
+}
+
+/* Compute the inverse of the van der Monde matrix of T multiplied by den */
+GEN
+vandermondeinverse(GEN L, GEN T, GEN den, GEN V)
+{
+  pari_sp av = avma;
+  long i, n = lg(L)-1;
+  GEN M = cgetg(n+1, t_MAT);
+
+  if (!V) V = vandermondeinverseinit(L);
+  if (den && equali1(den)) den = NULL;
+  for (i = 1; i <= n; i++)
+  {
+    GEN d = gel(V,i), P = RgX_Rg_mul(RgX_div_by_X_x(T, gel(L,i), NULL),
+                                     den? gdiv(den,d): ginv(d));
+    gel(M,i) = RgX_to_RgC(P, n);
+  }
+  return gerepilecopy(av, M);
+}
+
 static GEN
 RgV_polint_fast(GEN X, GEN Y, long v)
 {
@@ -1167,18 +1250,18 @@ GEN
 RgV_polint(GEN X, GEN Y, long v)
 {
   pari_sp av0 = avma, av;
-  GEN Q, P = NULL;
+  GEN Q, L, P = NULL;
   long i, l = lg(Y);
   if ((Q = RgV_polint_fast(X,Y,v))) return Q;
   if (!X) X = identity_ZV(l-1);
+  L = vandermondeinverseinit(X);
   Q = roots_to_pol(X, v); av = avma;
   for (i=1; i<l; i++)
   {
-    GEN inv, T, dP;
+    GEN T, dP;
     if (gequal0(gel(Y,i))) continue;
     T = RgX_div_by_X_x(Q, gel(X,i), NULL);
-    inv = ginv(poleval(T,gel(X,i)));
-    dP = RgX_Rg_mul(T, gmul(gel(Y,i),inv));
+    dP = RgX_Rg_mul(T, gdiv(gel(Y,i), gel(L,i)));
     P = P? RgX_add(P, dP): dP;
     if (gc_needed(av,2))
     {
@@ -1762,8 +1845,8 @@ vecsearch(GEN v, GEN x, GEN k)
   if (tv == t_VECSMALL)
     x = (GEN)itos(x);
   else if (!is_matvec_t(tv)) pari_err_TYPE("vecsearch", v);
-  r = CMP? gen_search(v, x, 0, E, CMP): key_search(v, x, k);
-  return gc_long(av, r);
+  r = CMP? gen_search(v, x, E, CMP): key_search(v, x, k);
+  return gc_long(av, r < 0? 0: r);
 }
 
 GEN
@@ -1774,6 +1857,18 @@ GEN
 ZV_sort_uniq(GEN L) { return gen_sort_uniq(L, (void*)&cmpii, &cmp_nodata); }
 void
 ZV_sort_inplace(GEN L) { gen_sort_inplace(L, (void*)&cmpii, &cmp_nodata,NULL); }
+GEN
+ZV_sort_uniq_shallow(GEN L)
+{
+  GEN v = gen_indexsort_uniq(L, (void*)&cmpii, &cmp_nodata);
+  return vecpermute(L, v);
+}
+GEN
+ZV_sort_shallow(GEN L)
+{
+  GEN v = gen_indexsort(L, (void*)&cmpii, &cmp_nodata);
+  return vecpermute(L, v);
+}
 
 GEN
 vec_equiv(GEN F)
@@ -1833,20 +1928,19 @@ tablesearch(GEN T, GEN x, int (*cmp)(GEN,GEN))
 
 /* looks if x belongs to the set T and returns the index if yes, 0 if no */
 long
-gen_search(GEN T, GEN x, long flag, void *data, int (*cmp)(void*,GEN,GEN))
+gen_search(GEN T, GEN x, void *data, int (*cmp)(void*,GEN,GEN))
 {
   long u = lg(T)-1, i, l, s;
 
-  if (!u) return flag? 1: 0;
+  if (!u) return -1;
   l = 1;
   do
   {
-    i = (l+u)>>1; s = cmp(data, x, gel(T,i));
-    if (!s) return flag? 0: i;
-    if (s<0) u=i-1; else l=i+1;
-  } while (u>=l);
-  if (!flag) return 0;
-  return (s<0)? i: i+1;
+    i = (l+u) >> 1; s = cmp(data, x, gel(T,i));
+    if (!s) return i;
+    if (s < 0) u = i-1; else l = i+1;
+  } while (u >= l);
+  return -((s < 0)? i: i+1);
 }
 
 long
@@ -2093,7 +2187,7 @@ setisset(GEN x)
 long
 setsearch(GEN T, GEN y, long flag)
 {
-  long lx;
+  long i, lx;
   switch(typ(T))
   {
     case t_VEC: lx = lg(T); break;
@@ -2104,7 +2198,9 @@ setsearch(GEN T, GEN y, long flag)
       return 0; /*LCOV_EXCL_LINE*/
   }
   if (lx==1) return flag? 1: 0;
-  return gen_search(T,y,flag,(void*)cmp_universal,cmp_nodata);
+  i = gen_search(T,y,(void*)cmp_universal,cmp_nodata);
+  if (i > 0) return flag? 0: i;
+  return flag ? -i: 0;
 }
 
 GEN
@@ -2120,6 +2216,26 @@ setunion(GEN x, GEN y)
   return gerepilecopy(av, setunion_i(x, y));
 }
 
+GEN
+setdelta(GEN x, GEN y)
+{
+  long ix = 1, iy = 1, iz = 1, lx = lg(x), ly = lg(y);
+  pari_sp av = avma;
+  GEN z = cgetg(lx + ly - 1,t_VEC);
+  if (typ(x) != t_VEC) pari_err_TYPE("setdelta",x);
+  if (typ(y) != t_VEC) pari_err_TYPE("setdelta",y);
+  while (ix < lx && iy < ly)
+  {
+    int c = cmp_universal(gel(x,ix), gel(y,iy));
+    if      (c < 0) gel(z, iz++) = gel(x,ix++);
+    else if (c > 0) gel(z, iz++) = gel(y,iy++);
+    else { ix++; iy++; }
+  }
+  while (ix<lx) gel(z,iz++) = gel(x,ix++);
+  while (iy<ly) gel(z,iz++) = gel(y,iy++);
+  setlg(z,iz); return gerepilecopy(av,z);
+}
+
 GEN
 setintersect(GEN x, GEN y)
 {
