@@ -20,6 +20,8 @@ Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA. */
 #include "pari.h"
 #include "paripriv.h"
 
+#define DEBUGLEVEL DEBUGLEVEL_bnr
+
 static GEN
 bnr_get_El(GEN bnr) { return gel(bnr,3); }
 static GEN
@@ -406,13 +408,15 @@ ZM_content_mul(GEN u, GEN c, GEN *pd)
   return u;
 }
 
-/* bnr natural generators: bnf gens made coprime to modulus + bid gens */
+/* bnr natural generators: bnf gens made coprime to modulus + bid gens.
+ * Beware: if bnr includes MOD, we may have #El < #bnf.ge*/
 static GEN
 get_Gen(GEN bnf, GEN bid, GEN El)
 {
-  GEN Gen = shallowconcat(bnf_get_gen(bnf), bid_get_gen(bid));
-  GEN nf = bnf_get_nf(bnf);
+  GEN nf = bnf_get_nf(bnf), gen = bnf_get_gen(bnf), Gen;
   long i, l = lg(El);
+  if (lg(gen) > l) gen = vec_shorten(gen, l-1);
+  Gen = shallowconcat(gen, bid_get_gen(bid));
   for (i = 1; i < l; i++)
   {
     GEN e = gel(El,i);
@@ -444,15 +448,16 @@ Buchraymod_i(GEN bnf, GEN module, long flag, GEN MOD)
   cycbid = bid_get_cyc(bid);
   if (MOD)
   {
-    cyc = ZV_snf_gcd(cyc, MOD);
+    cyc = ZV_snfclean(ZV_snf_gcd(cyc, MOD));
     cycbid = ZV_snf_gcd(cycbid, MOD);
   }
   Ri = lg(cycbid)-1;
   if (Ri || add_gen || do_init)
   {
     GEN fx = bid_get_fact(bid);
-    El = cgetg(ngen+1,t_VEC);
-    for (j=1; j<=ngen; j++)
+    long n = Ri? ngen: lg(cyc)-1;
+    El = cgetg(n+1, t_VEC);
+    for (j = 1; j <= n; j++)
     {
       GEN c = idealcoprimefact(nf, gel(gen,j), fx);
       gel(El,j) = nf_to_scalar_or_basis(nf,c);
@@ -460,8 +465,9 @@ Buchraymod_i(GEN bnf, GEN module, long flag, GEN MOD)
   }
   if (!Ri)
   {
-    GEN hK = bnf_get_no(bnf);
-    clg = add_gen? mkvec3(hK, cyc, get_Gen(bnf, bid, El)): mkvec2(hK, cyc);
+    GEN no, Gen = add_gen? get_Gen(bnf, bid, El): NULL;
+    if (MOD) { ngen = lg(cyc)-1; no = ZV_prod(cyc); } else no = bnf_get_no(bnf);
+    clg = add_gen? mkvec3(no, cyc, Gen): mkvec2(no, cyc);
     if (!do_init) return clg;
     U = matid(ngen);
     U = mkvec3(U, cgetg(1,t_MAT), U);
@@ -547,7 +553,7 @@ bnrclassno(GEN bnf,GEN ideal)
   bnf = checkbnf(bnf);
   h = bnf_get_no(bnf);
   bid = checkbid_i(ideal);
-  if (!bid) bid = Idealstar(bnf, ideal, nf_INIT);
+  if (!bid) bid = Idealstar(bnf_get_nf(bnf), ideal, nf_INIT);
   cycbid = bid_get_cyc(bid);
   if (lg(cycbid) == 1) { set_avma(av); return icopy(h); }
   D = ideallog_units(bnf, bid); /* (Z_K/f)^* / units ~ Z^n / D */
@@ -608,7 +614,11 @@ bnrisprincipalmod(GEN bnr, GEN x, GEN MOD, long flag)
   bnf = bnr_get_bnf(bnr); nf = bnf_get_nf(bnf);
   bid = bnr_get_bid(bnr);
   trivialbid = lg(bid_get_cyc(bid)) == 1;
-  if (trivialbid) ex = isprincipal(bnf, x);
+  if (trivialbid)
+  {
+    ex = isprincipal(bnf, x);
+    setlg(ex, lg(cycray)); /* can happen with MOD */
+  }
   else
   {
     GEN v = bnfisprincipal0(bnf, x, nf_FORCE|nf_GENMAT);
@@ -620,7 +630,7 @@ bnrisprincipalmod(GEN bnr, GEN x, GEN MOD, long flag)
     if (!MOD && !(flag & nf_GEN)) MOD = gel(cycray,1);
     ex = ZM2_ZC2_mul(bnr_get_U(bnr), e, ideallogmod(nf, b, bid, MOD));
   }
-  ex = vecmodii(ex, cycray);
+  ex = ZV_ZV_mod(ex, cycray);
   if (!(flag & (nf_GEN|nf_GENMAT))) return gerepileupto(av, ex);
 
   /* compute generator */
@@ -722,8 +732,8 @@ zimmertbound(GEN D, long N, long R2)
 }
 
 /* return \gamma_n^n if known, an upper bound otherwise */
-static GEN
-hermiteconstant(long n)
+GEN
+Hermite_bound(long n, long prec)
 {
   GEN h,h1;
   pari_sp av;
@@ -731,17 +741,18 @@ hermiteconstant(long n)
   switch(n)
   {
     case 1: return gen_1;
-    case 2: return mkfrac(utoipos(4), utoipos(3));
+    case 2: retmkfrac(utoipos(4), utoipos(3));
     case 3: return gen_2;
     case 4: return utoipos(4);
     case 5: return utoipos(8);
-    case 6: return mkfrac(utoipos(64), utoipos(3));
+    case 6: retmkfrac(utoipos(64), utoipos(3));
     case 7: return utoipos(64);
     case 8: return utoipos(256);
+    case 24: return int2n(48);
   }
   av = avma;
-  h  = powru(divur(2,mppi(DEFAULTPREC)), n);
-  h1 = sqrr(ggamma(sstoQ(n+4,2),DEFAULTPREC));
+  h  = powru(divur(2,mppi(prec)), n);
+  h1 = sqrr(ggamma(uutoQ(n+4,2),prec));
   return gerepileuptoleaf(av, mulrr(h,h1));
 }
 
@@ -799,7 +810,7 @@ regulatorbound(GEN bnf)
 
   p1 = sqrr(glog(gdiv(dK,c1),DEFAULTPREC));
   p1 = divru(gmul2n(powru(divru(mulru(p1,3),N*(N*N-1)-6*R2),R),R2), N);
-  p1 = sqrtr(gdiv(p1, hermiteconstant(R)));
+  p1 = sqrtr(gdiv(p1, Hermite_bound(R, DEFAULTPREC)));
   if (DEBUGLEVEL>1) err_printf("Mahler bound for regulator: %Ps\n",p1);
   return gmax_shallow(p1, dbltor(0.2));
 }
@@ -812,7 +823,7 @@ is_unit(GEN M, long r1, GEN x)
   return gc_bool(av, is_pm1(Nx));
 }
 
-/* FIXME: should use smallvectors */
+/* True nf. FIXME: should use smallvectors */
 static double
 minimforunits(GEN nf, long BORNE, ulong w)
 {
@@ -1054,7 +1065,7 @@ lowerboundforregulator(GEN bnf, GEN units)
   if (DEBUGLEVEL>1) err_printf("M* = %Ps\n", dbltor(bound));
   M0 = compute_M0(dbltor(bound), N);
   if (DEBUGLEVEL>1) err_printf("M0 = %.28Pg\n",M0);
-  M = gmul2n(divru(gdiv(powrs(M0,RU),hermiteconstant(RU)),N),R2);
+  M = gmul2n(divru(gdiv(powrs(M0,RU),Hermite_bound(RU, DEFAULTPREC)),N),R2);
   if (cmprr(M, dbltor(0.04)) < 0) return NULL;
   M = sqrtr(M);
   if (DEBUGLEVEL>1)
@@ -1288,22 +1299,31 @@ char_denormalize(GEN cyc, GEN D, GEN chic)
 static GEN
 bnrchar_i(GEN bnr, GEN g, GEN v)
 {
-  long i, h, l = lg(g);
-  GEN CH, D, U, U2, H, cyc, cycD, dv, dchi;
-  checkbnr(bnr);
+  long i, h, l = lg(g), t = typ_NULL;
+  GEN CH, D, U, U2, H, cycD, dv, dchi, cyc = NULL;
+
+  if (checkbnr_i(bnr)) { t = typ_BNR; cyc = bnr_get_cyc(bnr); }
+  else if (checkznstar_i(bnr)) { t = typ_BIDZ; cyc = znstar_get_cyc(bnr); }
+  else if (typ(bnr) == t_VEC && RgV_is_ZV(bnr)) cyc = bnr;
+  else pari_err_TYPE("bnrchar", bnr);
   switch(typ(g))
   {
     GEN G;
     case t_VEC:
       G = cgetg(l, t_MAT);
-      for (i = 1; i < l; i++) gel(G,i) = isprincipalray(bnr, gel(g,i));
+      if (t == typ_BNR)
+      {
+        for (i = 1; i < l; i++) gel(G,i) = isprincipalray(bnr, gel(g,i));
+        cyc = bnr_get_cyc(bnr);
+      }
+      else
+        for (i = 1; i < l; i++) gel(G,i) = Zideallog(bnr, gel(g,i));
       g = G; break;
     case t_MAT:
       if (RgM_is_ZM(g)) break;
     default:
       pari_err_TYPE("bnrchar",g);
   }
-  cyc = bnr_get_cyc(bnr);
   H = ZM_hnfall_i(shallowconcat(g,diagonal_shallow(cyc)), v? &U: NULL, 1);
   dv = NULL;
   if (v)
@@ -1409,14 +1429,14 @@ bnrsurjection(GEN bnr1, GEN bnr2)
       M = shallowconcat(T, M);
     }
   }
-  /* could reduce the matrix mod cyc2 */
-  return mkvec3(ZM_mul(M, bnr_get_Ui(bnr1)), bnr_get_cyc(bnr1), cyc2);
+  M = ZM_ZV_mod(ZM_mul(M, bnr_get_Ui(bnr1)), cyc2);
+  return mkvec3(M, bnr_get_cyc(bnr1), cyc2);
 }
 
 /* nchi a normalized character, S a surjective map ; return S(nchi)
  * still normalized wrt the original cyclic structure (S[2]) */
 static GEN
-ag_nchar_image(GEN S, GEN nchi)
+abmap_nchar_image(GEN S, GEN nchi)
 {
   GEN U, M = gel(S,1), Mc = diagonal_shallow(gel(S,3));
   long l = lg(M);
@@ -1426,10 +1446,10 @@ ag_nchar_image(GEN S, GEN nchi)
   return char_simplify(gel(nchi,1), ZV_ZM_mul(gel(nchi,2), U));
 }
 static GEN
-ag_char_image(GEN S, GEN chi)
+abmap_char_image(GEN S, GEN chi)
 {
   GEN nchi = char_normalize(chi, cyc_normalize(gel(S,2)));
-  GEN DC = ag_nchar_image(S, nchi);
+  GEN DC = abmap_nchar_image(S, nchi);
   return char_denormalize(gel(S,3), gel(DC,1), gel(DC,2));
 }
 
@@ -1457,15 +1477,15 @@ bnrmap(GEN A, GEN B)
       B = ZM_hnfmodid(B, C); break;
     case t_MAT: /* subgroup */
       if (!RgM_is_ZM(B)) pari_err_TYPE("bnrmap [not a subgroup]", B);
-      B = ZM_hnfmodid(B, c); B = ag_subgroup_image(A, B); break;
+      B = ZM_hnfmodid(B, c); B = abmap_subgroup_image(A, B); break;
     case t_VEC: /* character */
       if (!char_check(c, B))
         pari_err_TYPE("bnrmap [not a character mod mA]", B);
-      B = ag_char_image(A, B); break;
+      B = abmap_char_image(A, B); break;
     case t_COL: /* discrete log mod mA */
       if (lg(B) != lg(c) || !RgV_is_ZV(B))
         pari_err_TYPE("bnrmap [not a discrete log]", B);
-      B = vecmodii(ZM_ZC_mul(M, B), C);
+      B = ZV_ZV_mod(ZM_ZC_mul(M, B), C);
       return gerepileupto(av, B);
   }
   return gerepilecopy(av, B);
@@ -1477,7 +1497,7 @@ bnrmap(GEN A, GEN B)
  *   chic(genc[i]) = zeta_C^chic[i]), C = cyc_normalize(bnr.cyc)[1] */
 GEN
 bnrchar_primitive(GEN bnr, GEN nchi, GEN bnrc)
-{ return ag_nchar_image(bnrsurjection(bnr, bnrc), nchi); }
+{ return abmap_nchar_image(bnrsurjection(bnr, bnrc), nchi); }
 
 /* s: <gen> = Cl_f -> Cl_f2 -> 0, H subgroup of Cl_f (generators given as
  * HNF on [gen]). Return subgroup s(H) in Cl_f2 */
@@ -1485,14 +1505,11 @@ static GEN
 imageofgroup(GEN bnr, GEN bnr2, GEN H)
 {
   if (!H) return diagonal_shallow(bnr_get_cyc(bnr2));
-  return ag_subgroup_image(bnrsurjection(bnr, bnr2), H);
+  return abmap_subgroup_image(bnrsurjection(bnr, bnr2), H);
 }
 GEN
 bnrchar_primitive_raw(GEN bnr, GEN bnrc, GEN chi)
-{
-  GEN S = bnrsurjection(bnr, bnrc);
-  return ag_char_image(S, chi);
-}
+{ return abmap_char_image(bnrsurjection(bnr, bnrc), chi); }
 
 /* convert A,B,C to [bnr, H] */
 GEN
@@ -1517,7 +1534,7 @@ bnrconductor0(GEN A, GEN B, GEN C, long flag)
 {
   pari_sp av = avma;
   GEN H, bnr = ABC_to_bnr(A,B,C,&H, 0);
-  return gerepilecopy(av, bnrconductor_i(bnr, H, flag));
+  return gerepilecopy(av, bnrconductor(bnr, H, flag));
 }
 
 long
@@ -1529,7 +1546,7 @@ bnrisconductor0(GEN A,GEN B,GEN C)
 
 static GEN
 ideallog_to_bnr_i(GEN Ubid, GEN cyc, GEN z)
-{ return (lg(Ubid)==1)? zerocol(lg(cyc)-1): vecmodii(ZM_ZC_mul(Ubid,z), cyc); }
+{ return (lg(Ubid)==1)? zerocol(lg(cyc)-1): ZV_ZV_mod(ZM_ZC_mul(Ubid,z), cyc); }
 /* return bnrisprincipal(bnr, (x)), assuming z = ideallog(x); allow a
  * t_MAT for z, understood as a collection of ideallog(x_i) */
 static GEN
@@ -1838,56 +1855,6 @@ nf_deg1_prime(GEN nf)
   return NULL;
 }
 
-static long
-rnfisabelian_i(GEN nf, GEN pol)
-{
-  GEN modpr, pr, T, Tnf, pp, ro, nfL, C, a, sig, eq;
-  long i, j, l, v;
-  ulong p, k, ka;
-
-  if (typ(nf) == t_POL)
-    Tnf = nf;
-  else {
-    nf = checknf(nf);
-    Tnf = nf_get_pol(nf);
-  }
-  v = varn(Tnf);
-  if (degpol(Tnf) != 1 && typ(pol) == t_POL && RgX_is_QX(pol)
-                       && rnfisabelian_i(pol_x(v), pol)) return 1;
-  pol = RgX_nffix("rnfisabelian",Tnf,pol,1);
-  eq = nf_rnfeq(nf,pol); /* init L := K[x]/(pol), nf attached to K */
-  C = gel(eq,1); setvarn(C, v); /* L = Q[t]/(C) */
-  a = gel(eq,2); setvarn(a, v); /* root of K.pol in L */
-  nfL = C;
-  ro = nfroots_if_split(&nfL, QXX_QXQ_eval(pol, a, C));
-  if (!ro) return 0;
-  l = lg(ro)-1;
-  /* small groups are abelian, as are groups of prime order */
-  if (l < 6 || uisprime(l)) return 1;
-
-  pr = nf_deg1_prime(nfL);
-  modpr = nf_to_Fq_init(nfL, &pr, &T, &pp);
-  p = itou(pp);
-  k = umodiu(gel(eq,3), p);
-  ka = (k * itou(nf_to_Fq(nfL, a, modpr))) % p;
-  sig= cgetg(l+1, t_VECSMALL);
-  /* image of c = ro[1] + k a [distinguished root of C] by the l automorphisms
-   * sig[i]: ro[1] -> ro[i] */
-  for (i = 1; i <= l; i++)
-    sig[i] = Fl_add(ka, itou(nf_to_Fq(nfL, gel(ro,i), modpr)), p);
-  ro = Q_primpart(ro);
-  for (i=2; i<=l; i++) { /* start at 2, since sig[1] = identity */
-    gel(ro,i) = ZX_to_Flx(gel(ro,i), p);
-    for (j=2; j<i; j++)
-      if (Flx_eval(gel(ro,j), sig[i], p)
-       != Flx_eval(gel(ro,i), sig[j], p)) return 0;
-  }
-  return 1;
-}
-long
-rnfisabelian(GEN nf, GEN pol)
-{ pari_sp av = avma; return gc_long(av, rnfisabelian_i(nf, pol)); }
-
 /* Given bnf and T defining an abelian relative extension, compute the
  * corresponding conductor and congruence subgroup. Return
  * [cond,bnr(cond),H] where cond=[ideal,arch] is the conductor. */
@@ -1895,8 +1862,10 @@ GEN
 rnfconductor0(GEN bnf, GEN T, long flag)
 {
   pari_sp av = avma;
-  GEN D, nf, module, bnr, H, lim, Tr, MOD;
+  GEN P, E, D, nf, module, bnr, H, lim, Tr, MOD;
+  long i, l, degT = degpol(T);
 
+  if (flag < 0 || flag > 2) pari_err_FLAG("rnfconductor");
   bnf = checkbnf(bnf); nf = bnf_get_nf(bnf);
   Tr = rnfdisc_get_T(nf, T, &lim);
   T = nfX_to_monic(nf, Tr, NULL);
@@ -1904,36 +1873,38 @@ rnfconductor0(GEN bnf, GEN T, long flag)
     D = rnfdisc_factored(nf, T, NULL);
   else
   {
-    GEN P, E, Ez;
-    long i, l, degT = degpol(T);
-    D = idealfactor_partial(nf, nfX_disc(nf, Q_primpart(Tr)), lim);
-    P = gel(D,1); l = lg(P);
-    E = gel(D,2); Ez = ZV_to_zv(E);
-    if (l > 1 && vecsmall_max(Ez) > 1)
-    { /* cheaply update tame primes */
-      for (i = 1; i < l; i++)
-      { /* v_pr(f) = 1 + \sum_{0 < i < l} g_i/g_0
-                   <= 1 + max_{i>0} g_i/(g_i-1) \sum_{0 < i < l} g_i -1
-                   <= 1 + (p/(p-1)) * v_P(e(L/K, pr)), P | pr | p */
-        GEN pr = gel(P,i), p = pr_get_p(pr), e = gen_1;
-        long q, v = z_pvalrem(degT, p, &q);
-        if (v)
-        { /* e = e_tame * e_wild, e_wild | p^v */
-          long ee, pp = itou(p);
-          long t = ugcd(umodiu(subiu(pr_norm(pr),1), q), q); /* e_tame | t */
-          /* upper bound for 1 + p/(p-1) * v * e(L/Q,p) */
-          ee = 1 + (pp * v * pr_get_e(pr) * upowuu(pp,v) * t) / (pp-1);
-          e = utoi(minss(ee, Ez[i]));
-        }
-        gel(E,i) = e;
+    D = nfX_disc(nf, Q_primpart(Tr));
+    if (gequal0(D))
+      pari_err_DOMAIN("rnfconductor","issquarefree(pol)","=",gen_0, Tr);
+    D = idealfactor_partial(nf, D, lim);
+  }
+  P = gel(D,1); l = lg(P);
+  E = gel(D,2);
+  for (i = 1; i < l; i++) /* cheaply update tame primes */
+  { /* v_pr(f) = 1 + \sum_{0 < i < l} g_i/g_0
+               <= 1 + max_{i>0} g_i/(g_i-1) \sum_{0 < i < l} g_i -1
+               <= 1 + (p/(p-1)) * v_P(e(L/K, pr)), P | pr | p */
+    GEN pr = gel(P,i), p = pr_get_p(pr), e = gen_1;
+    ulong q, e0 = itou(gel(E,i));
+    if (e0 > 1 && cmpiu(p, degT) <= 0)
+    {
+      long v, pp = itou(p);
+      if ((v = u_lvalrem(degT, pp, &q)))
+      { /* e = e_tame * e_wild, e_wild | p^v */
+        ulong t = ugcd(umodiu(subiu(pr_norm(pr),1), q), q); /* e_tame | t */
+        /* upper bound for 1 + p/(p-1) * v * e(L/Q,p) */
+        e0 = minuu(e0, 1 + (pp * v * pr_get_e(pr) * upowuu(pp,v) * t) / (pp-1));
+        e = utoipos(e0);
       }
     }
+    gel(E,i) = e;
   }
   module = mkvec2(D, identity_perm(nf_get_r1(nf)));
   MOD = flag? utoipos(degpol(T)): NULL;
   bnr = Buchraymod_i(bnf, module, nf_INIT|nf_GEN, MOD);
   H = rnfnormgroup_i(bnr,T); if (!H) return gc_const(av,gen_0);
-  return gerepilecopy(av, bnrconductormod(bnr, H, MOD));
+  return gerepilecopy(av, flag == 2? bnrconductor_factored(bnr, H)
+                                   : bnrconductormod(bnr, H, MOD));
 }
 GEN
 rnfconductor(GEN bnf, GEN T) { return rnfconductor0(bnf, T, 0); }
@@ -2304,9 +2275,6 @@ fasmall_append(GEN fa, long p, long e)
   retmkvec2(vecsmall_append(P,p), vecsmall_append(E,e));
 }
 
-static GEN
-sprk_get_cyc(GEN s) { return gel(s,1); }
-
 /* sprk = sprkinit(pr,k), b zsimp with modulus coprime to pr */
 static GEN
 zsimpjoin(GEN b, GEN sprk, GEN U_pr, long prcode, long e)
@@ -2741,9 +2709,7 @@ bnrautmatrix(GEN bnr, GEN aut)
   for (i = 1; i < l; i++)
     gel(M,i) = isprincipalray(bnr, nfgaloismatrixapply(nf, aut, gel(Gen,i)));
   M = ZM_mul(M, bnr_get_Ui(bnr));
-  l = lg(M);
-  for (i = 1; i < l; i++) gel(M,i) = vecmodii(gel(M,i), cyc);
-  return gerepilecopy(av, M);
+  return gerepilecopy(av, ZM_ZV_mod(M, cyc));
 }
 
 GEN
@@ -2821,3 +2787,51 @@ bnrisgalois(GEN bnr, GEN M, GEN H)
   }
   return gc_long(av,1);
 }
+
+static GEN
+bnrlcmcond(GEN bnr1, GEN bnr2)
+{
+  GEN I1 = bnr_get_bid(bnr1), f1 = bid_get_fact(I1), a1 = bid_get_arch(I1);
+  GEN I2 = bnr_get_bid(bnr2), f2 = bid_get_fact(I2), a2 = bid_get_arch(I2);
+  GEN f, a;
+  long i, l;
+  if (!gidentical(bnr_get_nf(bnr1), bnr_get_nf(bnr2)))
+    pari_err_TYPE("bnrcompositum[different fields]", mkvec2(bnr1,bnr2));
+  f = merge_factor(f1, f2, (void*)&cmp_prime_ideal, &cmp_nodata);
+  a = cgetg_copy(a1, &l);
+  for (i = 1; i < l; i++)
+    gel(a,i) = (signe(gel(a1,i)) || signe(gel(a2,i)))? gen_1: gen_0;
+  return mkvec2(f, a);
+}
+/* H subgroup (of bnr.clgp) in HNF; lift to BNR */
+static GEN
+bnrliftsubgroup(GEN BNR, GEN bnr, GEN H)
+{
+  GEN E = gel(bnrsurjection(BNR, bnr), 1), K = kerint(shallowconcat(E, H));
+  return ZM_hnfmodid(rowslice(K, 1, lg(E)-1), bnr_get_cyc(BNR));
+}
+static GEN
+ZM_intersect(GEN A, GEN B)
+{
+  GEN K = kerint(shallowconcat(A, B));
+  return ZM_mul(A, rowslice(K, 1, lg(A)-1));
+}
+GEN
+bnrcompositum(GEN fH1, GEN fH2)
+{
+  pari_sp av = avma;
+  GEN bnr1, bnr2, bnr, H1, H2, H, n1, n2;
+  if (typ(fH1) != t_VEC || lg(fH2) != 3) pari_err_TYPE("bnrcompositum", fH1);
+  if (typ(fH2) != t_VEC || lg(fH2) != 3) pari_err_TYPE("bnrcompositum", fH2);
+  bnr1 = gel(fH1,1); if (!checkbnr_i(bnr1)) pari_err_TYPE("bnrcompositum",bnr1);
+  bnr2 = gel(fH2,1); if (!checkbnr_i(bnr2)) pari_err_TYPE("bnrcompositum",bnr2);
+  H1 = bnr_subgroup_check(bnr1, gel(fH1,2), &n1);
+  if (!H1) H1 = diagonal_shallow(bnr_get_cyc(bnr1));
+  H2 = bnr_subgroup_check(bnr2, gel(fH2,2), &n2);
+  if (!H2) H2 = diagonal_shallow(bnr_get_cyc(bnr2));
+  bnr = bnrinitmod(bnr_get_bnf(bnr1), bnrlcmcond(bnr1, bnr2), 0, lcmii(n1,n2));
+  H1 = bnrliftsubgroup(bnr, bnr1, H1);
+  H2 = bnrliftsubgroup(bnr, bnr2, H2);
+  H = ZM_intersect(H1, H2);
+  return gerepilecopy(av, mkvec2(bnr, ZM_hnfmodid(H, bnr_get_cyc(bnr))));
+}
