@@ -85,16 +85,20 @@ char_normalize(GEN chi, GEN ncyc)
 static GEN
 get_cyc(GEN x, GEN chi, const char *s)
 {
-  if (nftyp(x) == typ_BIDZ)
+  switch(nftyp(x))
   {
-    if (!zncharcheck(x, chi)) pari_err_TYPE(s, chi);
-    return NULL;
-  }
-  else
-  {
-    if (typ(x) != t_VEC || !RgV_is_ZV(x)) x = member_cyc(x);
-    if (!char_check(x, chi)) pari_err_TYPE(s, chi);
-    return x;
+    case typ_BIDZ:
+      if (!zncharcheck(x, chi)) pari_err_TYPE(s, chi);
+      return NULL;
+    case typ_GCHAR:
+      x = gchar_get_cyc(x);
+      if (!is_vec_t(typ(chi)) || lg(chi) != lg(x) || !RgV_is_ZV(chi))
+        pari_err_TYPE(s, chi); /* FIXME: handle norm component */
+      return x;
+    default:
+      if (typ(x) != t_VEC || !RgV_is_ZV(x)) x = member_cyc(x);
+      if (!char_check(x, chi)) pari_err_TYPE(s, chi);
+      return x;
   }
 }
 
@@ -128,6 +132,8 @@ charorder(GEN cyc, GEN x)
     if (signe(gel(x,i)))
     {
       GEN c, o = gel(cyc,i);
+      if (!signe(o))
+        return gerepileupto(av,mkoo());
       c = gcdii(o, gel(x,i));
       if (!is_pm1(c)) o = diviiexact(o,c);
       f = lcmii(f, o);
@@ -303,6 +309,8 @@ chareval(GEN G, GEN chi, GEN x, GEN z)
     case typ_BIDZ:
       if (checkznstar_i(G)) return gerepileupto(av, znchareval(G, chi, x, z));
       /* don't implement chars on general bid: need an nf... */
+    case typ_GCHAR:
+        pari_err_TYPE("chareval [use gchareval to evaluate a grossencharacter]", G);
     default:
       pari_err_TYPE("chareval", G);
       return NULL;/* LCOV_EXCL_LINE */
@@ -459,7 +467,7 @@ cyc2elts_normal(GEN cyc, long maxord, GEN ORD)
   long i, n, o, N, j = 1;
   GEN z, vcoprime;
 
-  if (typ(cyc) != t_VECSMALL) cyc = gtovecsmall(cyc);
+  if (typ(cyc) != t_VECSMALL) cyc = vec_to_vecsmall(cyc);
   n = lg(cyc)-1;
   N = zv_prod(cyc);
   z = cgetg(N+1, t_VEC);
@@ -798,7 +806,7 @@ znconreylog(GEN bid, GEN x)
     case t_COL: /* log_bid(x) */
       Ui = znstar_get_Ui(bid);
       if (!RgV_is_ZV(x) || lg(x) != lg(Ui)) pari_err_TYPE("znconreylog", x);
-      return gerepileupto(av, vecmodii(ZM_ZC_mul(Ui,x), cycg));
+      return gerepileupto(av, ZV_ZV_mod(ZM_ZC_mul(Ui,x), cycg));
     case t_VEC:
       return gerepilecopy(av, znconreyfromchar(bid, x));
     default: pari_err_TYPE("znconreylog", x);
@@ -1026,8 +1034,7 @@ znconreyconductor(GEN bid, GEN chi, GEN *pm)
     E = zc_to_ZC(E);
     q = mkvec2(factorback2(P,E), mkmat2(P,E));
   }
-  gerepileall(av, pm? 2: 1, &q, pm);
-  return q;
+  return gc_all(av, pm? 2: 1, &q, pm);
 }
 
 GEN
