@@ -15,6 +15,8 @@ Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA. */
 #include "pari.h"
 #include "paripriv.h"
 
+#define DEBUGLEVEL DEBUGLEVEL_isprime
+
 #define dbg_mode()  if (DEBUGLEVEL >= 2)
 #define dbg_mode1() if (DEBUGLEVEL >= 3)
 
@@ -143,11 +145,13 @@ ecpp_param_get_maxdisc(GEN param) { return umael3(param, 1, 1, 2); }
 INLINE ulong
 ecpp_param_get_maxpcdg(GEN param) { return umael3(param, 1, 1, 3); }
 INLINE GEN
-ecpp_param_get_primelist(GEN param) { return gmael(param, 1, 2); }
+ecpp_param_get_uprimelist(GEN param) { return gmael(param, 1, 2); }
 INLINE GEN
-ecpp_param_get_disclist(GEN param) { return gmael(param, 1, 3); }
+ecpp_param_get_primelist(GEN param) { return gmael(param, 1, 3); }
 INLINE GEN
-ecpp_param_get_primeord(GEN param) { return gmael(param, 1, 4); }
+ecpp_param_get_disclist(GEN param) { return gmael(param, 1, 4); }
+INLINE GEN
+ecpp_param_get_primeord(GEN param) { return gmael(param, 1, 5); }
 INLINE GEN
 ecpp_param_get_primorial_vec(GEN param) { return gel(param, 2); }
 INLINE GEN
@@ -256,32 +260,51 @@ sort_Dmq_by_cnum(void *data, GEN x, GEN y)
   return sort_Dmq_by_q(data, x, y);
 }
 
-/* return H s.t if -maxD <= D < 0 is fundamental then H[(-D)>>1] is the
- * ordinary class number of Q(sqrt(D)); junk at other entries. */
-static GEN
-allh(ulong maxD)
+static void
+allh_r(GEN H, ulong minD, ulong maxD)
 {
-  ulong a, A = usqrt(maxD/3), maxD2 = maxD >> 1;
-  GEN H = zero_zv(maxD2);
+  ulong a, A = usqrt(maxD/3), minD2 = (minD-1) >> 1, maxD2 = maxD >> 1;
   for (a = 1; a <= A; a++)
   {
-    ulong a2 = a << 1, aa = a*a, aa4 = aa << 2, b, c;
+    ulong a2 = a << 1, aa = a*a, aa4 = aa << 2, b, B;
     { /* b = 0 */
       ulong D = aa << 1;
-      for (c = a; D <= maxD2; c++) { H[D]++; D += a2; }
+      if (D <= minD2)
+        D += a2 * ((minD2 - D + a2) / a2);
+      for (; D <= maxD2; D += a2) H[D]++;
     }
+    B = aa4 - 1; /* 4a^2 - b^2 */
     for (b = 1; b < a; b++)
     {
-      ulong B = b*b, D = (aa4 - B) >> 1;
-      if (D > maxD2) continue;
-      H[D]++; D += a2; /* c = a */
-      for (c = a+1; D <= maxD2; c++) { H[D] += 2; D += a2; }
+      ulong D = B >> 1; /* (4a^2 - b^2) / 2 */
+      B -= (b << 1) + 1; if (D > maxD2) continue;
+      if (D > minD2)
+      {
+        H[D]++; D += a2; /* c = a */
+      } else
+        D += a2 * ((minD2 - D + a2) / a2);
+      for (; D <= maxD2; D += a2) H[D] += 2;
     }
     { /* b = a */
       ulong D = (aa4 - aa) >> 1;
-      for (c = a; D <= maxD2; c++) { H[D]++; D += a2; }
+      if (D <= minD2)
+        D += a2 * ((minD2 - D + a2) / a2);
+      for (; D <= maxD2; D += a2) H[D] ++;
     }
   }
+}
+
+/* return H s.t if -maxD <= D < 0 is fundamental then H[(-D)>>1] is the
+ * ordinary class number of Q(sqrt(D)); junk at other entries. */
+
+static GEN
+allh(ulong maxD)
+{
+  ulong maxD2 = maxD>>1, s = 1UL<<16;
+  GEN H = zero_zv(maxD2);
+  ulong a;
+  for (a = 0; a < maxD; a += s)
+    allh_r(H, a+1, minss(a+s,maxD));
   return H;
 }
 
@@ -291,7 +314,7 @@ mkDinfo(GEN c, long D, long h)
   long bi, pd, p1, p2;
   bi = disc_best_modinv(D);
   pd = (modinv_degree(&p1,&p2,bi) && (-D)%p1==0 && (-D)%p2==0)? h/2: h;
-  gel(c,1) = mkvecsmall4(D, h, bi, pd); return c;
+  return mkvec2(mkvecsmall4(D, h, bi, pd), c);
 }
 
 static GEN
@@ -301,21 +324,7 @@ ecpp_disclist_init(ulong maxdisc, GEN primelist)
   long t, ip, u, lp, lmerge;
   GEN merge, ev, od, Harr = allh(maxdisc); /* table of class numbers*/
   long lenv = maxdisc/4; /* max length of od/ev */
-  long N; /* maximum number of positive prime factors */
-
-  /* tuning paramaters blatantly copied from vecfactoru */
-  if (maxdisc < 510510UL) N = 7;
-  else if (maxdisc < 9699690UL) N = 8;
-    else if (maxdisc < 223092870UL) N = 9;
-  #ifdef LONG_IS_64BIT
-    else if (maxdisc < 6469693230UL) N = 10;
-    else if (maxdisc < 200560490130UL) N = 11;
-    else if (maxdisc < 7420738134810UL) N = 12;
-    else if (maxdisc < 304250263527210UL) N = 13;
-    else N = 16; /* don't bother */
-  #else
-    else N = 10;
-  #endif
+  long N = maxomegau(maxdisc) + 1;
 
   /* od[t] attached to discriminant 1-4*t, ev[t] attached to -4*t */
   od = cgetg(lenv + 1, t_VEC); /* contains 'long' at first: save memory */
@@ -358,7 +367,7 @@ ecpp_disclist_init(ulong maxdisc, GEN primelist)
   /* second pass: sieve by primes and record factorization */
   for (t = 1; t <= lenv; t++)
   {
-    if (od[t]) gel(od,t) = mkvec2(NULL, vecsmalltrunc_init(N));
+    if (od[t]) gel(od,t) = vecsmalltrunc_init(N);
     if (ev[t])
     {
       GEN F = vecsmalltrunc_init(N);
@@ -370,7 +379,7 @@ ecpp_disclist_init(ulong maxdisc, GEN primelist)
         default:id = 2; break;
       }
       vecsmalltrunc_append(F, id);
-      gel(ev,t) = mkvec2(NULL, F);
+      gel(ev,t) = F;
     }
   }
   lp = lg(primelist);
@@ -381,13 +390,13 @@ ecpp_disclist_init(ulong maxdisc, GEN primelist)
     for (t = (s*p+1)>>2; t <= lenv; t += p, s += 4)
     {
       GEN c = gel(od,t);
-      if (c) vecsmalltrunc_append(gel(c,2), ip);
+      if (c) vecsmalltrunc_append(c, ip);
     }
     s = 1;
     for (t = p; t <= lenv; t += p, s++)
     {
       GEN c = gel(ev,t);
-      if (c) vecsmalltrunc_append(gel(c,2), ip);
+      if (c) vecsmalltrunc_append(c, ip);
     }
   }
   /* merging the two arrays */
@@ -433,15 +442,29 @@ ecpp_primeord_init(GEN primelist, GEN disclist)
 static GEN
 ecpp_param_set(GEN tune, GEN x)
 {
+  pari_timer ti;
   pari_sp av = avma;
   ulong maxsqrt = uel(x,1), maxpcdg = uel(x,2), tdivexp = uel(x,3);
   ulong maxdisc = maxsqrt * maxsqrt;
-  GEN T = mkvecsmall3(maxsqrt, maxdisc, maxpcdg);
-  GEN Plist = ecpp_primelist_init(maxsqrt);
-  GEN Dlist = ecpp_disclist_init(maxdisc, Plist);
-  GEN Olist = ecpp_primeord_init(Plist, Dlist);
-  GEN primorial = primorial_vec(tdivexp);
-  return gerepilecopy(av, mkvec3(mkvec4(T,Plist,Dlist,Olist), primorial, tune));
+  GEN Plist, U, Utree, Dlist, Olist, primorial, T;
+  T = mkvecsmall3(maxsqrt, maxdisc, maxpcdg);
+  dbg_mode() {
+    err_printf(ANSI_BRIGHT_WHITE "\nECPP: parameters %Ps:\n"
+               "init time: " ANSI_RESET, x);
+    timer_start(&ti);
+  }
+  Plist = ecpp_primelist_init(maxsqrt);
+  U = zv_abs(Plist);
+  Utree = mkvec2(U, ZV_producttree(U));
+  dbg_mode() err_printf("%ld",timer_delay(&ti));
+  Dlist = ecpp_disclist_init(maxdisc, Plist);
+  dbg_mode() err_printf(", %ld",timer_delay(&ti));
+  Olist = ecpp_primeord_init(Plist, Dlist);
+  dbg_mode() err_printf(", %ld",timer_delay(&ti));
+  primorial = primorial_vec(tdivexp);
+  dbg_mode() err_printf(", %ld\n",timer_delay(&ti));
+  return gerepilecopy(av, mkvec3(mkvec5(T,Utree,Plist,Dlist,Olist),
+                                 primorial, tune));
 }
 
 /* cert contains [N, t, s, a4, [x, y]] as documented in ??ecpp; the following
@@ -469,6 +492,8 @@ cert_get_s(GEN x) { return gel(x,3); }
 INLINE GEN
 cert_get_q(GEN x) { return diviiexact(cert_get_m(x), cert_get_s(x)); }
 INLINE GEN
+cert_get_qlast(GEN x) { return cert_get_q( gel(x, lg(x)-1) ); }
+INLINE GEN
 cert_get_a4(GEN x) { return gel(x,4); }
 INLINE GEN
 cert_get_P(GEN x) { return gel(x,5); }
@@ -620,13 +645,12 @@ find_EP(GEN N, long D, GEN q, GEN g, GEN J, GEN s)
   }
 }
 
-/* Convert the disc. factorisation of a genus field to the
- * disc. factorisation of its real part. */
+/* Return the genus field. If real is set only the real part */
 static GEN
-realgenusfield(GEN Dfac, GEN sq, GEN p)
+genusfield(GEN Dfac, GEN sq, GEN p, long real)
 {
   long i, j, l = lg(Dfac), dn, n = 0;
-  GEN sn, s = gen_0, R = cgetg(l-1, t_VECSMALL);
+  GEN sn, s = gen_0, R = cgetg(l-1+!real, t_VECSMALL);
   for (i = 1; i < l; i++)
     if (Dfac[i] < 0) { n = i; break; }
   if (n == 0) pari_err_BUG("realgenusfield");
@@ -639,17 +663,40 @@ realgenusfield(GEN Dfac, GEN sq, GEN p)
       R[j++] = d > 0 ? d : d * dn;
       s = Fp_add(s, d > 0? si: Fp_mul(si, sn, p), p);
     }
+  if (!real)
+  {
+    R[j++] = dn;
+    s = Fp_add(s,sn, p);
+  }
   return mkvec2(R, s);
 }
 
 static GEN
-FpX_classtower_oneroot(GEN P, GEN Dfac, GEN sq, GEN p)
+realpart(GEN P, GEN R)
 {
-  pari_sp av = avma;
-  GEN C;
+  GEN G = galoisinit(rnfequation(P,R), NULL);
+  GEN T = galoisfixedfield(G, gel(gal_get_gen(G),2), 1, -1);
+  setvarn(T,1); return polredbest(T,0);
+}
+
+static GEN
+dihedralpol(long D, long l)
+{
+  GEN P = quadpoly0(utoineg(-D), 1);
+  GEN bnf = Buchall(P, 1, BIGDEFAULTPREC);
+  GEN x = bnrclassfield(bnf, utoipos(l), 0, BIGDEFAULTPREC);
+  pari_APPLY_same(realpart(P, gel(x,i)))
+}
+
+static GEN
+FpX_classtower_oneroot(GEN P, GEN Dinfo, GEN Dfac, GEN sq, GEN p)
+{
+  long D = Dinfo_get_D(Dinfo), h = Dinfo_get_h(Dinfo);
   if (degpol(P) > 1)
   {
-    GEN N = NULL, V = realgenusfield(Dfac, sq, p), v = gel(V,1), R = gel(V,2);
+    long real = !modinv_is_double_eta(Dinfo_get_bi(Dinfo));
+    GEN V = genusfield(Dfac, sq, p, real), v = gel(V,1), R = gel(V,2);
+    GEN N = NULL;
     long i, l = lg(v);
     for (i = 1; i < l; i++)
     {
@@ -658,19 +705,37 @@ FpX_classtower_oneroot(GEN P, GEN Dfac, GEN sq, GEN p)
       else
       {
         GEN cm = polcompositum0(N,Q,3);
-        N = gel(cm,1); P = gsubst(P, 1, gel(cm,2));
+        N = gel(cm,1); P = QXY_QXQ_evalx(P,gmael(cm,2,2),N);
       }
       P = liftpol_shallow(gmael(nffactor(N,P),1,1));
     }
+    if (h%3 == 0 && (!N || degpol(N)*3<degpol(P)))
+    {
+      GEN v = dihedralpol(D,3);
+      long i, l = lg(v);
+      for (i = 1; i < l; i++)
+      {
+        GEN Q = gel(v,i);
+        if (!N) N = Q;
+        else
+        {
+          GEN cm = polcompositum0(N, Q, 3);
+          N = gel(cm,1); P = QXY_QXQ_evalx(P,gmael(cm,2,2),N);
+          R = Fp_mul(R, gel(cm,4),p);
+        }
+        R = Fp_add(R, FpX_oneroot_split(Q, p), p);
+        P = liftpol_shallow(gmael(nffactor(N,P),1,1));
+        if (degpol(N)*3>=degpol(P)) break;
+      }
+    }
     if (N)
       P = FpXY_evalx(Q_primpart(P), R, p);
   }
-  C = FpX_oneroot_split(P, p);
-  return gerepileupto(av, C);
+  return P;
 }
 
 GEN
-ecpp_step2_worker(GEN S, GEN HD, GEN primelist)
+ecpp_step2_worker(GEN S, GEN HD, GEN primelist, long dbg)
 {
   pari_sp av = avma;
   pari_timer ti;
@@ -680,24 +745,26 @@ ecpp_step2_worker(GEN S, GEN HD, GEN primelist)
   GEN g = NDmqg_get_g(S), sq = NDmqg_get_sqrt(S);
   long D = Dinfo_get_D(Dinfo), inv = Dinfo_get_bi(Dinfo);
   GEN Dfacp = Dfac_to_p(Dinfo_get_Dfac(Dinfo), primelist);
-  long C2 = 0, C3 = 0, D1 = 0;
-  GEN c = getrand();
+  long C2 = 0, C3 = 0, C4 = 0, D1 = 0;
+  GEN P, c = getrand();
   setrand(gen_1); /* for reproducibility */
   /* C2: Find a root modulo N of polclass(D,inv) */
-  dbg_mode() timer_start(&ti);
-  rt = FpX_classtower_oneroot(HD, Dfacp, sq, N);
-  dbg_mode() C2 = timer_delay(&ti);
+  if (dbg >= 2) timer_start(&ti);
+  P = FpX_classtower_oneroot(HD, Dinfo, Dfacp, sq, N);
+  if (dbg >= 2) C2 = timer_delay(&ti);
+  rt = FpX_oneroot_split(P, N);
+  if (dbg >= 2) C3 = timer_delay(&ti);
   /* C3: Convert root from previous step into the appropriate j-invariant */
   J = Fp_modinv_to_j(rt, inv, N); /* root of polclass(D) */
-  dbg_mode() C3 = timer_delay(&ti);
+  if (dbg >= 2) C4 = timer_delay(&ti);
   /* D1: Find an elliptic curve E with a point P satisfying the theorem */
   s = diviiexact(m, q);
   EP = find_EP(N, D, q, g, J, s);
-  dbg_mode() D1 = timer_delay(&ti);
+  if (dbg >= 2) D1 = timer_delay(&ti);
 
   /* D2: Compute for t and s */
   t = subii(addiu(N, 1), m); /* t = N+1-m */
-  res = mkvec2(mkvec5(N, t, s, gel(EP,1), gel(EP,2)),mkvecsmall3(C2,C3,D1));
+  res = mkvec2(mkvec5(N, t, s, gel(EP,1), gel(EP,2)),mkvecsmall4(C2,C3,C4,D1));
   setrand(c);
   return gerepilecopy(av, res);
 }
@@ -736,7 +803,8 @@ ecpp_step2(GEN step1, GEN *X0, GEN primelist)
     gel(Hi, i) = HD;
   }
   gunclone_deep(db);
-  worker = snm_closure(is_entry("_ecpp_step2_worker"),mkvec(primelist));
+  worker = snm_closure(is_entry("_ecpp_step2_worker"),
+                       mkvec2(primelist,stoi(DEBUGLEVEL)));
   mt_queue_start_lim(&pt, worker, ls-1);
   for (j = 1; j < ls || pending; j++)
   {
@@ -751,7 +819,8 @@ ecpp_step2(GEN step1, GEN *X0, GEN primelist)
       err_printf(ANSI_BRIGHT_GREEN "\n[ %3d | %4ld bits]" ANSI_RESET, jj, expi(N));
       err_printf(" %6ld", time_record(X0, "C2", T[1]));
       err_printf(" %6ld", time_record(X0, "C3", T[2]));
-      err_printf(" %6ld", time_record(X0, "D1", T[3]));
+      err_printf(" %6ld", time_record(X0, "C4", T[3]));
+      err_printf(" %6ld", time_record(X0, "D1", T[4]));
     }
     gel(step2, jj) = gel(done,1);
   }
@@ -778,8 +847,6 @@ ecpp_parsqrt(GEN N, GEN param, GEN kroP, GEN sqrtlist, GEN g, long nb, long *nbs
   for (i=n, k=1; k<=nb && i<l; i++)
   {
     long pi = ordinv[i];
-    long s = kroP[pi];
-    if (s > 1) kroP[pi] = s = krosi(primelist[pi], N); /* update cache */
     if (kroP[pi] > 0)
       gel(W,k++) = stoi(primelist[pi]);
   }
@@ -858,17 +925,14 @@ D_collectcards(GEN N, GEN param, GEN* X0, GEN Dinfo, GEN sqrtlist, GEN g, GEN Dm
 {
   long i, l, corn_succ, wD, D = Dinfo_get_D(Dinfo);
   GEN U, V, sqrtofDmodN, Dfac = Dinfo_get_Dfac(Dinfo);
-  GEN primelist = ecpp_param_get_primelist(param);
   pari_timer ti;
 
   /* A1: Check (p*|N) = 1 for all primes dividing D */
   l = lg(Dfac);
   for (i = 1; i < l; i++)
   {
-    long j = Dfac[i], s = kroP[j];
-    if (s > 1) kroP[j] = s = krosi(primelist[j], N); /* update cache */
-    if (s == 0) return -1; /* N is composite */
-    if (s < 0) return 0;
+    long j = Dfac[i];
+    if (kroP[j] < 0) return 0;
   }
   /* A3: Get square root of D mod N */
   dbg_mode() timer_start(&ti);
@@ -1003,18 +1067,19 @@ mkNDmqg(GEN z, GEN N, GEN Dmq, GEN g, GEN sqrtlist)
  * pseudoprime, recursive call with N = q. May return gen_0 at toplevel
  * => N has a small prime divisor */
 static GEN
-N_downrun(GEN N, GEN param, GEN worker, GEN *X0, long *depth, long persevere)
+N_downrun(GEN N, GEN param, GEN worker, GEN *X0, long *depth, long persevere, long stopat)
 {
   pari_timer T, ti;
   long lgdisclist, lprimelist, nbsqrt = 0, t, i, j, expiN = expi(N);
   long persevere_next = 0, FAIL = 0;
   ulong maxpcdg;
-  GEN primelist, disclist, sqrtlist, g, Dmbatch, kroP;
+  GEN primelist, uprimelist, disclist, sqrtlist, g, Dmbatch, kroP, Np;
 
   dbg_mode() timer_start(&T);
   (*depth)++; /* we're going down the tree. */
   maxpcdg = ecpp_param_get_maxpcdg(param);
   primelist = ecpp_param_get_primelist(param);
+  uprimelist = ecpp_param_get_uprimelist(param);
   disclist = ecpp_param_get_disclist(param);
   lgdisclist = lg(disclist);
   t = tunevec_batchsize(expiN, param);
@@ -1029,7 +1094,7 @@ N_downrun(GEN N, GEN param, GEN worker, GEN *X0, long *depth, long persevere)
   /* A2: Check (p*|N) = 1 for all p */
   dbg_mode() timer_start(&ti);
   /* kroP[i] will contain (primelist[i] | N) */
-  kroP = const_vecsmall(lprimelist-1, 2/*sentinel*/);
+  kroP = cgetg(lprimelist,t_VECSMALL);
   switch(mod8(N))
   { /* primelist[1,2,3] = [8,-4,-8]; N is odd */
     case 1: kroP[1] = kroP[2] = kroP[3] = 1; break;
@@ -1037,8 +1102,12 @@ N_downrun(GEN N, GEN param, GEN worker, GEN *X0, long *depth, long persevere)
     case 5: kroP[1] = -1; kroP[2] = 1; kroP[3] =-1; break;
     case 7: kroP[1] =  1; kroP[2] =-1; kroP[3] =-1; break;
   }
+  Np = Z_ZV_mod_tree(N, gel(uprimelist,1), gel(uprimelist,2));
   for(i=4; i<lprimelist; i++)
-    kroP[i]= krosi(primelist[i],N);
+  {
+    if (Np[i]==0) return gen_0;
+    kroP[i] = kross(Np[i],primelist[i]);
+  }
   dbg_mode() timer_record(X0, "A2", &ti);
 
   /* Print the start of this iteration. */
@@ -1127,16 +1196,17 @@ N_downrun(GEN N, GEN param, GEN worker, GEN *X0, long *depth, long persevere)
         err_printf(ANSI_BRIGHT_BLUE "  h = %ld" ANSI_RESET, Dmq_get_h(Dmq));
       }
       /* q is pseudoprime */
-      if (expi(q) < 64) z = gen_1; /* q is prime; sentinel */
+      if (expi(q) < stopat) z = gen_1; /* q is prime; sentinel */
       else
       {
-        z = N_downrun(q, param, worker, X0, depth, persevere_next);
+        z = N_downrun(q, param, worker, X0, depth, persevere_next, stopat);
         if (!z) {
           dbg_mode() { char o = persevere? '<': '[', c = persevere? '>': ']';
                        err_printf(ANSI_CYAN "\n%c %3d | %4ld bits%c "
                                   ANSI_RESET, o, *depth, expiN, c); }
           continue; /* downrun failed */
         }
+        if (isintzero(z)) return z;
       }
       return mkNDmqg(z, N, Dmq, g, sqrtlist); /* SUCCESS */
     }
@@ -1160,7 +1230,6 @@ ecpp_flattencert(GEN x, long depth)
 {
   long i, l = depth+1;
   GEN ret = cgetg(l, t_VEC);
-  if (typ(x) != t_VEC) return gen_0;
   for (i = 1; i < l; i++) { gel(ret, i) = gel(x,1); x = gel(x,2); }
   return ret;
 }
@@ -1171,20 +1240,21 @@ ecpp_flattencert(GEN x, long depth)
  * - gen_0 (if N is composite)
  * - NULL (if FAIL) */
 static GEN
-ecpp_step1(GEN N, GEN param, GEN* X0)
+ecpp_step1(GEN N, GEN param, GEN* X0, long stopat)
 {
   pari_sp av = avma;
   long depth = 0;
   GEN worker = strtofunction("_ecpp_ispsp_worker");
-  GEN downrun = N_downrun(N, param, worker, X0, &depth, 1);
+  GEN downrun = N_downrun(N, param, worker, X0, &depth, 1, stopat);
   if (downrun == NULL) return gc_NULL(av);
+  if (typ(downrun) != t_VEC) return gen_0;
   return gerepilecopy(av, ecpp_flattencert(downrun, depth));
 }
 
 /* The input is an integer N.
    It uses the precomputation step0 done in ecpp_step0. */
 static GEN
-ecpp0(GEN N, GEN param)
+ecpp_param(GEN N, GEN param, long stopat)
 {
 
   GEN step1, answer, Tv, Cv, X0;
@@ -1195,14 +1265,14 @@ ecpp0(GEN N, GEN param)
   if (!BPSW_psp(N)) return gen_0;
 
   /* Check if we should even prove it. */
-  if (expi(N) < 64) return N;
+  if (expi(N) < stopat) return N;
 
   /* Timers and Counters */
-  Tv = mkvec4(zero_zv(5), zero_zv(3), zero_zv(3), zero_zv(1));
-  Cv = mkvec4(zero_zv(5), zero_zv(3), zero_zv(3), zero_zv(1));
+  Tv = mkvec4(zero_zv(5), zero_zv(3), zero_zv(4), zero_zv(1));
+  Cv = mkvec4(zero_zv(5), zero_zv(3), zero_zv(4), zero_zv(1));
   X0 = mkvec3(Tv, Cv, zero_zv(1));
 
-  step1 = ecpp_step1(N, param, &X0);
+  step1 = ecpp_step1(N, param, &X0, stopat);
   if (step1 == NULL) return gc_NULL(av);
   if (typ(step1) != t_VEC) { set_avma(av); return gen_0; }
 
@@ -1252,37 +1322,46 @@ static const long ecpp_tune[][4]=
 
 /* assume N BPSW-pseudoprime */
 GEN
-ecpp(GEN N)
+ecpp0(GEN C, long stopat)
 {
+  pari_sp av = avma;
   long expiN, i, tunelen;
-  GEN tune;
+  GEN N = C, tune;
+
+  /* Is it a partial certificate? */
+  if (typ(C) == t_VEC && check_ecppcert(C))
+    N = cert_get_qlast(C);
+  if (typ(N) != t_INT)
+    pari_err_TYPE("ecpp",C);
 
   /* Check if we should even prove it. */
   expiN = expi(N);
-  if (expiN < 64) return N;
+  if (stopat < 64) stopat = 64;
+  if (expiN < stopat) return C;
 
   tunelen = (expiN+499)/500;
   tune = cgetg(tunelen+1, t_VEC);
   for (i = 1; i <= tunelen && ecpp_tune[i-1][3]; i++)
     gel(tune,i) = mkvecsmall4(ecpp_tune[i-1][0], ecpp_tune[i-1][1],
                               ecpp_tune[i-1][2], ecpp_tune[i-1][3]);
-  for (; i <= tunelen; i++) gel(tune,i) = mkvecsmall4(200*(i-1),6*i-4,30,500*i);
+  for (; i <= tunelen; i++) gel(tune,i) = mkvecsmall4(100*(i+20),4*i+42,30,500*i);
   for(;;)
   {
-    GEN C, param, x = gel(tune, tunelen);
-    pari_timer T;
-    dbg_mode() timer_start(&T);
+    GEN C2, param, x = gel(tune, tunelen);
     param = ecpp_param_set(tune, x);
-    dbg_mode() {
-      err_printf(ANSI_BRIGHT_WHITE "\n%Ps" ANSI_RESET, x);
-      err_printf(ANSI_WHITE "  %8ld" ANSI_RESET, timer_delay(&T));
-    }
-    if ((C = ecpp0(N, param))) return C;
+    C2 = ecpp_param(N, param, stopat);
+    if (C2)
+      return typ(C)==t_INT? C2: gerepilecopy(av, shallowconcat(C,C2));
     x[1] *= 2;
     x[2] *= 2;
     x[3] = minss(x[3]+1, 30);
   }
 }
+
+GEN
+ecpp(GEN N)
+{ return ecpp0(N, 0); }
+
 long
 isprimeECPP(GEN N)
 {
@@ -1495,7 +1574,7 @@ ecppisvalid_i(GEN cert)
     GEN certi = gel(cert, i);
     GEN qq = gel(check,i), N = cert_get_N(certi);
     if (isintzero(qq)) return 0;
-    if (i > 1 && !equalii(N, q)) return 0;
+    if (i > 1 && !equalii(N, q)) return 0; /* N of this entry doesn't match q of previous */
     q = qq;
   }
   return 1;
