@@ -20,6 +20,8 @@ Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA. */
 #include "pari.h"
 #include "paripriv.h"
 
+#define DEBUGLEVEL DEBUGLEVEL_ellanal
+
 struct baby_giant
 {
   GEN baby, giant, sum;
@@ -376,7 +378,7 @@ ellL1_eval(GEN e, GEN vec, struct lcritical *C, GEN t, long prec)
 static GEN
 ellL1_der(GEN e, GEN vec, struct lcritical *C, GEN t, long der, long prec)
 {
-  GEN r = polcoef(ellL1_eval(e, vec, C, t, prec), der, 0);
+  GEN r = polcoef_i(ellL1_eval(e, vec, C, t, prec), der, 0);
   r = gmul(r,powIs(C->real == 1 ? -der: 1-der));
   return gmul(real_i(r), mpfact(der));
 }
@@ -585,7 +587,7 @@ lambda1(GEN E, GEN nv, GEN p, long prec)
     long j, m = kod - 4, nl = 1 + (m >> 1L);
     res = cgetg(nl, t_VEC);
     for (j = 1; j < nl; j++)
-      gel(res, j) = gmul(lp, gsubgs(gdivgs(sqru(j), n), j)); /* j^2/n - j */
+      gel(res, j) = gmul(lp, gsubgs(gdivgu(sqru(j), n), j)); /* j^2/n - j */
   }
   else if (kod < -4)
     res = mkvec2(negr(lp), shiftr(mulrs(lp, kod), -2));
@@ -670,7 +672,7 @@ logplusQ(GEN t, long prec)
 
 /* See GTM239, p532, Th 8.1.18
  * Return M such that h_naive <= M */
-static GEN
+GEN
 hnaive_max(GEN ell, GEN ht)
 {
   const long prec = LOWDEFAULTPREC; /* minimal accuracy */
@@ -678,7 +680,7 @@ hnaive_max(GEN ell, GEN ht)
   GEN logd   = glog(absi_shallow(ell_get_disc(ell)), prec);
   GEN logj   = logplusQ(j, prec);
   GEN hj     = heightQ(j, prec);
-  GEN logb2p = signe(b2)? addrr(logplusQ(gdivgs(b2, 12),prec), mplog2(prec))
+  GEN logb2p = signe(b2)? addrr(logplusQ(gdivgu(b2, 12),prec), mplog2(prec))
                         : real_1(prec);
   GEN mu     = addrr(divru(addrr(logd, logj),6), logb2p);
   return addrs(addrr(addrr(ht, divru(hj,12)), mu), 2);
@@ -705,23 +707,22 @@ qimag2(GEN Q)
 /***************************************************/
 
 static GEN
-qfb_mult(GEN Q, GEN M)
+qfb_mult(GEN Q, GEN a, GEN b, GEN c, GEN d)
 {
-  GEN A = gel(Q, 1) , B = gel(Q, 2) , C = gel(Q, 3);
-  GEN a = gcoeff(M, 1, 1), b = gcoeff(M, 1, 2);
-  GEN c = gcoeff(M, 2, 1), d = gcoeff(M, 2, 2);
-  GEN W1 = addii(addii(mulii(sqri(a), A), mulii(mulii(c, a), B)), mulii(sqri(c), C));
+  GEN A = gel(Q, 1) , B = gel(Q, 2), C = gel(Q, 3), D = qfb_disc(Q);
+  GEN a2 = sqri(a), b2 = sqri(b), c2 = sqri(c), d2 = sqri(d);
+  GEN ad = mulii(d, a), bc = mulii(b, c), e = subii(ad, bc);
+  GEN W1 = addii(addii(mulii(a2, A), mulii(mulii(c, a), B)), mulii(c2, C));
+  GEN W3 = addii(addii(mulii(b2, A), mulii(mulii(d, b), B)), mulii(d2, C));
   GEN W2 = addii(addii(mulii(mulii(shifti(b,1), a), A),
-                       mulii(addii(mulii(d, a), mulii(c, b)), B)),
+                       mulii(addii(ad, bc), B)),
                  mulii(mulii(shifti(d,1), c), C));
-  GEN W3 = addii(addii(mulii(sqri(b), A), mulii(mulii(d, b), B)), mulii(sqri(d), C));
-  GEN D = gcdii(W1, gcdii(W2, W3));
-  if (!equali1(D)) {
-    W1 = diviiexact(W1,D);
-    W2 = diviiexact(W2,D);
-    W3 = diviiexact(W3,D);
+  if (!equali1(e)) {
+    W1 = diviiexact(W1,e);
+    W2 = diviiexact(W2,e);
+    W3 = diviiexact(W3,e);
   }
-  return qfi(W1, W2, W3);
+  return mkqfb(W1, W2, W3, D);
 }
 
 #ifdef DEBUG
@@ -729,32 +730,24 @@ static void
 best_point_old(GEN Q, GEN NQ, GEN f, GEN *u, GEN *v)
 {
   long n, k;
-  GEN U, c, d;
-  GEN A = gel(f, 1);
-  GEN B = gel(f, 2);
-  GEN C = gel(f, 3);
-  GEN q = qfi(mulii(NQ, C), negi(B), diviiexact(A, NQ));
+  GEN U, c, d, A = gel(f,1), B = gel(f,2), C = gel(f,3), D = qfb_disc(f);
+  GEN q = mkqfb(mulii(NQ, C), negi(B), diviiexact(A, NQ), D);
   redimagsl2(q, &U);
   *u = c = gcoeff(U, 1, 1);
   *v = d = gcoeff(U, 2, 1);
-  if (equali1(gcdii(mulii(*u, NQ), mulii(*v, Q))))
-    return;
-  for (n = 1; ; n++)
+  if (equali1(gcdii(mulii(*u, NQ), mulii(*v, Q)))) return;
+  for (n = 1;; n++)
   {
-    for (k = -n; k<=n; k++)
+    for (k = -n; k <= n; k++)
     {
       *u = addis(c, k); *v = addiu(d, n);
-      if (equali1(ggcd(mulii(*u, NQ), mulii(*v, Q))))
-        return;
-      *v= subiu(d, n);
-      if (equali1(ggcd(mulii(*u, NQ), mulii(*v, Q))))
-        return;
+      if (equali1(gcdii(mulii(*u, NQ), mulii(*v, Q)))) return;
+      *v = subiu(d, n);
+      if (equali1(gcdii(mulii(*u, NQ), mulii(*v, Q)))) return;
       *u = addiu(c, n); *v = addis(d, k);
-      if (equali1(ggcd(mulii(*u, NQ), mulii(*v, Q))))
-        return;
+      if (equali1(gcdii(mulii(*u, NQ), mulii(*v, Q)))) return;
       *u = subiu(c, n);
-      if (equali1(ggcd(mulii(*u, NQ), mulii(*v, Q))))
-        return;
+      if (equali1(gcdii(mulii(*u, NQ), mulii(*v, Q)))) return;
     }
   }
 }
@@ -788,12 +781,13 @@ static void
 best_point(GEN Q, GEN NQ, GEN f, GEN *pu, GEN *pv)
 {
   GEN a = mulii(NQ, gel(f,3)), b = negi(gel(f,2)), c = diviiexact(gel(f,1), NQ);
-  GEN D = absi_shallow( qfb_disc(f) );
-  GEN U, qr = redimagsl2(qfi(a,b,c), &U);
+  GEN D = qfb_disc(f);
+  GEN U, qr = redimagsl2(mkqfb(a, b, c, D), &U);
   GEN A = gel(qr,1), B = gel(qr,2), A2 = shifti(A,1), AA4 = sqri(A2);
   GEN V, best;
   long y;
 
+  D = absi_shallow(D);
   /* 4A qr(x,y) = (2A x + By)^2 + D y^2
    * Write x = x0(y) + i, where x0 is an integer minimum
    * (the smallest in case of tie) of x-> qr(x,y), for given y.
@@ -867,7 +861,7 @@ best_point(GEN Q, GEN NQ, GEN f, GEN *pu, GEN *pv)
   }
 #ifdef DEBUG
   {
-    GEN oldu, oldv, F = qfi(a,b,c);
+    GEN oldu, oldv, F = mkqfb(a, b, c, qfb_disc(f));
     best_point_old(Q, NQ, f, &oldu, &oldv);
     if (!equalii(oldu, *pu) || !equalii(oldv, *pv))
     {
@@ -885,18 +879,17 @@ best_point(GEN Q, GEN NQ, GEN f, GEN *pu, GEN *pv)
 }
 
 static GEN
-best_lift(GEN N, GEN Q, GEN NQ, GEN f)
+best_lift(GEN Q, GEN NQ, GEN f)
 {
-  GEN a,b,c,d,M;
+  GEN a, b, c, d, dQ, cNQ;
   best_point(Q, NQ, f, &c, &d);
-  (void)bezout(mulii(d, Q), mulii(NQ, c), &a, &b);
-  M = mkmat2( mkcol2(mulii(d, Q), mulii(negi(N), c)),
-              mkcol2(b, mulii(a, Q)));
-  return qfb_mult(f, M);
+  dQ = mulii(d, Q); cNQ = mulii(NQ, c);
+  (void)bezout(dQ, cNQ, &a, &b);
+  return qfb_mult(f, dQ, b, mulii(negi(Q),cNQ), mulii(a,Q));
 }
 
 static GEN
-lift_points(GEN N, GEN listQ, GEN f, GEN *pt, GEN *pQ)
+lift_points(GEN listQ, GEN f, GEN *pt, GEN *pQ)
 {
   pari_sp av = avma;
   GEN yf = gen_0, tf = NULL, Qf = NULL;
@@ -904,11 +897,10 @@ lift_points(GEN N, GEN listQ, GEN f, GEN *pt, GEN *pQ)
   for (k = 1; k < l; ++k)
   {
     GEN c = gel(listQ, k), Q = gel(c,1), NQ = gel(c,2);
-    GEN t = best_lift(N, Q, NQ, f), y = qimag2(t);
+    GEN t = best_lift(Q, NQ, f), y = qimag2(t);
     if (gcmp(y, yf) > 0) { yf = y; Qf = Q; tf = t; }
   }
-  gerepileall(av, 3, &tf, &Qf, &yf);
-  *pt = tf; *pQ = Qf; return yf;
+  *pt = tf; *pQ = Qf; return gc_all(av, 3, &yf, pt, pQ);
 }
 
 /***************************/
@@ -919,8 +911,7 @@ static GEN
 ltwist1(GEN E, GEN d, long bitprec)
 {
   pari_sp av = avma;
-  GEN Ed = ellinit(elltwist(E, d), NULL, DEFAULTPREC);
-  GEN z = ellL1_bitprec(Ed, 0, bitprec);
+  GEN Ed = elltwist(E, d), z = ellL1_bitprec(Ed, 0, bitprec);
   obj_free(Ed); return gerepileuptoleaf(av, z);
 }
 
@@ -962,7 +953,76 @@ heegner_indexmultD(GEN faN, GEN a, long D, GEN sqrtD)
 }
 
 static GEN
-heegner_try_point(GEN E, GEN lambdas, GEN ht, GEN z, long prec)
+nf_to_basis(GEN nf, GEN x)
+{
+  x = nf_to_scalar_or_basis(nf, x);
+  if (typ(x)!=t_COL)
+    x = scalarcol(x, nf_get_degree(nf));
+  return x;
+}
+
+static GEN
+etnf_to_basis(GEN et, GEN x)
+{
+  long i, l = lg(et);
+  GEN V = cgetg(l, t_VEC);
+  for (i = 1; i < l; i++)
+    gel(V,i) = nf_to_basis(gel(et,i), x);
+  return shallowconcat1(V);
+}
+
+static GEN
+etnf_get_M(GEN et)
+{
+  long i, l = lg(et);
+  GEN V = cgetg(l, t_VEC);
+  for (i = 1; i < l; i++)
+    gel(V,i)=nf_get_M(gel(et,i));
+  return shallowmatconcat(diagonal(V));
+}
+
+static long
+etnf_get_varn(GEN et)
+{
+  return nf_get_varn(gel(et,1));
+}
+
+static GEN
+heegner_descent_try_point(GEN nfA, GEN z, GEN den, long prec)
+{
+  pari_sp av = avma;
+  GEN etal = gel(nfA,1), A = gel(nfA,2), cb = gel(nfA,3);
+  GEN al = gel(nfA,4), th = gel(nfA, 5);
+  GEN et = gel(etal,1), zk = gel(etal, 2), T = gel(etal,3);
+  GEN M = etnf_get_M(et);
+  long i, j, n = lg(th)-1, l = lg(al);
+  GEN u2 = gsqr(gel(cb,1)), r = gel(cb,2);
+  GEN zz = gdiv(gsub(z,r), u2);
+  GEN be = cgetg(n+1, t_COL);
+  for (j = 1; j < l; j++)
+  {
+    GEN aj = gel(al, j), Aj = gel(A,j);
+    for (i = 1; i <= n; i++)
+      gel(be,i) = gsqrt(gmul(gsub(zz, gel(th,i)), gel(aj,i)), prec);
+    for (i = 0; i <= (1<<(n-1))-1; i++)
+    {
+      long eps;
+      GEN s = gmul(den, RgM_solve_realimag(M, be));
+      GEN S = grndtoi(s, &eps), V, S2;
+      gel(be,1+odd(i)) = gneg(gel(be,1+odd(i)));
+      if (eps > -7) continue;
+      S2 = QXQ_sqr(RgV_RgC_mul(zk, S), T);
+      V = gdiv(QXQ_mul(S2, Aj, T), sqri(den));
+      if (typ(V) != t_POL || degpol(V) != 1) continue;
+      if (gequalm1(gel(V,3)))
+        return gerepileupto(av,gadd(gmul(gel(V,2),u2),r));
+    }
+  }
+  return gc_NULL(av);
+}
+
+static GEN
+heegner_try_point(GEN E, GEN nfA, GEN lambdas, GEN ht, GEN z, long prec)
 {
   long l = lg(lambdas);
   long i, eps;
@@ -972,23 +1032,25 @@ heegner_try_point(GEN E, GEN lambdas, GEN ht, GEN z, long prec)
   {
     GEN logd = shiftr(gsub(rh, gel(lambdas, i)), -1);
     GEN d, approxd = gexp(logd, prec);
-    if (DEBUGLEVEL > 2)
-      err_printf("Trying lambda number %ld, logd=%Ps, approxd=%Ps\n", i, logd, approxd);
     d = grndtoi(approxd, &eps);
     if (signe(d) > 0 && eps<-10)
     {
-      GEN X, ylist, d2 = sqri(d), approxn = mulir(d2, x);
-      if (DEBUGLEVEL > 2) err_printf("approxn=%Ps  eps=%ld\n", approxn,eps);
-      X = gdiv(ground(approxn), d2);
-      ylist = ellordinate(E, X, prec);
-      if (lg(ylist) > 1)
+      GEN X, ylist;
+      if (DEBUGLEVEL > 2)
+        err_printf("\nTrying lambda number %ld, logd=%Ps, approxd=%Ps\n", i, logd, approxd);
+      X = heegner_descent_try_point(nfA, x, d, prec);
+      if (X)
       {
-        GEN P = mkvec2(X, gel(ylist, 1));
-        GEN hp = ghell(E,P,prec);
-        if (signe(hp) && cmprr(hp, shiftr(ht,1)) < 0 && cmprr(hp, shiftr(ht,-1)) > 0)
-          return P;
-        if (DEBUGLEVEL)
-          err_printf("found non-Heegner point %Ps\n", P);
+        ylist = ellordinate(E, X, prec);
+        if (lg(ylist) > 1)
+        {
+          GEN P = mkvec2(X, gel(ylist, 1));
+          GEN hp = ellheight(E,P,prec);
+          if (signe(hp) && cmprr(hp, shiftr(ht,1)) < 0 && cmprr(hp, shiftr(ht,-1)) > 0)
+            return P;
+          if (DEBUGLEVEL)
+            err_printf("found non-Heegner point %Ps\n", P);
+        }
       }
     }
   }
@@ -996,25 +1058,25 @@ heegner_try_point(GEN E, GEN lambdas, GEN ht, GEN z, long prec)
 }
 
 static GEN
-heegner_find_point(GEN e, GEN om, GEN ht, GEN z1, long k, long prec)
+heegner_find_point(GEN e, GEN nfA, GEN om, GEN ht, GEN z1, long k, long prec)
 {
   GEN lambdas = lambdalist(e, prec);
   pari_sp av = avma;
   long m;
   GEN Ore = gel(om, 1), Oim = gel(om, 2);
   if (DEBUGLEVEL)
-    err_printf("%ld*%ld multipliers to test\n",k,lg(lambdas)-1);
+    err_printf("%ld*%ld multipliers to test: ",k,lg(lambdas)-1);
   for (m = 0; m < k; m++)
   {
     GEN P, z2 = divru(addrr(z1, mulsr(m, Ore)), k);
     if (DEBUGLEVEL > 2)
-      err_printf("Trying multiplier %ld\n",m);
-    P = heegner_try_point(e, lambdas, ht, z2, prec);
+      err_printf("%ld ",m);
+    P = heegner_try_point(e, nfA, lambdas, ht, z2, prec);
     if (P) return P;
     if (signe(ell_get_disc(e)) > 0)
     {
       z2 = gadd(z2, gmul2n(Oim, -1));
-      P = heegner_try_point(e, lambdas, ht, z2, prec);
+      P = heegner_try_point(e, nfA, lambdas, ht, z2, prec);
       if (P) return P;
     }
     set_avma(av);
@@ -1080,10 +1142,9 @@ testDisc(GEN bad, long d) { return !bad || ugcdiu(bad, -d) == 1; }
 /* bad = product of bad primes. Return the NDISC largest fundamental
  * discriminants D < d such that (D,bad) = 1 and d is a square mod 4N */
 static GEN
-listDisc(GEN fa4N, GEN bad, long d)
+listDisc(GEN fa4N, GEN bad, long d, long ndisc)
 {
-  const long NDISC = 10;
-  GEN v = cgetg(NDISC+1, t_VECSMALL);
+  GEN v = cgetg(ndisc+1, t_VECSMALL);
   pari_sp av = avma;
   long j = 1;
   for(;;)
@@ -1092,7 +1153,7 @@ listDisc(GEN fa4N, GEN bad, long d)
     if (testDisc(bad,d) && unegisfundamental(-d) && Zn_issquare(stoi(d), fa4N))
     {
       v[j++] = d;
-      if (j > NDISC) break;
+      if (j > ndisc) break;
     }
     set_avma(av);
   }
@@ -1101,10 +1162,10 @@ listDisc(GEN fa4N, GEN bad, long d)
 /* L = vector of [q1,q2] or [q1,q2,q2']
  * cd = (b^2 - D)/(4N) */
 static void
-listfill(GEN N, GEN b, GEN c, GEN d, GEN L, long *s)
+listfill(GEN N, GEN b, GEN c, GEN d, GEN D, GEN L, long *s)
 {
   long k, l = lg(L);
-  GEN add, frm2, a = mulii(d, N), V = mkqfi(a,b,c), frm = redimag(V);
+  GEN add, frm2, a = mulii(d, N), V = mkqfb(a,b,c,D), frm = qfbred_i(V);
   for (k = 1; k < l; ++k)
   { /* Lk = [v,frm] or [v,frm,frm2] */
     GEN Lk = gel(L,k);
@@ -1117,7 +1178,7 @@ listfill(GEN N, GEN b, GEN c, GEN d, GEN L, long *s)
         return;
       }
   }
-  frm2 = redimag( mkqfi(d, negi(b), mulii(c,N)) );
+  frm2 = qfbred_i(mkqfb(d, negi(b), mulii(c,N), D));
   add = gequal(frm, frm2)? mkvec2(V,frm): mkvec3(V,frm,frm2);
   vectrunc_append(L, add);
   *s += lg(add) - 2;
@@ -1128,7 +1189,7 @@ listheegner(GEN N, GEN faN4, GEN listQ, GEN D)
 {
   pari_sp av = avma;
   const long kmin = 30;
-  long h = itos(gel(quadclassunit0(D, 0, NULL, DEFAULTPREC), 1));
+  long h = itos(quadclassno(D));
   GEN ymin, b = Zn_sqrt(D, faN4), L = vectrunc_init(h+1);
   long l, k, s = 0;
   for (k = 0; k < kmin || s < h; k++)
@@ -1140,14 +1201,14 @@ listheegner(GEN N, GEN faN4, GEN listQ, GEN D)
     for (i = 1; i < l; i++)
     {
       GEN d = gel(div, i), c = gel(div, l-i); /* cd = C */
-      listfill(N, bk, c, d, L, &s);
+      listfill(N, bk, c, d, D, L, &s);
     }
   }
   l = lg(L); ymin = NULL;
   for (k = 1; k < l; k++)
   {
     GEN t, Q, Lk = gel(L,k), f = gel(Lk,1);
-    GEN y = lift_points(N, listQ, f, &t, &Q);
+    GEN y = lift_points(listQ, f, &t, &Q);
     gel(L, k) = mkvec3(t, stoi(lg(Lk) - 2), Q);
     if (!ymin || gcmp(y, ymin) < 0) ymin = y;
   }
@@ -1170,7 +1231,7 @@ rootno(GEN Q, GEN P, GEN R)
 
 static void
 heegner_find_disc(GEN *points, GEN *coefs, long *pind, GEN E,
-                  GEN indmult, long prec)
+                  GEN indmult, long ndisc, long prec)
 {
   long d = 0;
   GEN faN4, bad, N, faN, listQ, listR;
@@ -1183,7 +1244,7 @@ heegner_find_disc(GEN *points, GEN *coefs, long *pind, GEN E,
   for(;;)
   {
     pari_sp av = avma;
-    GEN list, listD = listDisc(faN4, bad, d);
+    GEN list, listD = listDisc(faN4, bad, d, ndisc);
     long k, l = lg(listD);
     list = cgetg(l, t_VEC);
     for (k = 1; k < l; ++k)
@@ -1248,22 +1309,55 @@ ellanal_globalred_all(GEN e, GEN *cb, GEN *N, GEN *tam)
   return E;
 }
 
+static GEN
+vecelnfembed(GEN x, GEN M, GEN et)
+{ pari_APPLY_same(gmul(M, etnf_to_basis(et, gel(x,i)))) }
+
+static GEN
+QXQV_inv(GEN x, GEN T)
+{ pari_APPLY_same(QXQ_inv(gel(x,i), T)) }
+
+static GEN
+etnfnewprec(GEN x, long prec)
+{ pari_APPLY_same(nfnewprec(gel(x,i),prec)) }
+
+static GEN
+vec_etnf_to_basis(GEN et, GEN x)
+{ pari_APPLY_same(etnf_to_basis(et,gel(x,i))) }
+
+static GEN
+makenfA(GEN sel, GEN A, GEN cb)
+{
+  GEN etal = gel(sel,1), T = gel(etal,3);
+  GEN et = gel(etal,1), M = etnf_get_M(et);
+  long v = etnf_get_varn(et);
+  GEN al = vecelnfembed(A, M, et);
+  GEN th = gmul(M, etnf_to_basis(et, pol_x(v)));
+  return mkvec5(etal,QXQV_inv(A, T),cb,al,th);
+}
+
 GEN
 ellheegner(GEN E)
 {
   pari_sp av = avma;
   GEN z, P, ht, points, coefs, s, om, indmult;
-  long ind, lint, k, l, wtor, etor;
+  GEN sel, etal, et, cbb, A, dAi, T, Ag, At;
+  long ind, indx, lint, k, l, wtor, etor, ndisc, ltors2, selrank;
   long bitprec = 16, prec = nbits2prec(bitprec)+1;
   pari_timer ti;
-  GEN N, cb, tam, torsion;
-
+  GEN N, cb, tam, torsion, nfA;
   E = ellanal_globalred_all(E, &cb, &N, &tam);
   if (ellrootno_global(E) == 1)
     pari_err_DOMAIN("ellheegner", "(analytic rank)%2","=",gen_0,E);
   torsion = elltors(E);
   wtor = itos( gel(torsion,1) ); /* #E(Q)_tor */
   etor = wtor > 1? itou(gmael(torsion, 2, 1)): 1; /* exponent of E(Q)_tor */
+  sel = ell2selmer_basis(E, &cbb, prec);
+  etal = gel(sel,1); A = gel(sel,2); et = gel(etal,1); T = gel(etal,3);
+  ltors2 = lg(et)-2; selrank = lg(A)-1;
+  Ag = selrank > ltors2+1 ? pol_1(etnf_get_varn(et)): gel(A,selrank);
+  At = vecslice(A,1,ltors2);
+  dAi = gsupnorm(vec_etnf_to_basis(et,A),prec);
   while (1)
   {
     GEN hnaive, l1;
@@ -1278,24 +1372,33 @@ ellheegner(GEN E)
     if (DEBUGLEVEL) err_printf("Expected height=%Ps\n", ht);
     hnaive = hnaive_max(E, ht);
     if (DEBUGLEVEL) err_printf("Naive height <= %Ps\n", hnaive);
-    bitneeded = itos(gceil(divrr(hnaive, mplog2(prec)))) + 10;
+    hnaive = gadd(shiftr(hnaive,-1),glog(dAi,prec));
+    bitneeded = itos(gceil(divrr(hnaive, mplog2(prec)))) + 12;
     if (DEBUGLEVEL) err_printf("precision = %ld\n", bitneeded);
     if (bitprec>=bitneeded) break;
     bitprec = bitneeded;
     prec = nbits2prec(bitprec)+1;
   }
   indmult = heegner_indexmult(om, wtor, tam, prec);
-  heegner_find_disc(&points, &coefs, &ind, E, indmult, prec);
+  ndisc = maxss(10, (long) rtodbl(ht)/10);
+  heegner_find_disc(&points, &coefs, &ind, E, indmult, ndisc, prec);
   if (DEBUGLEVEL) timer_start(&ti);
   s = heegner_psi(E, N, points, bitprec);
   if (DEBUGLEVEL) timer_printf(&ti,"heegner_psi");
   l = lg(points);
   z = mulsr(coefs[1], gel(s, 1));
   for (k = 2; k < l; ++k) z = addrr(z, mulsr(coefs[k], gel(s, k)));
-  if (DEBUGLEVEL) err_printf("z=%Ps\n", z);
   z = gsub(z, gmul(gel(om,1), ground(gdiv(z, gel(om,1)))));
+  if (DEBUGLEVEL) err_printf("z=%.*Pg\n",nbits2ndec(bitprec), z);
   lint = wtor > 1 ? ugcd(ind, etor): 1;
-  P = heegner_find_point(E, om, ht, gmulsg(2*lint, z), lint*2*ind, prec);
+  indx = lint*2*ind;
+  if (vals(indx) >= vals(etor))
+    A = mkvec(Ag);
+  else
+    A = mkvec2(Ag, QXQ_mul(Ag, gel(At,1), T));
+  gmael(sel,1,1) = etnfnewprec(et, prec);
+  nfA = makenfA(sel, A, cbb);
+  P = heegner_find_point(E, nfA, om, ht, gmulsg(2*lint, z), indx, prec);
   if (DEBUGLEVEL) timer_printf(&ti,"heegner_find_point");
   if (cb) P = ellchangepointinv(P, cb);
   return gerepilecopy(av, P);
