@@ -212,7 +212,7 @@ doellQp_root(GEN E, long prec)
   }
   a = cvtop(a, p, prec);
   if (alpha) setvalp(a, valp(a)+alpha);
-  return gsub(a, gdivgs(ell_get_b2(E), 12));
+  return gsub(a, gdivgu(ell_get_b2(E), 12));
 }
 GEN
 ellQp_root(GEN E, long prec)
@@ -223,7 +223,7 @@ static void
 doellQp_ab(GEN E, GEN *pta, GEN *ptb, long prec)
 {
   GEN b2 = ell_get_b2(E), b4 = ell_get_b4(E), e1 = ellQp_root(E, prec);
-  GEN w, u, t = gadd(gdivgs(b2,4), gmulsg(3,e1)), p = ellQp_get_p(E);
+  GEN w, u, t = gadd(gdivgu(b2,4), gmulsg(3,e1)), p = ellQp_get_p(E);
   w = Qp_sqrt(gmul2n(gadd(b4,gmul(e1,gadd(b2,gmulsg(6,e1)))),1));
   u = gadd(t,w);
   /* Decide between w and -w: we want v(a-b) > v(b) */
@@ -328,7 +328,7 @@ ellformalw(GEN e, long n, long v)
   t = pol_x(v);
   checkell(e);
   a1 = ell_get_a1(e); a2 = ell_get_a2(e); a3 = ell_get_a3(e);
-  a4 = ell_get_a4(e); a6 = ell_get_a6(e); a63 = gmulgs(a6,3);
+  a4 = ell_get_a4(e); a6 = ell_get_a6(e); a63 = gmulgu(a6,3);
   w[1] = evalsigne(1)|evalvarn(v)|evalvalp(3);
   gel(w,2) = gen_1; /* t^3 + O(t^4) */
   /* use Newton iteration, doubling accuracy at each step
@@ -782,7 +782,7 @@ ellpadics2_tate(GEN Ep, long n)
     if ((m & 31) == 0) s = gerepileuptoint(av, s);
   }
   s = subui(1, muliu(s,24));
-  s = gdivgs(gsub(b2, gdiv(s,u2)), 12);
+  s = gdivgu(gsub(b2, gdiv(s,u2)), 12);
   return precp_fix(s,n);
 }
 
@@ -790,12 +790,12 @@ GEN
 ellpadicfrobenius(GEN E, ulong p, long n)
 {
   checkell(E);
-  if (p < 2) pari_err_DOMAIN("ellpadicfrobenius","p", "<", gen_2, utoi(p));
+  if (p < 2) pari_err_DOMAIN("ellpadicfrobenius","p", "<", gen_2, utoipos(p));
   switch(ell_get_type(E))
   {
     case t_ELL_Q: break;
     case t_ELL_Qp: if (equaliu(ellQp_get_p(E), p)) break;
-    default: pari_err_TYPE("ellpadicfrobenius",utoi(p));
+    default: pari_err_TYPE("ellpadicfrobenius", E);
   }
   return hyperellpadicfrobenius(ec_bmodel(E), p, n);
 }
@@ -913,12 +913,14 @@ ellpadicbsd(GEN E, GEN p, long n, GEN D)
 {
   const long MAXR = 30;
   pari_sp av = avma;
-  GEN W, ED, tam, ND, C, apD, U = NULL;/*-Wall*/
+  GEN D0, W, ED, tam, ND, C, apD, U = NULL;/*-Wall*/
   long r, vN;
   checkell(E);
   if (D && isint1(D)) D = NULL;
+  D0 = ellminimaltwistcond(E); if (isint1(D0)) D0 = NULL;
+  if (D0) { E = elltwist(E, D0); D = D? coredisc(mulii(D, D0)): D0; }
   W = ellpadicL_symbol(E, p, gen_0, D);
-  ED = D? ellinit(elltwist(E,D), gen_1, 0): E;
+  ED = D? elltwist(E, D): E;
   ED = ellanal_globalred_all(ED, NULL, &ND, &tam);
   vN = Z_pval(ND, p); /* additive reduction ? */
   if (vN >= 2) pari_err_DOMAIN("ellpadicbsd","v_p(N(E_D))",">",gen_1,stoi(vN));
@@ -945,7 +947,7 @@ ellpadicbsd(GEN E, GEN p, long n, GEN D)
   else if (dvdii(ND,p))
   {
     if (equalim1(apD)) /* divide by 1-1/a_p */
-      U = gdivgs(U, 2);
+      U = gdivgu(U, 2);
     else
     { /* ap = 1 */
       GEN EDp = ellinit(ED, zeropadic(p,n), 0);
