@@ -498,14 +498,17 @@ lfunthetacost(GEN ldata, GEN tdom, long m, long bitprec)
   if (al)
   { /* t = rho e^(i*al), T^(1/c) = Re(t^(1/c)) > 0, T = rho cos^c(al/c) */
     double z = cos(al/c);
-    T = (d == 2 && typ(tdom) != t_VEC)? gtodouble(real_i(tdom)): rho*pow(z,c);
     if (z <= 0)
       pari_err_DOMAIN("lfunthetaneed", "arg t", ">", dbltor(c*M_PI/2), tdom);
+    T = (d == 2 && typ(tdom) != t_VEC)? gtodouble(real_i(tdom)): rho*pow(z,c);
     B -= log(z) * (c * (k1+A+1) + m);
   }
   else
     T = rho;
-  return B <= 0? 0: floor(0.9 + dblcoro526(a,c,B) / T * sqrt(N));
+  if (B <= 0) return 0;
+  A = floor(0.9 + dblcoro526(a,c,B) / T * sqrt(N));
+  if (dblexpo(A) >= BITS_IN_LONG-1) pari_err_OVERFLOW("lfunthetacost");
+  return (long)A;
 }
 long
 lfunthetacost0(GEN L, GEN tdom, long m, long bitprec)
