@@ -240,28 +240,6 @@ get_S_even(long N)
   }
 }
 
-static GEN Lfeq(long D, long k);
-/* Euler numbers: 1, 0, -1, 0, 5, 0, -61,... */
-GEN
-eulerfrac(long n)
-{
-  pari_sp av = avma;
-  if (n < 0) pari_err_DOMAIN("eulerfrac", "index", "<", gen_0, stoi(n));
-  if (odd(n)) return gen_0;
-  switch(n)
-  {
-    case 0: return gen_1;
-    case 2: return gen_m1;
-    case 4: return utoipos(5);
-    case 6: return utoineg(61);
-    case 8: return utoipos(1385);
-    case 10:return utoineg(50521);
-    case 12:return utoipos(2702765);
-    case 14:return utoineg(199360981);
-  }
-  return gerepileuptoint(av, gmul2n(Lfeq(-4, n+1), 1));
-}
-
 static GEN
 mfDcoefs(GEN F, GEN vD, long d)
 {
@@ -282,6 +260,7 @@ myinverseimage(GEN M, GEN R, GEN *pden)
   return c;
 }
 
+static GEN Lfeq(long D, long k);
 static GEN
 Hcol(GEN k, long r, GEN vD, long d, long N2)
 {
@@ -327,14 +306,14 @@ static GEN
 modulareven(long D, long r, long N0)
 {
   long B, d, i, l, N = labs(N0);
-  GEN V, vs, R, M, C, den, L, vP, vD, k = sstoQ(2*r+1, 2);
+  GEN V, vs, R, M, C, den, L, vP, vD, k = uutoQ(2*r+1, 2);
   SIGMA_F S = get_S_even(N0);
 
   d = dimeven(r, N);
   B = muleven(N) * mfsturmNgk(N, k);
   vD = Dpos(d, N0, B);
   vP = vecRCpol(r, d);
-  l = lg(vD); B = vD[l-1] / N; V = vecfactoru(1, B);
+  l = lg(vD); B = vD[l-1] / N; V = vecfactoru_i(1, B);
   vs = cgetg(B+2, t_VEC); gel(vs,1) = usumdivk_0_all(r, d);
   for (i = 1; i <= B; i++) gel(vs, i+1) = usumdivk_fact_all(gel(V,i), r, d);
   M = cgetg(l, t_MAT);
@@ -347,7 +326,7 @@ modulareven(long D, long r, long N0)
   if (r == 2*d)
   { /* r = 2 or (r = 4 and N = 4) */
     GEN v = mfDcoefs(mfderiv(mfTheta(NULL), d+1), vD, 1);
-    gel(M, d) = gadd(gel(M, d), gdivgs(v, N*(2*d - 1)));
+    gel(M, d) = gadd(gel(M, d), gdivgu(v, N*(2*d - 1)));
   }
   R = Hcol(k, r, vD, 1, (N == 8 || N0 == 12)? N >> 2: 1);
   /* Cost is O(d^2) * bitsize(result) ~ O(d^3.8) [heuristic] */
@@ -532,7 +511,7 @@ static GEN
 modularodd(long D, long r, long N0)
 {
   long B, d, i, l, dim, kro = kross(D, 2), Da = labs(D), N = labs(N0);
-  GEN V, vs, R, M, C, den, L, vP, vD, vD4, k = sstoQ(2*r+1, 2);
+  GEN V, vs, R, M, C, den, L, vP, vD, vD4, k = uutoQ(2*r+1, 2);
   SIGMA_Fodd S = get_S_odd(N);
 
   dim = dimodd(r, kro, N); d = (dim + 1) >> 1;
@@ -540,7 +519,7 @@ modularodd(long D, long r, long N0)
   B = mulodd(N, kro) * mfsturmNgk(4*N, k);
   vD = Dneg(B, kro, dim + 5, N);
   vD4 = kro ? vD : div4(vD);
-  l = lg(vD); B = vD4[l-1] / N; V = vecfactoru(1, B);
+  l = lg(vD); B = vD4[l-1] / N; V = vecfactoru_i(1, B);
   vs = cgetg(B+2, t_VEC); gel(vs,1) = NULL; /* unused */
   for (i = 1; i <= B; i++) gel(vs,i+1) = usumdivktwist_fact_all(gel(V,i), r, d);
   M = cgetg(l, t_MAT);
