@@ -163,7 +163,7 @@ Kderivsmallinit(GEN ldata, GEN Vga, long m, long bit)
   L = cgetg(N+1, t_VECSMALL);
   M = cgetg(N+1, t_VEC);
   mat = cgetg(N+1, t_VEC);
-  limn = ceil(2*M_LN2*bit / (d * dbllambertW0(C2/(M_PI*M_E))));
+  limn = ceil(2*M_LN2*bit / (d * dbllambertW0(C2/(M_PI*M_E*d))));
   l = limn + 2;
   for (j = 1; j <= N; j++)
   {
@@ -364,7 +364,7 @@ Kderivlarge(GEN K, GEN t, GEN t2d, long bitprec0)
     S = contfraceval_inv(M, z, nlim/2);
     if (DEBUGLEVEL>3)
     {
-      GEN S0 = contfraceval_inv(M, z, nlim/2 + 1);
+      GEN S0 = contfraceval_inv(M, z, minss(nlim/2 + 1, minss(lg(gel(M, 1)) - 1, lg(gel(M, 2)))));
       long e = gexpo(gmul(P, gsub(S,S0)));
       if (-e < bitprec0)
         err_printf("Kderivlarge: e = %ld, bit = %ld\n",e,bitprec0);
@@ -527,9 +527,23 @@ gammamellininvasymp_i(GEN Vga, long nlimmax, long m, long *status, long prec)
                          gmul(gel(M, n), gsub(A, uutoQ(n-1, d))));
   stripzeros(M); return M;
 }
+
+INLINE int
+RgV_is_CV(GEN x)
+{
+  long i;
+  for (i = lg(x)-1; i > 0; i--)
+  {
+    long t = typ(gel(x,i));
+    if (!is_real_t(t) && t!= t_COMPLEX) return 0;
+  }
+  return 1;
+}
+
 static GEN
 get_Vga(GEN x, GEN *ldata)
 {
+  if (typ(x)==t_VEC && RgV_is_CV(x)) { *ldata = NULL; return x; }
   *ldata = lfunmisc_to_ldata_shallow_i(x);
   if (*ldata) x = ldata_get_gammavec(*ldata);
   return x;
@@ -560,6 +574,20 @@ ishankelspec(GEN Vga, GEN M)
       if (!gequal(gel(Vga,i), v1)) break;
     if (i > d) return 1;
   }
+  else if (d==10 || d==14)
+  { /* [ a x 5 , (a+1) x 5 ] */
+    long d2 = d>>1;
+    long s0 = 1, s1 = 0, sm1 = 0;
+    GEN v1 = gel(Vga, 1);
+    for (i = 2; i <= d; i++)
+    {
+      GEN s = gsub(gel(Vga,i),v1);
+      if (gequal0(s)) s0++;
+      else if(gequal1(s)) s1++;
+      else if(gequalm1(s)) sm1++;
+    }
+    if (s0==d2 && (s1==d2 || sm1==d2)) return 1;
+  }
   status = 0;
   /* Heuristic: if 6 first terms in contfracinit don't fail, assume it's OK */
   pari_CATCH(e_INV) { status = 1; }
