@@ -136,8 +136,16 @@ grem(GEN x, GEN y)
         case t_POLMOD: return polmod_mod(x,y);
         case t_POL: return RgX_rem(x,y);
         case t_RFRAC:
-          av = avma; z = RgXQ_inv(RgX_rem(gel(x,2), y), y);
-          return gerepileupto(av, grem(gmul(gel(x,1), z), y));
+        {
+          GEN a = gel(x,1), b = gel(x,2), p, pol;
+          if (typ(a) == t_POL && RgX_is_ZX(y) && ZX_is_monic(y))
+          {
+            long pa, t = RgX_type2(a,b, &p,&pol,&pa);
+            if (t == t_FRAC || t == t_INT) return QXQ_div(a, b, y);
+          }
+          av = avma; z = RgXQ_inv(RgX_rem(b, y), y);
+          return gerepileupto(av, grem(gmul(a, z), y));
+        }
         case t_SER:
           if (RgX_is_monomial(y))
           {
@@ -313,22 +321,22 @@ QpXQX_to_ZXY(GEN f, GEN p)
 /*                                                                 */
 /*******************************************************************/
 
-/* f primitive ZX, squarefree, leading term prime to p; a in Z such that
+/* f primitive ZX, squarefree, leading term prime to p; 0 <= a < p such that
  * f(a) = 0 mod p. Return p-adic roots of f equal to a mod p, in
  * precision >= prec */
 GEN
 ZX_Zp_root(GEN f, GEN a, GEN p, long prec)
 {
-  GEN z, R, a0 = modii(a, p);
+  GEN z, R;
   long i, j, k;
 
-  if (signe(FpX_eval(FpX_deriv(f, p), a0, p)))
+  if (signe(FpX_eval(FpX_deriv(f, p), a, p)))
   { /* simple zero mod p, go all the way to p^prec */
-    if (prec > 1) a0 = ZpX_liftroot(f, a0, p, prec);
-    return mkcol(a0);
+    if (prec > 1) a = ZpX_liftroot(f, a, p, prec);
+    return mkcol(a);
   }
 
-  f = ZX_unscale_div(RgX_translate(f,a), p); /* f(pX + a) / p */
+  f = ZX_unscale_div(ZX_translate(f,a), p); /* f(pX + a) / p */
   (void)ZX_pvalrem(f,p,&f);
   z = cgetg(degpol(f)+1,t_COL);
 
@@ -347,12 +355,14 @@ Zp_appr(GEN f, GEN a)
 {
   pari_sp av = avma;
   GEN z, p = gel(a,2);
-  long prec = gequal0(a)? valp(a): precp(a);
+  long v = valp(a), prec = v;
 
+  if (signe(gel(a,4))) prec += precp(a);
   f = QpX_to_ZX(f, p);
   if (degpol(f) <= 0) pari_err_CONSTPOL("Zp_appr");
+  if (v < 0) pari_err_DOMAIN("padicappr", "v(a)", "<", gen_0, stoi(v));
   f = ZX_radical(f);
-  a = padic_to_Q(a);
+  a = padic_to_Fp(a, p);
   if (signe(FpX_eval(f,a,p))) { set_avma(av); return cgetg(1,t_COL); }
   z = ZX_Zp_root(f, a, p, prec);
   return gerepilecopy(av, ZV_to_ZpV(z, p, prec));
@@ -373,7 +383,7 @@ pnormalize(GEN f, GEN p, GEN T, long prec, long n,
     if (v1 < v)
     {
       *prev = 1;
-      f = RgX_recip_shallow(f); /* f(0) != 0 so degree is the same */
+      f = RgX_recip_i(f); /* f(0) != 0 so degree is the same */
      /* beware loss of precision from lc(factor), whose valuation is <= v */
       *pprec += v; v = v1;
     }
@@ -434,7 +444,7 @@ ZXY_ZpQ_root(GEN f, GEN a, GEN T, GEN p, long prec)
 GEN
 padicappr(GEN f, GEN a)
 {
-  GEN p, z, T;
+  GEN p, z, T, Tp;
   long prec;
   pari_sp av = avma;
 
@@ -457,8 +467,9 @@ padicappr(GEN f, GEN a)
   /* ensure that f /= (f,f') is separable */
   (void)nfgcd_all(f, RgX_deriv(f), T, NULL, &f);
 
-  if (!gequal0(FqX_eval(FqX_red(f,T,p), a, T,p))) /* check f(a) = 0 (mod p,T) */
-  { set_avma(av); return cgetg(1,t_COL); }
+  Tp = FpX_red(T, p); a = FqX_red(a, Tp, p);
+  if (!gequal0(FqX_eval(FqX_red(f,Tp,p), a, Tp,p)))
+  { set_avma(av); return cgetg(1,t_COL); } /* f(a) != 0 (mod p,T) */
   z = ZXY_ZpQ_root(f, a, T, p, prec);
   return gerepilecopy(av, ZXV_to_ZpXQV(z, T, p, prec));
 }
@@ -474,7 +485,7 @@ ZX_Zp_roots(GEN f, GEN p, long prec)
   r = FpX_roots(f, p);
   l = lg(r); if (l == 1) return r;
   for (i = 1; i < l; i++) gel(r,i) = ZX_Zp_root(f, gel(r,i), p, prec);
-  settyp(r, t_VEC); return ZV_to_ZpV(shallowconcat1(r), p, prec);
+  return ZV_to_ZpV(shallowconcat1(r), p, prec);
 }
 /* vector of p-adic roots of the ZXX f, leading term prime to p. Shallow */
 static GEN
@@ -487,7 +498,7 @@ ZXY_ZpQ_roots(GEN f, GEN T, GEN p, long prec)
   r = FqX_roots(f, FpX_red(T,p), p);
   l = lg(r); if (l == 1) return r;
   for (i = 1; i < l; i++) gel(r,i) = ZXY_ZpQ_root(f, gel(r,i), T, p, prec);
-  settyp(r, t_VEC); return ZXV_to_ZpXQV(shallowconcat1(r), T, p, prec);
+  return ZXV_to_ZpXQV(shallowconcat1(r), T, p, prec);
 }
 
 /* return p-adic roots of f, precision prec */
@@ -547,10 +558,9 @@ famat_flatten(GEN fa)
   {
     GEN p = gel(P,i);
     long e = itou(gel(E,i));
-    gel(y,i) = const_vec(e, p);
+    gel(y,i) = const_col(e, p);
   }
-  y = shallowconcat1(y); settyp(y, t_COL);
-  return mkmat2(y, const_col(lg(y)-1, gen_1));
+  y = shallowconcat1(y); return mkmat2(y, const_col(lg(y)-1, gen_1));
 }
 
 GEN
@@ -587,7 +597,7 @@ factorpadic(GEN f, GEN p, long r)
     for (i=1; i<l; i++)
     {
       GEN t = gel(P,i);
-      if (reverse) t = normalizepol(RgX_recip_shallow(t));
+      if (reverse) t = RgX_recip_shallow(t);
       gel(P,i) = ZX_to_ZpX_normalized(t,p,ppow,r);
     }
   }
