@@ -359,7 +359,7 @@ quadunit_mod(GEN D, GEN N)
     GEN M = shifti(mulii(q, N), 1);
     quadunit_uvmod(D, d, M, &u, &v);
     u = diviiexact(u, q);
-    v = diviiexact(v, q); u = shifti(u,-1);
+    v = modii(diviiexact(v, q), N); u = shifti(u,-1);
   }
   return deg1pol_shallow(v, u, 0);
 }
@@ -738,7 +738,7 @@ unegquadclassnoF(ulong x, ulong *pD)
   GEN E, P;
   ulong D = coredisc2u_fact(factoru(x), -1, &P, &E);
   long H = uquadclassnoF_fact(D, -1, P, E);
-  switch(D)
+  if (x != D) switch(D)
   { /* divide by [ O_K^* : O^* ] */
     case 4: H >>= 1; break;
     case 3: H /= 3; break;
