@@ -2,40 +2,38 @@
 #define __AGM_ATAN_LIMIT                 56
 #define __DIVRR_GMP_LIMIT                -1
 #define __EXPNEWTON_LIMIT                66
-#define __F2x_MUL_KARATSUBA_LIMIT        15
-#define __F2xqX_BARRETT_LIMIT            17
-#define __F2xqX_DIVREM_BARRETT_LIMIT     46
-#define __F2xqX_INVBARRETT_LIMIT         22
-#define __F2xqX_REM_BARRETT_LIMIT        48
-#define __Flx_BARRETT_HALFMULII_LIMIT    29
-#define __Flx_BARRETT_KARATSUBA_LIMIT    2561
-#define __Flx_BARRETT_MULII2_LIMIT       30
-#define __Flx_BARRETT_MULII_LIMIT        1681
-#define __Flx_BARRETT_QUARTMULII_LIMIT   29
-#define __Flx_DIVREM_BARRETT_LIMIT       2804
-#define __Flx_EXTGCD_LIMIT               284
-#define __Flx_GCD_LIMIT                  1890
-#define __Flx_HALFGCD_HALFMULII_LIMIT    60
-#define __Flx_HALFGCD_KARATSUBA_LIMIT    58
-#define __Flx_HALFGCD_MULII2_LIMIT       64
-#define __Flx_HALFGCD_MULII_LIMIT        90
-#define __Flx_HALFGCD_QUARTMULII_LIMIT   60
-#define __Flx_INVBARRETT_HALFMULII_LIMIT 424
-#define __Flx_INVBARRETT_KARATSUBA_LIMIT 5120
-#define __Flx_INVBARRETT_MULII2_LIMIT    36
-#define __Flx_INVBARRETT_MULII_LIMIT     3840
-#define __Flx_INVBARRETT_QUARTMULII_LIMIT 424
-#define __Flx_MUL_HALFMULII_LIMIT        5
-#define __Flx_MUL_KARATSUBA_LIMIT        147
-#define __Flx_MUL_MULII2_LIMIT           5
-#define __Flx_MUL_MULII_LIMIT            1639
-#define __Flx_MUL_QUARTMULII_LIMIT       5
-#define __Flx_REM_BARRETT_LIMIT          3577
-#define __Flx_SQR_HALFSQRI_LIMIT         3
-#define __Flx_SQR_KARATSUBA_LIMIT        330
-#define __Flx_SQR_QUARTSQRI_LIMIT        3
-#define __Flx_SQR_SQRI2_LIMIT            8
-#define __Flx_SQR_SQRI_LIMIT             5
+#define __EXTGCD_HALFGCD_LIMIT           548
+#define __F2x_MUL_KARATSUBA_LIMIT        11
+#define __F2x_MUL_MULII_LIMIT            662
+#define __F2xqX_BARRETT_LIMIT            56
+#define __F2xqX_DIVREM_BARRETT_LIMIT     127
+#define __F2xqX_EXTGCD_LIMIT             127
+#define __F2xqX_GCD_LIMIT                1292
+#define __F2xqX_HALFGCD_LIMIT            127
+#define __F2xqX_INVBARRETT_LIMIT         50
+#define __F2xqX_REM_BARRETT_LIMIT        127
+#define __Flx_BARRETT2_LIMIT             63
+#define __Flx_BARRETT_LIMIT              131
+#define __Flx_DIVREM2_BARRETT_LIMIT      28
+#define __Flx_DIVREM_BARRETT_LIMIT       407
+#define __Flx_EXTGCD2_LIMIT              79
+#define __Flx_EXTGCD_LIMIT               267
+#define __Flx_GCD2_LIMIT                 4121
+#define __Flx_GCD_LIMIT                  2060
+#define __Flx_HALFGCD2_LIMIT             58
+#define __Flx_HALFGCD_LIMIT              265
+#define __Flx_INVBARRETT2_LIMIT          28
+#define __Flx_INVBARRETT_LIMIT           316
+#define __Flx_MUL2_KARATSUBA_LIMIT       10
+#define __Flx_MUL2_MULII_LIMIT           14
+#define __Flx_MUL_KARATSUBA_LIMIT        38
+#define __Flx_MUL_MULII_LIMIT            66
+#define __Flx_REM2_BARRETT_LIMIT         241
+#define __Flx_REM_BARRETT_LIMIT          407
+#define __Flx_SQR2_KARATSUBA_LIMIT       15
+#define __Flx_SQR2_SQRI_LIMIT            28
+#define __Flx_SQR_KARATSUBA_LIMIT        94
+#define __Flx_SQR_SQRI_LIMIT             77
 #define __FlxqX_BARRETT_LIMIT            17
 #define __FlxqX_DIVREM_BARRETT_LIMIT     46
 #define __FlxqX_EXTGCD_LIMIT             44
@@ -59,6 +57,8 @@
 #define __FpX_REM_BARRETT_LIMIT          306
 #define __Fp_POW_BARRETT_LIMIT           101
 #define __Fp_POW_REDC_LIMIT              99
+#define __GCD_HALFGCD_LIMIT              141
+#define __HALFGCD_LIMIT                  6
 #define __INVMOD_GMP_LIMIT               -1
 #define __INVNEWTON_LIMIT                550
 #define __LOGAGMCX_LIMIT                 13
@@ -75,40 +75,38 @@
 #define __AGM_ATAN_LIMIT                 159
 #define __DIVRR_GMP_LIMIT                -1
 #define __EXPNEWTON_LIMIT                66
-#define __F2x_MUL_KARATSUBA_LIMIT        23
-#define __F2xqX_BARRETT_LIMIT            17
-#define __F2xqX_DIVREM_BARRETT_LIMIT     46
-#define __F2xqX_INVBARRETT_LIMIT         22
-#define __F2xqX_REM_BARRETT_LIMIT        48
-#define __Flx_BARRETT_HALFMULII_LIMIT    244
-#define __Flx_BARRETT_KARATSUBA_LIMIT    905
-#define __Flx_BARRETT_MULII2_LIMIT       1004
-#define __Flx_BARRETT_MULII_LIMIT        2715
-#define __Flx_BARRETT_QUARTMULII_LIMIT   244
-#define __Flx_DIVREM_BARRETT_LIMIT       3942
-#define __Flx_EXTGCD_LIMIT               850
-#define __Flx_GCD_LIMIT                  7165
-#define __Flx_HALFGCD_HALFMULII_LIMIT    116
-#define __Flx_HALFGCD_KARATSUBA_LIMIT    230
-#define __Flx_HALFGCD_MULII2_LIMIT       1045
-#define __Flx_HALFGCD_MULII_LIMIT        168
-#define __Flx_HALFGCD_QUARTMULII_LIMIT   116
-#define __Flx_INVBARRETT_HALFMULII_LIMIT 898
-#define __Flx_INVBARRETT_KARATSUBA_LIMIT 3471
-#define __Flx_INVBARRETT_MULII2_LIMIT    3672
-#define __Flx_INVBARRETT_MULII_LIMIT     4196
-#define __Flx_INVBARRETT_QUARTMULII_LIMIT 898
-#define __Flx_MUL_HALFMULII_LIMIT        8
-#define __Flx_MUL_KARATSUBA_LIMIT        85
-#define __Flx_MUL_MULII2_LIMIT           3755
-#define __Flx_MUL_MULII_LIMIT            698
-#define __Flx_MUL_QUARTMULII_LIMIT       8
-#define __Flx_REM_BARRETT_LIMIT          3942
-#define __Flx_SQR_HALFSQRI_LIMIT         6
-#define __Flx_SQR_KARATSUBA_LIMIT        159
-#define __Flx_SQR_QUARTSQRI_LIMIT        6
-#define __Flx_SQR_SQRI2_LIMIT            4139
-#define __Flx_SQR_SQRI_LIMIT             1276
+#define __EXTGCD_HALFGCD_LIMIT           320
+#define __F2x_MUL_KARATSUBA_LIMIT        13
+#define __F2x_MUL_MULII_LIMIT            6265
+#define __F2xqX_BARRETT_LIMIT            58
+#define __F2xqX_DIVREM_BARRETT_LIMIT     127
+#define __F2xqX_EXTGCD_LIMIT             127
+#define __F2xqX_GCD_LIMIT                1292
+#define __F2xqX_HALFGCD_LIMIT            78
+#define __F2xqX_INVBARRETT_LIMIT         38
+#define __F2xqX_REM_BARRETT_LIMIT        127
+#define __Flx_BARRETT2_LIMIT             102
+#define __Flx_BARRETT_LIMIT              433
+#define __Flx_DIVREM2_BARRETT_LIMIT      194
+#define __Flx_DIVREM_BARRETT_LIMIT       1888
+#define __Flx_EXTGCD2_LIMIT              313
+#define __Flx_EXTGCD_LIMIT               1357
+#define __Flx_GCD2_LIMIT                 3407
+#define __Flx_GCD_LIMIT                  8245
+#define __Flx_HALFGCD2_LIMIT             401
+#define __Flx_HALFGCD_LIMIT              817
+#define __Flx_INVBARRETT2_LIMIT          1396
+#define __Flx_INVBARRETT_LIMIT           1233
+#define __Flx_MUL2_KARATSUBA_LIMIT       9
+#define __Flx_MUL2_MULII_LIMIT           21
+#define __Flx_MUL_KARATSUBA_LIMIT        57
+#define __Flx_MUL_MULII_LIMIT            601
+#define __Flx_REM2_BARRETT_LIMIT         296
+#define __Flx_REM_BARRETT_LIMIT          1888
+#define __Flx_SQR2_KARATSUBA_LIMIT       19
+#define __Flx_SQR2_SQRI_LIMIT            67
+#define __Flx_SQR_KARATSUBA_LIMIT        109
+#define __Flx_SQR_SQRI_LIMIT             469
 #define __FlxqX_BARRETT_LIMIT            17
 #define __FlxqX_DIVREM_BARRETT_LIMIT     46
 #define __FlxqX_EXTGCD_LIMIT             44
@@ -132,6 +130,8 @@
 #define __FpX_REM_BARRETT_LIMIT          306
 #define __Fp_POW_BARRETT_LIMIT           97
 #define __Fp_POW_REDC_LIMIT              99
+#define __GCD_HALFGCD_LIMIT              170
+#define __HALFGCD_LIMIT                  12
 #define __INVMOD_GMP_LIMIT               -1
 #define __INVNEWTON_LIMIT                66
 #define __LOGAGMCX_LIMIT                 58
