@@ -19,8 +19,8 @@ ASM addll mulll bfffo divll
 /* Written by Bill Allombert from the ix86 version by Bruno Haible. Basically
  * change insl to insq*/
 #ifdef ASMINLINE
-#define LOCAL_HIREMAINDER  register ulong hiremainder
-#define LOCAL_OVERFLOW     register ulong overflow
+#define LOCAL_HIREMAINDER  ulong hiremainder
+#define LOCAL_OVERFLOW     ulong overflow
 
 #define addll(a,b) \
 __extension__ ({ ulong __value, __arg1 = (a), __arg2 = (b); \
