@@ -419,13 +419,13 @@ sd_colors(const char *v, long flag)
     l = strlen(v);
     if (l <= 2 && strncmp(v, "no", l) == 0)
       v = "";
-    if (l <= 6 && strncmp(v, "darkbg", l) == 0)
-      v = "1, 5, 3, 7, 6, 2, 3"; /* Assume recent ReadLine. */
-    if (l <= 7 && strncmp(v, "lightbg", l) == 0)
-      v = "1, 6, 3, 4, 5, 2, 3"; /* Assume recent ReadLine. */
-    if (l <= 8 && strncmp(v, "brightfg", l) == 0)      /* Good for windows consoles */
+    else if (l <= 6 && strncmp(v, "darkbg", l) == 0)
+      v = "1, 5, 3, 7, 6, 2, 3"; /* assume recent readline. */
+    else if (l <= 7 && strncmp(v, "lightbg", l) == 0)
+      v = "1, 6, 3, 4, 5, 2, 3"; /* assume recent readline. */
+    else if (l <= 8 && strncmp(v, "brightfg", l) == 0) /* windows console */
       v = "9, 13, 11, 15, 14, 10, 11";
-    if (l <= 6 && strncmp(v, "boldfg", l) == 0)        /* Good for darkbg consoles */
+    else if (l <= 6 && strncmp(v, "boldfg", l) == 0) /* darkbg console */
       v = "[1,,1], [5,,1], [3,,1], [7,,1], [6,,1], , [2,,1]";
     s = gp_filter(v);
     for (c=c_ERR; c < c_LAST; c++) gp_colors[c] = gp_get_color(&s);
@@ -516,11 +516,15 @@ sd_secure(const char *v, long flag)
 
 GEN
 sd_debug(const char *v, long flag)
-{ return sd_ulong(v,flag,"debug",&DEBUGLEVEL, 0,20,NULL); }
+{
+  GEN r = sd_ulong(v,flag,"debug",&DEBUGLEVEL, 0,20,NULL);
+  if (v) setalldebug(DEBUGLEVEL);
+  return r;
+}
 
 GEN
 sd_debugfiles(const char *v, long flag)
-{ return sd_ulong(v,flag,"debugfiles",&DEBUGFILES, 0,20,NULL); }
+{ return sd_ulong(v,flag,"debugfiles",&DEBUGLEVEL_io, 0,20,NULL); }
 
 GEN
 sd_debugmem(const char *v, long flag)
