@@ -167,7 +167,7 @@ print_fun_list(char **list, long nbli)
   if (i) pari_putc('\n');
 }
 
-static const long MAX_SECTION = 16;
+static const long MAX_SECTION = 17;
 static void
 commands(long n)
 {
@@ -249,8 +249,7 @@ gentypes(void)
   t_POL    : polynomials       [ cod1 ] [ cod2 ] [ man_1 ] ... [ man_k ]\n\
   t_SER    : power series      [ cod1 ] [ cod2 ] [ man_1 ] ... [ man_k ]\n\
   t_RFRAC  : irred. rat. func. [ code ] [ num. ] [ den. ]\n\
-  t_QFR    : real qfb          [ code ] [ a ] [ b ] [ c ] [ del ]\n\
-  t_QFI    : imaginary qfb     [ code ] [ a ] [ b ] [ c ]\n\
+  t_QFB    : qfb               [ code ] [ a ] [ b ] [ c ] [ disc ]\n\
   t_VEC    : row vector        [ code ] [  x_1  ] ... [  x_k  ]\n\
   t_COL    : column vector     [ code ] [  x_1  ] ... [  x_k  ]\n\
   t_MAT    : matrix            [ code ] [ col_1 ] ... [ col_k ]\n\
@@ -282,6 +281,7 @@ menu_commands(void)
   "Associative and central simple ALGEBRAS",
   "ELLIPTIC CURVES",
   "L-FUNCTIONS",
+  "HYPERGEOMETRIC MOTIVES",
   "MODULAR FORMS",
   "MODULAR SYMBOLS",
   "GRAPHIC functions",
@@ -483,12 +483,21 @@ gphelp_keyword_list(void)
   "bnr",
   "ell",
   "rnf",
+  "hgm",
+  "HGM",
+  "ideal",
+  "idele",
+  "CFT",
   "bid",
   "modulus",
   "prototype",
   "Lmath",
   "Ldata",
   "Linit",
+  "character",
+  "sums",
+  "products",
+  "integrals",
   NULL};
   return L;
 }
@@ -934,6 +943,22 @@ parse_key_val(char *src, char **ps, char **pt)
   if (*t == '"') (void)pari_translate_string(t, t, src);
   *s_end = 0; *ps = src; *pt = t;
 }
+/* parse src of the form (s,t) (or "s", or "t"), set *ps to s, and *pt to t. */
+static void
+parse_key_val_paren(char *src, char **ps, char **pt)
+{
+  char *s, *t, *s_end, *t_end;
+  s = t = src + 1; while (*t && *t != ',') t++;
+  if (*t != ',') err_gprc("missing ','",t,src);
+  s_end = t;
+  t++; while (*t && *t != ')') t++;
+  if (*t != ')') err_gprc("missing ')'",t,src);
+  if (t[1])  err_gprc("unexpected character",t+1,src);
+  t_end = t; t = s_end + 1;
+  if (*t == '"') (void)pari_translate_string(t, t, src);
+  if (*s == '"') (void)pari_translate_string(s, s, src);
+  *s_end = 0; *t_end = 0; *ps = s; *pt = t;
+}
 
 void
 gp_initrc(pari_stack *p_A)
@@ -986,6 +1011,16 @@ gp_initrc(pari_stack *p_A)
         if (*s == '"') (void)pari_translate_string(s, t, s-4); else strcpy(t,s);
         pari_stack_pushp(p_A,t);
       }
+      else if (!strncmp(s, "default(", 8))
+      {
+        s += 7; parse_key_val_paren(s, &s,&t);
+        (void)setdefault(s,t,d_INITRC);
+      }
+      else if (!strncmp(s, "setdebug(", 9))
+      {
+        s += 8; parse_key_val_paren(s, &s,&t);
+        setdebug(s, atol(t));
+      }
       else
       { /* set default */
         parse_key_val(s, &s,&t);
@@ -1733,6 +1768,7 @@ print_all_user_fun(int member)
   set_avma(av);
 }
 
+/* get_sep, removing enclosing quotes */
 static char *
 get_name(const char *s)
 {
@@ -1744,18 +1780,22 @@ get_name(const char *s)
   }
   return t;
 }
+static void
+ack_debug(const char *s, long d) {pari_printf("   debug(\"%s\") = %ld\n",s,d);}
+static void
+ack_setdebug(const char *s, long d) {setdebug(s, d); ack_debug(s, d);}
 
 static void
 escape(const char *tch, int ismain)
 {
   const char *s = tch;
+  long d;
   char c;
+  GEN x;
   switch ((c = *s++))
   {
     case 'w': case 'x': case 'a': case 'b': case 'B': case 'm':
     { /* history things */
-      long d;
-      GEN x;
       if (c != 'w' && c != 'x') d = get_int(s,0);
       else
       {
@@ -1785,12 +1825,40 @@ escape(const char *tch, int ismain)
       if (!*s) s = (GP_DATA->echo)? "0": "1";
       (void)sd_echo(s,d_ACKNOWLEDGE); break;
     case 'g':
-      switch (*s)
-      {
-        case 'm': s++; (void)sd_debugmem(*s? s: NULL,d_ACKNOWLEDGE); break;
-        case 'f': s++; (void)sd_debugfiles(*s? s: NULL,d_ACKNOWLEDGE); break;
-        default : (void)sd_debug(*s? s: NULL,d_ACKNOWLEDGE); break;
-      }
+        if (isdigit(*s))
+        {
+          const char *t = s + 1;
+          if (isdigit(*t)) t++; /* atol(s) < 99 */
+          t = get_name(t);
+          if (*t) { d = atol(s); ack_setdebug(t, d); break; }
+        }
+        else if (*s == '"' || isalpha(*s))
+        {
+          char *t = get_name(s);
+          if (t[1] && !isdigit(t[1]))
+          {
+            char *T = t + strlen(t) - 1;
+            if (isdigit(*T))
+            {
+              if (isdigit(T[-1])) T--; /* < 99 */
+              d = atol(T); *T = 0;
+              ack_setdebug(get_name(t), d); /* get_name in case of ".." */
+            }
+            else
+            {
+              x = setdebug(t, -1); ack_debug(t, itos(x));
+            }
+          }
+          else switch (*t)
+          {
+            case 'm':
+              s++; (void)sd_debugmem(*s? s: NULL,d_ACKNOWLEDGE); break;
+            case 'f':
+              s++; (void)sd_debugfiles(*s? s: NULL,d_ACKNOWLEDGE); break;
+          }
+          break;
+        }
+        (void)sd_debug(*s? s: NULL,d_ACKNOWLEDGE); break;
       break;
     case 'h': print_functions_hash(s); break;
     case 'l':
