@@ -52,6 +52,7 @@ Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA. */
 #if defined(USE_TIMES)
 #  include <sys/times.h>
 #endif
+#define PARI_INIT
 #include "pari.h"
 #include "paripriv.h"
 #include "anal.h"
@@ -84,13 +85,13 @@ static const ulong readonly_constants[] = {
   (ulong)(readonly_constants+4),
   (ulong)(readonly_constants+7)
 };
-THREAD GEN zetazone, bernzone, primetab;
+THREAD GEN zetazone, bernzone, eulerzone, primetab;
 byteptr diffptr;
 FILE    *pari_outfile, *pari_errfile, *pari_logfile, *pari_infile;
 char    *current_logfile, *current_psfile, *pari_datadir;
 long    gp_colors[c_LAST];
 int     disable_color;
-ulong   DEBUGFILES, DEBUGLEVEL, DEBUGMEM;
+ulong   DEBUGLEVEL, DEBUGMEM;
 long    DEBUGVAR;
 ulong   pari_mt_nbthreads;
 long    precreal;
@@ -276,6 +277,20 @@ blocksearch(GEN x, GEN bl)
   return NULL; /* Unknown address */
 }
 
+static int
+check_clone(GEN x)
+{
+  GEN bl = root_block;
+  if (isonstack(x) || is_universal_constant(x)) return 1;
+  while (bl)
+  {
+    if (x >= bl  && x < bl + bl_size(bl))
+      return 1;
+    bl = x < bl ? bl_left(bl): bl_right(bl);
+  }
+  return 0; /* Unknown address */
+}
+
 void
 clone_lock(GEN x)
 {
@@ -773,6 +788,13 @@ pari_init_errcatch(void)
   global_err_data = NULL;
 }
 
+void
+setalldebug(long n)
+{
+  long i, l = numberof(pari_DEBUGLEVEL_ptr);
+  for (i = 0; i < l; i++) *pari_DEBUGLEVEL_ptr[i] = n;
+}
+
 /*********************************************************************/
 /*                           INIT DEFAULTS                           */
 /*********************************************************************/
@@ -789,7 +811,8 @@ pari_init_defaults(void)
 #endif
 
   precdl = 16;
-  DEBUGFILES = DEBUGLEVEL = 0;
+  DEBUGLEVEL = 0;
+  setalldebug(0);
   DEBUGMEM = 1;
   disable_color = 1;
   pari_logstyle = logstyle_none;
@@ -1175,10 +1198,12 @@ pari_thread_init(void)
   pari_init_errcatch();
   pari_init_rand();
   pari_init_floats();
+  pari_init_hgm();
   pari_init_parser();
   pari_init_compiler();
   pari_init_evaluator();
   pari_init_files();
+  pari_init_ellcondfile();
 }
 
 void
@@ -1189,6 +1214,7 @@ pari_thread_close(void)
   pari_close_compiler();
   pari_close_parser();
   pari_close_floats();
+  pari_close_hgm();
   pari_close_blocks();
 }
 
@@ -1198,6 +1224,7 @@ pari_thread_start(struct pari_thread *t)
   pari_mainstack_use(&t->st);
   pari_thread_init();
   pari_thread_set_global(&t->gs);
+  mt_thread_init();
   return t->data;
 }
 
@@ -1639,7 +1666,7 @@ type_dim(GEN x)
     case t_MAT:
     {
       long l = lg(x), r = (l == 1)? 1: lgcols(x);
-      sprintf(v, "t_MAT (%ldx%ld)", r-1,l-1);
+      sprintf(v, "t_MAT (%ld x %ld)", r-1,l-1);
       break;
     }
     case t_COL:
@@ -2448,22 +2475,33 @@ obj_init(long d, long n)
   gel(S, d+1) = zerovec(n);
   return S;
 }
-/* insert O in S [last position] at position K, return it */
-GEN
-obj_insert(GEN S, long K, GEN O)
-{ return obj_insert_shallow(S, K, gclone(O)); }
+
 /* as obj_insert. WITHOUT cloning (for libpari, when creating a *new* obj
  * from an existing one) */
 GEN
-obj_insert_shallow(GEN S, long K, GEN O)
+obj_insert(GEN S, long K, GEN O)
 {
   GEN o, v = gel(S, lg(S)-1);
   if (typ(v) != t_VEC) pari_err_TYPE("obj_insert", S);
+  if (!check_clone(v))
+  {
+    if (DEBUGLEVEL) pari_warn(warner,"trying to update parent object");
+    return gclone(O);
+  }
   o = gel(v,K);
-  gel(v,K) = O; /*SIGINT: before unclone(o)*/
+  gel(v,K) = gclone(O); /*SIGINT: before unclone(o)*/
   if (isclone(o)) gunclone(o); return gel(v,K);
 }
 
+GEN
+obj_insert_shallow(GEN S, long K, GEN O)
+{
+  GEN v = gel(S, lg(S)-1);
+  if (typ(v) != t_VEC) pari_err_TYPE("obj_insert", S);
+  gel(v,K) = O;
+  return gel(v,K);
+}
+
 /* Does S [last position] contain data at position K ? Return it, or NULL */
 GEN
 obj_check(GEN S, long K)
@@ -2895,6 +2933,35 @@ getwalltime(void)
 /*                   FUNCTIONS KNOWN TO THE ANALYZER               */
 /*                                                                 */
 /*******************************************************************/
+
+GEN
+setdebug(const char *s, long n)
+{
+  long i, l = numberof(pari_DEBUGLEVEL_str);
+  GEN V, V1, V2;
+  if (s)
+  {
+    if (n > 20)
+      pari_err_DOMAIN("setdebug", "n", ">", utoipos(20), stoi(n));
+    for (i = 0; i < l; i++)
+      if (!strcmp(s, pari_DEBUGLEVEL_str[i])) break;
+    if (i == l)
+      pari_err_DOMAIN("setdebug", s, "not a valid",
+                      strtoGENstr("debug domain"), strtoGENstr(s));
+    if (n >= 0) { *pari_DEBUGLEVEL_ptr[i] = n; return gnil; }
+    return stoi(*pari_DEBUGLEVEL_ptr[i]);
+  }
+  V = cgetg(3,t_MAT);
+  V1 = gel(V,1) = cgetg(l+1, t_COL);
+  V2 = gel(V,2) = cgetg(l+1, t_COL);
+  for (i = 0; i < l; i++)
+  {
+    gel(V1, i+1) = strtoGENstr(pari_DEBUGLEVEL_str[i]);
+    gel(V2, i+1) = stoi(*pari_DEBUGLEVEL_ptr[i]);
+  }
+  return V;
+}
+
 GEN
 pari_version(void)
 {
