@@ -65,6 +65,8 @@ Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA. */
 #include "pari.h"
 #include "paripriv.h"
 
+#define DEBUGLEVEL DEBUGLEVEL_mpqs
+
 /** DEBUG **/
 /* #define MPQS_DEBUG_VERBOSE 1 */
 #include "mpqs.h"
@@ -74,7 +76,6 @@ Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA. */
 #define MAX_PE_PAIR 60
 
 #ifdef HAS_SSE2
-#define AND(a,b) ((a)&(b))
 #define EXT0(a) ((ulong)__builtin_ia32_vec_ext_v2di((__v2di)(a), 0))
 #define EXT1(a) ((ulong)__builtin_ia32_vec_ext_v2di((__v2di)(a), 1))
 #define TEST(a) (EXT0(a) || EXT1(a))
@@ -83,7 +84,6 @@ const mpqs_bit_array mpqs_mask = { (long) 0x8080808080808080L, (long) 0x80808080
 #else
 /* Use ulong for the bit arrays */
 typedef ulong mpqs_bit_array;
-#define AND(a,b) ((a)&(b))
 #define TEST(a) (a)
 
 #ifdef LONG_IS_64BIT
@@ -326,6 +326,7 @@ mpqs_create_FB(mpqs_handle_t *h, ulong *f)
   mpqs_uint32_t k = h->_k->k;
   forprime_t S;
 
+  h->largest_FB_p = 0; /* -Wall */
   FB[2].fbe_p = 2;
   /* the fbe_logval and the fbe_sqrt_kN for 2 are never used */
   FB[2].fbe_flags = MPQS_FBE_CLEAR;
@@ -891,7 +892,7 @@ INLINE void
 mpqs_sieve_p(unsigned char *B, unsigned char *E, long p4, long p,
              unsigned char logp)
 {
-  register unsigned char *e = E - p4;
+  unsigned char *e = E - p4;
   /* Unrolled loop. It might be better to let the compiler worry about this
    * kind of optimization, based on its knowledge of whatever useful tricks the
    * machine instruction set architecture is offering */
@@ -921,7 +922,7 @@ INLINE void
 mpqs_sieve_p2(unsigned char *B, unsigned char *E, long p4, long s1, long s2,
              unsigned char logp)
 {
-  register unsigned char *e = E - p4;
+  unsigned char *e = E - p4;
   /* Unrolled loop. It might be better to let the compiler worry about this
    * kind of optimization, based on its knowledge of whatever useful tricks the
    * machine instruction set architecture is offering */
@@ -995,7 +996,7 @@ mpqs_eval_sieve(mpqs_handle_t *h)
   while (count < MPQS_CANDIDATE_ARRAY_SIZE - 1)
   {
     long j, y;
-    while (!TEST(AND(U[x],mpqs_mask))) x++;
+    while (!TEST(U[x]&mpqs_mask)) x++;
     y = x*sizemask;
     for (j=0; j<sizemask; j++, y++)
     {
