@@ -20,6 +20,8 @@ Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA. */
 #include "pari.h"
 #include "paripriv.h"
 
+#define DEBUGLEVEL DEBUGLEVEL_stark
+
 /* ComputeCoeff */
 typedef struct {
   GEN L0, L1, L11, L2; /* VECSMALL of p */
@@ -221,7 +223,7 @@ LiftChar(GEN Qt, GEN cyc, GEN chi)
 
 /* Let C be a subgroup, system of representatives of the quotient */
 static GEN
-ag_subgroup_classes(GEN C)
+subgroup_classes(GEN C)
 {
   GEN U, D = ZM_snfall_i(C, &U, NULL, 1), e = cyc2elts(D);
   long i, l = lg(e);
@@ -238,7 +240,7 @@ ag_subgroup_classes(GEN C)
 
 /* Let s: A -> B given by [P,cycA,cycB] A and B, compute the kernel of s. */
 GEN
-ag_kernel(GEN S)
+abmap_kernel(GEN S)
 {
   GEN U, P = gel(S,1), cycA = gel(S,2), DB = diagonal_shallow(gel(S,3));
   long nA = lg(cycA)-1, rk;
@@ -248,7 +250,7 @@ ag_kernel(GEN S)
 }
 /* let H be a subgroup of A; return s(H) */
 GEN
-ag_subgroup_image(GEN S, GEN H)
+abmap_subgroup_image(GEN S, GEN H)
 { return ZM_hnfmodid(ZM_mul(gel(S,1), H),  gel(S,3)); }
 
 /* Let m and n be two moduli such that n|m and let C be a congruence
@@ -260,7 +262,7 @@ ComputeKernel(GEN bnrm, GEN bnrn, GEN dtQ)
   pari_sp av = avma;
   GEN S = bnrsurjection(bnrm, bnrn);
   GEN P = ZM_mul(gel(dtQ,3), gel(S,1));
-  return gerepileupto(av, ag_kernel(mkvec3(P, gel(S,2), gel(dtQ,2))));
+  return gerepileupto(av, abmap_kernel(mkvec3(P, gel(S,2), gel(dtQ,2))));
 }
 
 static long
@@ -301,7 +303,7 @@ IsGoodSubgroup(GEN H, GEN bnr, GEN map)
   PH = divcond(bnrH);
   S = bnrsurjection(bnr, bnrH);
   /* H as a subgroup of bnrH */
-  iH = ag_subgroup_image(S, p1);
+  iH = abmap_subgroup_image(S, p1);
   qH = InitQuotient(iH);
   for (j = 1; j < lg(P); j++)
   {
@@ -310,7 +312,7 @@ IsGoodSubgroup(GEN H, GEN bnr, GEN map)
     if (tablesearch(PH, pr, cmp_prime_ideal)) continue;
     /* inertia degree of pr in bnr(modH)/H is charorder(e, cycH) */
     e = ZM_ZC_mul(gel(qH,3), isprincipalray(bnrH, pr));
-    e = vecmodii(e, gel(qH,2));
+    e = ZV_ZV_mod(e, gel(qH,2));
     if (ZV_equal0(e)) return gc_long(av,0); /* f = 1 */
   }
   return gc_long(av,1);
@@ -487,7 +489,7 @@ FindModulus(GEN bnr, GEN dtQ, long *newprec)
               if (!ok) continue;
             }
             CR = InitChar(bnrm, AllChars(bnrm, QD, 1), 0, DEFAULTPREC);
-            data = mkvec4(bnrm, D, ag_subgroup_classes(Cm), CR);
+            data = mkvec4(bnrm, D, subgroup_classes(Cm), CR);
             if (DEBUGLEVEL>1)
               err_printf("\nTrying modulus = %Ps and subgroup = %Ps\n",
                          bnr_get_mod(bnrm), D);
@@ -1078,7 +1080,7 @@ EvalCoeff(GEN z, int* c, long deg)
   return e;
 }
 
-/* copy the n * (m+1) array matan */
+/* a2 <- copy the n x m array */
 static void
 CopyCoeff(int** a, int** a2, long n, long m)
 {
@@ -1176,15 +1178,13 @@ ComputeCoeff(GEN dtcr, LISTray *R, long n, long deg)
 /********************************************************************/
 static void
 deg11(LISTray *R, long p, GEN bnr, GEN pr) {
-  GEN z = isprincipalray(bnr, pr);
   vecsmalltrunc_append(R->L1, p);
-  vectrunc_append(R->L1ray, z);
+  vectrunc_append(R->L1ray, isprincipalray(bnr, pr));
 }
 static void
-deg12(LISTray *R, long p, GEN bnr, GEN Lpr) {
-  GEN z = isprincipalray(bnr, gel(Lpr,1));
+deg12(LISTray *R, long p, GEN bnr, GEN pr) {
   vecsmalltrunc_append(R->L11, p);
-  vectrunc_append(R->L11ray, z);
+  vectrunc_append(R->L11ray, isprincipalray(bnr, pr));
 }
 static void
 deg0(LISTray *R, long p) { vecsmalltrunc_append(R->L0, p); }
@@ -1195,50 +1195,58 @@ static void
 InitPrimesQuad(GEN bnr, ulong N0, LISTray *R)
 {
   pari_sp av = avma;
-  GEN bnf = bnr_get_bnf(bnr), cond = gel(bnr_get_mod(bnr), 1);
-  long p,i,l, condZ = itos(gcoeff(cond,1,1)), contZ = itos(content(cond));
-  GEN prime, Lpr, nf = bnf_get_nf(bnf), dk = nf_get_disc(nf);
+  GEN bnf = bnr_get_bnf(bnr), F = gel(bnr_get_mod(bnr), 1);
+  GEN v, N, nf = bnf_get_nf(bnf), dk = nf_get_disc(nf);
+  long l = 1 + primepi_upper_bound(N0);
+  ulong i, p, FZ = itou(gcoeff(F,1,1)), FZ2 = itou(gcoeff(F,2,2));
   forprime_t T;
 
-  l = 1 + primepi_upper_bound(N0);
+  FZ2 = ugcd(FZ, FZ2); /* content(F) */
   R->L0 = vecsmalltrunc_init(l);
-  R->L2 = vecsmalltrunc_init(l); R->condZ = condZ;
+  R->L2 = vecsmalltrunc_init(l); R->condZ = FZ;
   R->L1 = vecsmalltrunc_init(l); R->L1ray = vectrunc_init(l);
   R->L11= vecsmalltrunc_init(l); R->L11ray= vectrunc_init(l);
-  prime = utoipos(2);
+  N = utoipos(2);
   u_forprime_init(&T, 2, N0);
   while ( (p = u_forprime_next(&T)) )
   {
-    prime[2] = p;
+    N[2] = p;
     switch (kroiu(dk, p))
     {
     case -1: /* inert */
-      if (condZ % p == 0) deg0(R,p); else deg2(R,p);
+      if (FZ % p == 0) deg0(R,p); else deg2(R,p);
       break;
     case 1: /* split */
-      Lpr = idealprimedec(nf, prime);
-      if      (condZ % p != 0) deg12(R, p, bnr, Lpr);
-      else if (contZ % p == 0) deg0(R,p);
+      if (FZ2 % p == 0) deg0(R,p);
       else
       {
-        GEN pr = idealval(nf, cond, gel(Lpr,1))? gel(Lpr,2): gel(Lpr,1);
-        deg11(R, p, bnr, pr);
+        GEN Lpr = idealprimedec(nf, N);
+        if (FZ % p) deg12(R, p, bnr, gel(Lpr,1));
+        else
+        {
+          long t = ZC_prdvd(gel(F,2), gel(Lpr,1))? 2: 1;
+          deg11(R, p, bnr, gel(Lpr,t));
+        }
       }
       break;
     default: /* ramified */
-      if (condZ % p == 0)
+      if (FZ % p == 0)
         deg0(R,p);
       else
-        deg11(R, p, bnr, idealprimedec_galois(nf,prime));
+        deg11(R, p, bnr, idealprimedec_galois(nf, N));
       break;
     }
   }
   /* precompute isprincipalray(x), x in Z */
-  R->rayZ = cgetg(condZ, t_VEC);
-  for (i=1; i<condZ; i++)
-    gel(R->rayZ,i) = (ugcd(i,condZ) == 1)? isprincipalray(bnr, utoipos(i)): gen_0;
+  v = coprimes_zv(FZ);
+  R->rayZ = cgetg(FZ, t_VEC);
+  for (i = 1; i < FZ; i++)
+  {
+    N[2] = i;
+    gel(R->rayZ,i) = v[i]? isprincipalray(bnr, N): gen_0;
+  }
   gerepileall(av, 7, &(R->L0), &(R->L2), &(R->rayZ),
-              &(R->L1), &(R->L1ray), &(R->L11), &(R->L11ray) );
+              &(R->L1), &(R->L1ray), &(R->L11), &(R->L11ray));
 }
 
 static void
@@ -1547,7 +1555,7 @@ RecCoeff2(GEN nf,  RC_data *d,  long prec)
 
   d->G = minss(-20, -prec2nbits(prec) >> 4);
 
-  vec  = shallowconcat(mkvec(gneg(beta)), row(M, d->v));
+  vec  = vec_prepend(row(M, d->v), gneg(beta));
   min = (long)prec2nbits_mul(prec, 0.75);
   max = (long)prec2nbits_mul(prec, 0.98);
   av = avma;
@@ -2248,7 +2256,7 @@ LABDOUB:
       p1 = real_0(newprec); p2 = gel(ch_CHI(dtcr), 2);
       for (j = 1; j <= n; j++)
         if ( (an = EvalCoeff(p2, matan[j], d)) )
-          p1 = gadd(p1, gdivgs(an, j));
+          p1 = gadd(p1, gdivgu(an, j));
       gel(L1,i) = gerepileupto(av2, p1);
       FreeMat(matan, n);
     }
@@ -2276,7 +2284,7 @@ LABDOUB:
       if (chi_get_deg(CHI) != 2) t = gmul2n(t, 1); /* character not real */
       z = gadd(z, t);
     }
-    gel(vzeta,i) = gmul2n(gcosh(gdivgs(z,den), newprec), 1);
+    gel(vzeta,i) = gmul2n(gcosh(gdivgu(z,den), newprec), 1);
   }
   polrelnum = roots_to_pol(vzeta, 0);
   if (DEBUGLEVEL)
@@ -2506,7 +2514,7 @@ makescind(GEN nf, GEN P)
       ulong o = perm_orderu(perm);
       if (o != 2) perm = perm_powu(perm, o >> 1);
       /* perm has order two and doesn't belong to Gal(H_k/k) */
-      return galoisfixedfield(G, perm, 1, varn(P));
+      return polredbest(galoisfixedfield(G, perm, 1, varn(P)), 0);
     }
   }
   pari_err_BUG("makescind");
@@ -2516,16 +2524,14 @@ makescind(GEN nf, GEN P)
 /* pbnf = NULL if no bnf is needed, f = NULL may be passed for a trivial
  * conductor */
 static void
-quadray_init(GEN *pD, GEN f, GEN *pbnf, long prec)
+quadray_init(GEN *pD, GEN *pbnf, long prec)
 {
   GEN D = *pD, nf, bnf = NULL;
   if (typ(D) == t_INT)
   {
     int isfund;
     if (pbnf) {
-      long v = f? gvar(f): NO_VARIABLE;
-      if (v == NO_VARIABLE) v = 1;
-      bnf = Buchall(quadpoly0(D, v), nf_FORCE, prec);
+      bnf = Buchall(quadpoly0(D, 1), nf_FORCE, prec);
       nf = bnf_get_nf(bnf);
       isfund = equalii(D, nf_get_disc(nf));
     }
@@ -2550,16 +2556,15 @@ quadray_init(GEN *pD, GEN f, GEN *pbnf, long prec)
 static GEN
 quadhilbertreal(GEN D, long prec)
 {
-  pari_sp av = avma;
-  GEN bnf, pol, bnr, dtQ, data, M;
+  GEN bnf, bnr, dtQ, data, M;
   long newprec;
   pari_timer T;
 
-  quadray_init(&D, NULL, &bnf, prec);
+  quadray_init(&D, &bnf, prec);
   switch(itou_or_0(cyc_get_expo(bnf_get_cyc(bnf))))
   {
-    case 1: set_avma(av); return pol_x(0);
-    case 2: return gerepileupto(av, GenusFieldQuadReal(D));
+    case 1: return pol_x(0);
+    case 2: return GenusFieldQuadReal(D);
   }
   bnr  = Buchray(bnf, gen_1, nf_INIT);
   M = diagonal_shallow(bnr_get_cyc(bnr));
@@ -2568,10 +2573,8 @@ quadhilbertreal(GEN D, long prec)
   if (DEBUGLEVEL) timer_start(&T);
   data = FindModulus(bnr, dtQ, &newprec);
   if (DEBUGLEVEL) timer_printf(&T,"FindModulus");
-  if (!data) return gerepileupto(av, bnrstark_cyclic(bnr, dtQ, prec));
-  pol = AllStark(data, 0, newprec);
-  pol = makescind(bnf_get_nf(bnf), pol);
-  return gerepileupto(av, polredbest(pol, 0));
+  if (!data) return bnrstark_cyclic(bnr, dtQ, prec);
+  return makescind(bnf_get_nf(bnf), AllStark(data, 0, newprec));
 }
 
 /*******************************************************************/
@@ -2702,7 +2705,7 @@ init_pq(GEN D, struct gpq_data *T)
       break;
     }
     if (kroiu(D, q) < 0) continue; /* inert */
-    Q = redimag(primeform_u(D, q));
+    Q = qfbred_i(primeform_u(D, q));
     if (is_pm1(gel(Q,1))) continue; /* Q | q is principal */
 
     store = 1;
@@ -2756,12 +2759,11 @@ static GEN
 gpq(GEN form, struct gpq_data *T)
 {
   pari_sp av = avma;
-  long a = form[1], b = form[2], c = form[3];
-  long p = T->p, q = T->q;
+  long a = form[1], b = form[2], c = form[3], p = T->p, q = T->q;
   GEN form2, w, z;
   int fl, real = 0;
 
-  form2 = qficomp(T->qfpq, mkvec3s(a, -b, c));
+  form2 = qfbcomp_i(T->qfpq, mkqfb(stoi(a), stoi(-b), stoi(c), T->D));
   /* form2 and form yield complex conjugate roots : only compute for the
    * lexicographically smallest of the 2 */
   fl = cmpis(gel(form2,1), a);
@@ -2807,7 +2809,6 @@ static GEN
 quadhilbertimag(GEN D)
 {
   GEN L, P, Pi, Pr, qfp, u;
-  pari_sp av = avma;
   long h, i, prec;
   struct gpq_data T;
   pari_timer ti;
@@ -2841,17 +2842,17 @@ quadhilbertimag(GEN D)
   T.pq2 = shifti(T.pq,1);
   if (T.p == T.q)
   {
-    GEN qfbp2 = qficompraw(qfp, qfp);
+    GEN qfbp2 = qfbcompraw(qfp, qfp);
     u = gel(qfbp2,2);
     T.u = modii(u, T.pq2);
-    T.qfpq = redimag(qfbp2);
+    T.qfpq = qfbred_i(qfbp2);
   }
   else
   {
     GEN qfq = primeform_u(D, T.q), bp = gel(qfp,2), bq = gel(qfq,2);
     T.u = Z_chinese(bp, bq, utoipos(T.p << 1), utoipos(T.q << 1));
     /* T.u = bp (mod 2p), T.u = bq (mod 2q) */
-    T.qfpq = qficomp(qfp, qfq);
+    T.qfpq = qfbcomp_i(qfp, qfq);
   }
   /* u modulo 2pq */
   prec = LOWDEFAULTPREC;
@@ -2881,16 +2882,17 @@ quadhilbertimag(GEN D)
     set_avma(av0); prec += nbits2extraprec(prec2nbits(DEFAULTPREC)+exmax);
     if (DEBUGLEVEL) pari_warn(warnprec,"quadhilbertimag",prec);
   }
-  return gerepileupto(av,P);
+  return P;
 }
 
 GEN
 quadhilbert(GEN D, long prec)
 {
+  pari_sp av = avma;
   GEN d = D;
-  quadray_init(&d, NULL, NULL, 0);
-  return (signe(d)>0)? quadhilbertreal(D,prec)
-                     : quadhilbertimag(d);
+  quadray_init(&d, NULL, 0);
+  return gerepileupto(av, signe(d)>0? quadhilbertreal(D,prec)
+                                    : quadhilbertimag(d));
 }
 
 /* return a vector of all roots of 1 in bnf [not necessarily quadratic] */
@@ -3058,7 +3060,7 @@ ellphistinit(GEN om, long prec)
   if (gsigne(imag_i(gdiv(om1,om2))) < 0) { swap(om1,om2); om = mkvec2(om1,om2); }
   om1b = conj_i(om1);
   om2b = conj_i(om2); res = cgetg(4,t_VEC);
-  gel(res,1) = gdivgs(elleisnum(om,2,0,prec),12);
+  gel(res,1) = gdivgu(elleisnum(om,2,0,prec),12);
   gel(res,2) = gdiv(PiI2(prec), gmul(om2, imag_i(gmul(om1b,om2))));
   gel(res,3) = om2b; return res;
 }
@@ -3101,7 +3103,7 @@ PRECPB:
   if (!first)
   {
     if (DEBUGLEVEL) pari_warn(warnprec,"computeP2",prec);
-    nf = gerepilecopy(av2, nfnewprec_shallow(checknf(bnr),prec));
+    nf = gerepilecopy(av2, nfnewprec_shallow(bnr_get_nf(bnr),prec));
   }
   first = 0; lanum = to_approx(nf,la);
   P = cgetg(clrayno+1,t_VEC);
@@ -3174,17 +3176,20 @@ findquad_pol(GEN p, GEN a, GEN x)
   for (i=2; i<lx; i++) gel(y,i) = findquad(a, gel(x,i), p);
   y[1] = x[1]; return y;
 }
+/* m is 3, 4 or 12 */
 static GEN
-compocyclo(GEN nf, long m, long d)
+compocyclo(GEN D, long m)
+{ return do_compo(quadhilbertimag(D), polcyclo(m,0)); }
+/* m is prime or 4 * prime */
+static GEN
+compocyclop(GEN nf, long m)
 {
   GEN sb,a,b,s,p1,p2,p3,res,polL,polLK,nfL, D = nf_get_disc(nf);
   long ell,vx;
 
   p1 = quadhilbertimag(D);
   p2 = polcyclo(m,0);
-  if (d==1) return do_compo(p1,p2);
-
-  ell = m&1 ? m : (m>>2);
+  ell = odd(m)? m: (m>>2); /* prime */
   if (absequalui(ell,D)) /* ell = |D| */
   {
     p2 = gcoeff(nffactor(nf,p2),1,1);
@@ -3192,7 +3197,7 @@ compocyclo(GEN nf, long m, long d)
   }
   if (ell%4 == 3) ell = -ell;
   /* nf = K = Q(a), L = K(b) quadratic extension = Q(t) */
-  polLK = quadpoly(stoi(ell)); /* relative polynomial */
+  polLK = quadpoly_i(stoi(ell)); /* relative polynomial */
   res = rnfequation2(nf, polLK);
   vx = nf_get_varn(nf);
   polL = gsubst(gel(res,1),0,pol_x(vx)); /* = charpoly(t) */
@@ -3247,10 +3252,10 @@ treatspecialsigma(GEN bnr)
   Ds = smodis(D,48);
   if (i)
   {
-    if (i==2 && Ds%16== 8) return compocyclo(nf, 4,1);
-    if (i==3 && Ds% 3== 1) return compocyclo(nf, 3,1);
-    if (i==4 && Ds% 8== 1) return compocyclo(nf, 4,1);
-    if (i==6 && Ds   ==40) return compocyclo(nf,12,1);
+    if (i==2 && Ds%16== 8) return compocyclo(D, 4);
+    if (i==3 && Ds% 3== 1) return compocyclo(D, 3);
+    if (i==4 && Ds% 8== 1) return compocyclo(D, 4);
+    if (i==6 && Ds   ==40) return compocyclo(D,12);
     return NULL;
   }
 
@@ -3264,7 +3269,7 @@ treatspecialsigma(GEN bnr)
   }
   if (tryf <= 3 || umodiu(D, tryf) || !uisprime(tryf)) return NULL;
   if (fl) tryf <<= 2;
-  return compocyclo(nf,tryf,2);
+  return compocyclop(nf, tryf);
 }
 
 GEN
@@ -3274,7 +3279,9 @@ quadray(GEN D, GEN f, long prec)
   pari_sp av = avma;
 
   if (isint1(f)) return quadhilbert(D, prec);
-  quadray_init(&D, f, &bnf, prec);
+  if (typ(D) == t_INT && typ(f) != t_INT)
+    pari_err_TYPE("quadray [conductor]", f);
+  quadray_init(&D, &bnf, prec);
   bnr = Buchray(bnf, f, nf_INIT|nf_GEN);
   if (is_pm1(bnr_get_no(bnr))) { set_avma(av); return pol_x(0); }
   if (signe(D) > 0)
