@@ -15,6 +15,8 @@ Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA. */
 #include "pari.h"
 #include "paripriv.h"
 
+#define DEBUGLEVEL DEBUGLEVEL_thue
+
 /********************************************************************/
 /**                                                                **/
 /**             THUE EQUATION SOLVER (G. Hanrot)                   **/
@@ -86,7 +88,7 @@ LogHeight(GEN x, long prec)
     GEN t = gabs(gel(x,i), prec);
     if (gcmpgs(t,1) > 0) LH = gmul(LH, t);
   }
-  return gerepileupto(av, gdivgs(glog(LH,prec), n));
+  return gerepileupto(av, gdivgu(glog(LH,prec), n));
 }
 
 /* |x|^(1/n), x t_INT */
@@ -139,18 +141,18 @@ T_A_Matrices(GEN MatFU, long r, GEN *eps5, long prec)
   m1 = matslice(MatFU, 1,r, 1,r); /* minor order r */
   m1 = glog(gabs(m1, prec), prec); /* HIGH accuracy */
   A = RgM_inv(m1); if (!A) pari_err_PREC("thue");
-  IntM = RgM_Rg_add(RgM_mul(A,m1), gen_m1);
+  IntM = RgM_Rg_add_shallow(RgM_mul(A,m1), gen_m1);
   IntM = gabs(IntM, 0);
 
   eps2 = gadd(vecmax(IntM), real2n(-e, LOWDEFAULTPREC)); /* t_REAL */
   nia = vecmax(gabs(A, 0));
 
   /* Check for the precision in matrix inversion. See paper, Lemma 2.4.2. */
-  p1 = addrr(mulsr(r, gmul2n(nia, e)), eps2); /* t_REAL */
+  p1 = addrr(mulur(r, gmul2n(nia, e)), eps2); /* t_REAL */
   if (expo(p1) < -2*r) pari_err_PREC("thue");
 
-  p1 = addrr(mulsr(r, gmul2n(nia,-e)), eps2);
-  eps3 = mulrr(mulsr(2*r*r,nia), p1);
+  p1 = addrr(mulur(r, gmul2n(nia,-e)), eps2);
+  eps3 = mulrr(mulur(2*r*r,nia), p1);
   if (!signe(eps3))
     eps3 = real2n(expo(eps3), LOWDEFAULTPREC);
   else
@@ -167,7 +169,7 @@ static GEN
 get_prime_info(GEN bnf)
 {
   long n = 1, e = gexpo(bnf_get_reg(bnf)), nbp = e < 20? 1: 2;
-  GEN L = cgetg(nbp+1, t_VEC), nf = checknf(bnf), fu = bnf_get_fu(bnf);
+  GEN L = cgetg(nbp+1, t_VEC), nf = bnf_get_nf(bnf), fu = bnf_get_fu(bnf);
   GEN X = pol_x(nf_get_varn(nf));
   ulong p;
   for(p = 2147483659UL; n <= nbp; p = unextprime(p+1))
@@ -382,13 +384,13 @@ LLL_1stPass(GEN *pB0, GEN kappa, baker_s *BS, GEN *pBx)
   long e;
 
   C = grndtoi(mulir(mulii(BS->Ind, kappa),
-                    gpow(B0, dbltor(2.2), DEFAULTPREC)), &e);
+                    gpow(B0, dbltor(2.2), DEFAULTPREC)), NULL);
 
   if (DEBUGLEVEL > 1) err_printf("C (bitsize) : %d\n", expi(C));
   lllmat = matid(3);
   if (cmpri(B0, BS->Ind) > 0)
   {
-    gcoeff(lllmat, 1, 1) = grndtoi(divri(B0, BS->Ind), &e);
+    gcoeff(lllmat, 1, 1) = grndtoi(divri(B0, BS->Ind), NULL);
     triv = shiftr(sqrr(B0), 1);
   }
   else
@@ -453,7 +455,7 @@ CheckSol(GEN *pS, GEN z1, GEN z2, GEN P, GEN rhs, GEN ro)
   if (e > 0) return 0;
   if (e <= -13)
   { /* y != 0 and rhs != 0; check whether P(x,y) = rhs or P(-x,-y) = rhs */
-    GEN z = poleval(ZX_rescale(P,y),x);
+    GEN z = ZX_Z_eval(ZX_rescale(P,y),x);
     if (absequalii(z, rhs)) add_pm(pS, x,y, z, degpol(P), rhs);
   }
   return 1;
@@ -634,7 +636,7 @@ MiddleSols(GEN *pS, GEN bound, GEN roo, GEN P, GEN rhs, long s, GEN c1)
       if (cmpii(q, bound) > 0) break;
       if (DEBUGLEVEL >= 2) err_printf("Checking (+/- %Ps, +/- %Ps)\n",p, q);
       av = avma;
-      z = poleval(ZX_rescale(P,q), p); /* = P(p/q) q^dep(P) */
+      z = ZX_Z_eval(ZX_rescale(P,q), p); /* = P(p/q) q^dep(P) */
       Q = dvmdii(rhs, z, &R);
       if (R != gen_0) { set_avma(av); continue; }
       setabssign(Q);
@@ -744,15 +746,6 @@ fact(double x)
   return ft ;
 }
 
-static GEN
-RgX_homogenize(GEN P, long v)
-{
-  GEN Q = leafcopy(P);
-  long i, l = lg(P), d = degpol(P);
-  for (i = 2; i < l; i++) gel(Q,i) = monomial(gel(Q,i), d--, v);
-  return Q;
-}
-
 /* Compute all relevant constants needed to solve the equation P(x,y)=a given
  * the solutions of N_{K/Q}(x)=a (see inithue). */
 GEN
@@ -932,7 +925,7 @@ get_B0(long i1, GEN Delta2, GEN Lambda, GEN Deps5, long prec, baker_s *BS)
         l0 = subrr(errnum(BS->delta, q), ep);
         if (signe(l0) <= 0) break;
 
-        B0 = divrr(mplog(divrr(mulir(gel(Q,2), BS->c15), l0)),  BS->c13);
+        B0 = divrr(logr_abs(divrr(mulir(gel(Q,2), BS->c15), l0)),  BS->c13);
         if (DEBUGLEVEL>1) err_printf("Semirat. reduction: B0 -> %Ps\n",B0);
       }
       /* if no progress, stop */
@@ -1204,13 +1197,12 @@ maybe_warn(GEN bnf, GEN a, GEN Ind)
   if (!is_pm1(Ind) && !is_pm1(bnf_get_no(bnf)) && !is_pm1(a))
     pari_warn(warner, "The result returned by 'thue' is conditional on the GRH");
 }
-static GEN bnfisintnorm_fa(GEN bnf, GEN a, GEN fa, long sa);
-/* return solutions of Norm(x) = a mod U(K)^+ */
+/* true bnf; return solutions of Norm(x) = a mod U(K)^+ */
 static GEN
 get_ne(GEN bnf, GEN a, GEN fa, GEN Ind)
 {
   if (DEBUGLEVEL) maybe_warn(bnf,a,Ind);
-  return bnfisintnorm_fa(bnf, a, fa, signe(a));
+  return bnfisintnorm_i(bnf, a, signe(a), bnfisintnormabs(bnf, mkvec2(a, fa)));
 }
 /* return solutions of |Norm(x)| = |a| mod U(K) */
 static GEN
@@ -1330,7 +1322,7 @@ thue(GEN tnf, GEN rhs, GEN ne)
           GEN D = coredisc(ZX_disc(POL));
           if (cmpis(D, -4) >= 0)
           {
-            GEN F, T = quadpoly(D);
+            GEN F, T = quadpoly_i(D);
             w = equalis(D, -4)? 4: 6;
             setvarn(T, fetch_var_higher());
             F = gcoeff(nffactor(POL, T), 1, 1);
@@ -1511,9 +1503,8 @@ isintnorm_loop(struct sol_abs *T, long i)
 }
 
 static int
-get_sol_abs(struct sol_abs *T, GEN bnf, GEN fact, GEN *ptPR)
+get_sol_abs(struct sol_abs *T, GEN bnf, GEN nf, GEN fact, GEN *ptPR)
 {
-  GEN nf = bnf_get_nf(bnf);
   GEN P = gel(fact,1), E = gel(fact,2), PR;
   long N = nf_get_degree(nf), nP = lg(P)-1, Ngen, max, nPR, i, j;
 
@@ -1521,7 +1512,7 @@ get_sol_abs(struct sol_abs *T, GEN bnf, GEN fact, GEN *ptPR)
   T->f = new_chunk(max+1);
   T->n = new_chunk(max+1);
   T->next = new_chunk(max+1);
-  *ptPR = PR = cgetg(max+1, t_VEC); /* length to be fixed later */
+  *ptPR = PR = cgetg(max+1, t_COL); /* length to be fixed later */
 
   nPR = 0;
   for (i = 1; i <= nP; i++)
@@ -1553,8 +1544,8 @@ get_sol_abs(struct sol_abs *T, GEN bnf, GEN fact, GEN *ptPR)
 
   T->u = cgetg(nPR+1, t_VECSMALL);
   T->S = new_chunk(nPR+1);
-  T->cyc = bnf_get_cyc(bnf);
-  Ngen = lg(T->cyc)-1;
+  if (bnf) { T->cyc = bnf_get_cyc(bnf); Ngen = lg(T->cyc)-1; }
+  else { T->cyc = NULL; Ngen = 0; }
   if (Ngen == 0)
     T->rel = T->partrel = NULL; /* trivial Cl(K), no relations to check */
   else
@@ -1619,11 +1610,11 @@ bnfisintnormabs(GEN bnf, GEN a)
     a = typ(a) == t_VEC? gel(a,1): factorback(F);
     if (signe(a) < 0) F = clean_Z_factor(F);
   }
-  bnf = checkbnf(bnf); nf = bnf_get_nf(bnf);
+  nf = bnf_get_nf(bnf);
   if (!signe(a)) return mkvec(gen_0);
   if (is_pm1(a)) return mkvec(gen_1);
   if (!F) F = absZ_factor(a);
-  if (!get_sol_abs(&T, bnf, F, &PR)) return cgetg(1, t_VEC);
+  if (!get_sol_abs(&T, bnf, nf, F, &PR)) return cgetg(1, t_VEC);
   /* |a| > 1 => T.nPR > 0 */
   res = cgetg(T.sindex+1, t_VEC);
   for (i=1; i<=T.sindex; i++)
@@ -1635,11 +1626,34 @@ bnfisintnormabs(GEN bnf, GEN a)
   return res;
 }
 
-/* z = bnfisintnormabs(bnf,a), sa = 1 or -1, return bnfisintnorm(bnf,sa*|a|) */
+/* true nf */
+GEN
+ideals_by_norm(GEN nf, GEN a)
+{
+  struct sol_abs T;
+  GEN res, PR, F;
+  long i;
+
+  if ((F = check_arith_pos(a,"ideals_by_norm")))
+    a = typ(a) == t_VEC? gel(a,1): factorback(F);
+  if (is_pm1(a)) return mkvec(trivial_fact());
+  if (!F) F = absZ_factor(a);
+  if (!get_sol_abs(&T, NULL, nf, F, &PR)) return cgetg(1, t_VEC);
+  res = cgetg(T.sindex+1, t_VEC);
+  for (i=1; i<=T.sindex; i++)
+  {
+    GEN x = vecsmall_to_col( gel(T.normsol,i) );
+    gel(res,i) = famat_remove_trivial(mkmat2(PR, x));
+  }
+  return res;
+}
+
+/* true bnf; z = bnfisintnormabs(bnf,a), sa = 1 or -1,
+ * return bnfisintnorm(bnf,sa*|a|) */
 static GEN
 bnfisintnorm_i(GEN bnf, GEN a, long sa, GEN z)
 {
-  GEN nf = checknf(bnf), T = nf_get_pol(nf), f = nf_get_index(nf), unit = NULL;
+  GEN nf = bnf_get_nf(bnf), T = nf_get_pol(nf), f = nf_get_index(nf), unit=NULL;
   GEN Tp, A = signe(a) == sa? a: negi(a);
   long sNx, i, j, N = degpol(T), l = lg(z);
   long norm_1 = 0; /* gcc -Wall */
@@ -1684,16 +1698,13 @@ bnfisintnorm_i(GEN bnf, GEN a, long sa, GEN z)
   }
   setlg(z, j); return z;
 }
-/* bnfisintnorm sa * |a|, fa = factor(|a|) */
-static GEN
-bnfisintnorm_fa(GEN bnf, GEN a, GEN fa, long sa)
-{ return bnfisintnorm_i(bnf, a, sa, bnfisintnormabs(bnf, mkvec2(a, fa)));
-}
 GEN
 bnfisintnorm(GEN bnf, GEN a)
 {
   pari_sp av = avma;
-  GEN ne = bnfisintnormabs(bnf,a);
+  GEN ne;
+  bnf = checkbnf(bnf);
+  ne = bnfisintnormabs(bnf,a);
   switch(typ(a))
   {
     case t_VEC: a = gel(a,1); break;
