@@ -4,7 +4,8 @@ This file is part of the PARI/GP package.
 
 PARI/GP is free software; you can redistribute it and/or modify it under the
 terms of the GNU General Public License as published by the Free Software
-Foundation. It is distributed in the hope that it will be useful, but WITHOUT
+Foundation; either version 2 of the License, or (at your option) any later
+version. It is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY WHATSOEVER.
 
 Check the License for details. You should have received a copy of it, along
@@ -63,7 +64,7 @@ emscripten_draw(PARI_plot *T, GEN w, GEN x, GEN y)
   EM_ASM(rawPrint=true);(void)T;
   pari_printf("<img src=\"data:image/svg+xml;charset=utf-8;base64,%Ps\">\n", svg);
   EM_ASM(rawPrint=false);
-  avma = av;
+  set_avma(av);
 }
 
 static long plot_width, plot_height;
